/* Claerbout's conjugate-gradient iteration. */
/*
  Copyright (C) 2004 University of Texas at Austin
  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "stdafx.h"
#include "cdstep.h"

#include "cblas_RSF.h"
//#include "mkl_cblas.h"


using namespace Yang;

/* precision */
static const float EPSILON=1.e-12;

Yang::cdstepper::cdstepper(int nx, int ny): m_nx(nx), m_ny(ny), m_bAllocated(false) {
}

int Yang::cdstepper::step( bool forget     /* restart flag */, 
		int nx, int ny  /* model size, data size */, 
		float* x        /* current model [nx] */, 
		const float* g  /* gradient [nx] */, 
		float* rr       /* data residual [ny] */, 
		const float* gg /* conjugate gradient [ny], the peturbation of resiudal corresponding to model perturbation g */) 
/*< Step of conjugate-gradient iteration. >*/
{
  double sds, gdg, gds, determ, gdr, sdr, alfa, beta;
  int i;
  if (!m_bAllocated) {
    m_bAllocated = forget = true;
    this->S.resize(nx); this->ss.resize(ny);
  }
  //PRV2(nx,ny);
  //VERIFY(m_nx==nx); VERIFY(m_ny==ny);
  
  float g_nrm,s_nrm;
  g_nrm = sqrtf(cblas_dsdot(nx,g,1,g,1)); //try to normalize the gradient of input.
  s_nrm = sqrtf(cblas_dsdot(nx,va_ptr(S),1,va_ptr(S),1));
  if (forget) {
    this->S  =  0.f;
    this->ss = 0.f;
    beta = 0.0f;

    gdg = cblas_dsdot( ny, gg, 1, gg, 1); //<dr,dr>
    /* Solve G . ( R + G*alfa) = 0 */ //the optimatl step size for steepest descent search
    if (gdg <= 0.f) return -1;  //the gradient is invalid, it leads the residual perturbation to be zero.
    alfa = - cblas_dsdot( ny, gg, 1, rr, 1) / gdg;
  } else {
    // search plane by solving 2-by-2
    // G . (R + G*alfa + S*beta) = 0
    // S . (R + G*alfa + S*beta) = 0
    gdg = cblas_dsdot( ny, gg, 1, gg, 1);       
    sds = cblas_dsdot( ny, va_ptr(ss), 1, va_ptr(ss), 1);       
    gds = cblas_dsdot( ny, gg, 1, va_ptr(ss), 1);       
    if (gdg < 1e-12 || sds < 1e-12 ) return -2;
    determ = 1.0 - (gds/gdg)*(gds/sds);
    if (determ > EPSILON) determ *= gdg * sds;
    else determ = gdg * sds * EPSILON;
    gdr = - cblas_dsdot( ny, gg, 1, rr, 1);
    sdr = - cblas_dsdot( ny, va_ptr(ss), 1, rr, 1);
    alfa = ( sds * gdr - gds * sdr ) / determ;
    beta = (-gds * gdr + gdg * sdr ) / determ;
  }
  
  cblas_sscal(nx,beta,va_ptr(S),1);
  cblas_saxpy(nx,alfa,g,1,va_ptr(S),1);   //update the current search direction S

  cblas_sscal(ny,beta,va_ptr(ss),1);    
  cblas_saxpy(ny,alfa,gg,1,va_ptr(ss),1); //update dr corresponding to S

  //finally update current model and residual
  for (i = 0; i < nx; i++) {
    x[i] +=  this->S[i];
  }
  for (i = 0; i < ny; i++) {
    rr[i] += this->ss[i];
  }
  
  //Calculate the normalized values
  beta = beta * s_nrm; alfa = alfa * g_nrm;
  prt("normalized alfa=%e, beta(step on previous direct)=%e \n", alfa, beta);
  return 0;
}

