#include "stdafx.h"
#include "commonOps.h"
#include "cubeAlgo.h"
#include <omp.h>

using namespace Yang;
using namespace std;
using boost::shared_ptr;

//fwd: sample a bunch of points from a regular grid
//Adj: spread a bunch of points to a regular grid

void opInterpLin::oper(cubeData<float>& mod, cubeData<float>& dat, bool bAdj, bool bAdd) {
  addNull(mod,dat,bAdj,bAdd);
  if (!bAdj) {
    irregularSamp1d(dat, mod, this->mCoord, false, this->m_bExtend,bAdd);
  } else {
    irregularSamp1d(mod, dat, this->mCoord, true, this->m_bExtend,bAdd);
  }
}
 
void opInterpLin::init(const cubeData<float>& coordCube, const hypercube& regularGridShape, bool bExtend ) {
  this->setShape(regularGridShape,coordCube.domain());
  coordCube.copyTo(this->mCoord);
}

void opCausalInt::init(const hypercube& mdShp){
  this->setShape(mdShp, mdShp);
}

void opCausalInt::oper(cubeData<float>& mod, cubeData<float>& dat, bool bAdj, bool bAdd) {
  causalInt(mod,dat,bAdj,bAdd);
}

void opMask::oper(cubeData<float>& mod, cubeData<float>& dat, bool bAdj, bool bAdd) {
  assert(mod.sameSize(DatShape()));
  addNull(mod,dat,bAdj,bAdd);
  long i, N = ModShape().get_n123();
  float *pin,*pout;
  if (!bAdj) {
    pin   = mod.dataPtr(); pout = dat.dataPtr();
  } else {
    pout  = mod.dataPtr(); pin = dat.dataPtr();
  }
  for (i=0;i<N;i++) {
    if ((*msk_)[i]>0.5f)  {pout[i] += pin[i];}
  }
}

//C=A*B 
opChain20::opChain20(const shared_ptr<op>& opA, const shared_ptr<op>& opB) :op() {
  this->setShape(opB->ModShape(),opA->DatShape());
  //allocate the temproary cube.
  tCube_.alloc(opB->DatShape());
  VERIFY(tCube_.domain().sameSize(opA->ModShape()));
  m_opA = opA; m_opB = opB;
}

//Chain two ops together. Chain operator C=AB
void opChain20::oper(cubeData<float>& mod, cubeData<float>& dat, bool bAdj, bool bAdd) {
  addNull(mod,dat,bAdj,bAdd);
  if (!bAdj) {
    m_opB->oper(mod,tCube_,false,false);
    m_opA->oper(tCube_,dat,false,bAdd);
  } else { //do adjoint
    m_opA->oper(tCube_,dat,true,false);
    m_opB->oper(mod,tCube_,true,bAdd);
    //mod.dump("dump2"); tCube_.dump("dump1");
  }
  return;
}

void opChain20::oper(bool bAdj, bool bAdd, int nmod, int ndat, float* mod, float* dat) {
  VERIFY(nmod == m_opB->ModShape().get_n123());
  VERIFY(ndat == m_opA->DatShape().get_n123());
  cubeDataF modc,datc;
  modc.becomeProxyCube(m_opB->ModShape()); datc.becomeProxyCube(m_opA->DatShape());
  modc.proxyPtrIs(mod);  datc.proxyPtrIs(dat);
  this->oper(modc,datc,bAdj,bAdd);
  return;
}

//Do forward and backward integration, this op can turn a (-1,0,2,0,-1) into a triangle
opBidirectInt::opBidirectInt(const hypercube& md_domain):op() {
  this->setShape(md_domain,md_domain);
  shared_ptr<opCausalInt> opIntA(new opCausalInt());
  shared_ptr<opCausalInt> opIntB(new opCausalInt());
  opIntA->init(md_domain); opIntB->init(md_domain);
  shared_ptr<opAdj> opIntA_adj(new opAdj(opIntB));
  this->chain_.reset(new opChain20(opIntA_adj,opIntA)); //chain two causal integration operators together
}

//the op is a self adjoint op indeed
void opBidirectInt::oper(cubeData<float>& mod, cubeData<float>& dat, bool bAdj, bool bAdd) {
  addNull(mod,dat,bAdj,bAdd);
  this->chain_->oper(mod,dat,bAdj,bAdd);
}

//The semblance calculation operator
void opMoveoutSemblance::init(const hypercube& modTrcShape, const hypercube& datPanelShape, tagMoveOutType moveType, float fAntiAlias) {
  this->setShape(modTrcShape,datPanelShape);
  mMoType = moveType;
  mfAntiAlias = fAntiAlias;
}

//see Jon's BEI book antialias chap
//fwd spread a doublet on each trace and causal integration, reconstruct data
//adj anticausal integrate the data panel, and then subtract in a doublet pattern. calculate semblance

void opMoveoutSemblance::oper(cubeData<float>& mod, cubeData<float>& dat, bool bAdj, bool bAdd) {
  addNull(mod,dat,bAdj,bAdd);
  
  //mod.info("mod"); dat.info("dat");
  
  dat.allocMemoryTo(tCube);
  axis ax_m_z = mod.domain().get_axis(1);
  axis ax_m_x = mod.domain().get_axis(2);
  axis ax_d_z = dat.domain().get_axis(1);
  axis ax_d_x = dat.domain().get_axis(2);

  float dz = ax_d_z.d, dx = ax_d_x.d;
  float z,z1,z2; float hfpdh, hfRectWidth,amp;
  float zshift,slope,p; //p is moveout param
  float x;
  //Use antialias and sum use rectangular box smoothing, 
  if (bAdj) { //do anticausal integ first
    causalInt(tCube,dat,true,false); //tCube.info("tCube");
    //do stacking along the trajectory (z,hx) panel
    for (int ip=0; ip<ax_m_x.n; ip++) {
      p = ax_m_x.get_coord(ip);
      for (int ix=0; ix<ax_d_x.n; ix++) {
        x = ax_d_x.get_coord(ix);
        for (int iz=0; iz<ax_m_z.n; iz++) {
          z = ax_m_z.get_coord(iz);
          this->getMoveoutValue(zshift, slope, p,x,z);
          hfpdh = slope*dx*0.5f;
          hfRectWidth = dz*0.5f + mfAntiAlias*fabs(hfpdh);
          z1 = z+zshift-hfRectWidth; z2=z+zshift+hfRectWidth;
          //The rectangular integeration should be placed between z1 and z2
          //Corresponding, the doulbets should be place at z1+0.5dz->z2+0.5dz
          amp = 0.0f;
          placeDoubletLin(amp,&tCube(ix,0),ax_d_z, z1+0.5*dz, z2+0.5*dz,true);
          
          /* debug, turns out to be numerical roundoff error
          if (amp <0.0f) {
            cerr<<"xd"<<x<<",z0"<<z<<",zsh"<<zshift<<",imu"<<ip<<",mu,"<<p<<",slope"<<slope<<endl;
            cerr<<z1<<sep<<z2<<sep<<hfRectWidth<<sep<<dz<<endl;
            PRV(amp);
            PRV(z1+0.5*dz); PRV(z2+0.5*dz);ax_d_z.print_info("ax_z");
            PRV(ix); tCube.dump("dump10");
            float off1 = ax_d_z.get_offsetf(z1+0.5*dz);
            float off2 = ax_d_z.get_offsetf(z2+0.5*dz);
            PRV(off1); PRV(off2);
            seperr("amp<0!!");
          }
          */
          mod(ip,iz) += amp*dz*0.5/hfRectWidth;
        }
      }
    }
  } else {
    //do spreading along the trajectory (z,hx) panel
    for (int ip=0; ip<ax_m_x.n; ip++) {
      p = ax_m_x.get_coord(ip);
      for (int ix=0; ix<ax_d_x.n; ix++) {
        x = ax_d_x.get_coord(ix);
        for (int iz=0; iz<ax_m_z.n; iz++) {
          z = ax_m_z.get_coord(iz);
          this->getMoveoutValue(zshift, slope, p,x,z);
          hfpdh = slope*dx*0.5f;
          hfRectWidth = dz*0.5f + mfAntiAlias*fabs(hfpdh);
          z1 = z+zshift-hfRectWidth; z2=z+zshift+hfRectWidth;
          //The rectangular integeration should be placed between z1 and z2
          //Corresponding, the doulbets should be place at z1+0.5dz->z2+0.5dz
          amp = mod(ip,iz)*dz*0.5f/hfRectWidth;
          placeDoubletLin(amp,&tCube(ix,0),ax_d_z, z1+0.5*dz, z2+0.5*dz,false);
        }
      }
    }
    //Add to dat
    causalInt(tCube,dat,false,true);
  }
}

//dat gather should be in (z,angle/offset,x) axis, the moveout is operated on the 2nd axis, mod is (zout,mu,x)
void Yang::calcMoveoutSemblance(cubeData<float>& mod, cubeData<float>&dat, 
tagMoveOutType moveType, float fAntialias, bool bAdj) {
  FUNC_TRACE_BEG(calcMoveoutSemblance);

  opMoveoutSemblance semblance;
  hypercube shpm = mod.domain(), shpd = dat.domain(), shpm_old,shpd_old;
  stdVecInt reshapeIndex(3); reshapeIndex[0]=1; reshapeIndex[1]=2;
  bool bNeedReshape = shpm.get_ndim()>3;
  if (bNeedReshape) {
    reshapeIndex[2] = shpm.get_ndim();
    shpm.reshape(reshapeIndex); shpd.reshape(reshapeIndex);
    //reshape the >2nd axes to one axis, (otherwise calcMoveoutSemblance won't work properly) before put into the semblance calculation
    mod.reshape(shpm,&shpm_old); dat.reshape(shpd,&shpd_old);
  }

  const hypercube& datShape = dat.domain();
  axis ax_x = datShape.get_axis(3);
  int ix;
  //Get a proxy
  cubeData<float> xSlice,xSliceMod;
  hypercube xSliceShape = datShape.return_compatible(1,2);
  hypercube xSliceModShape = mod.domain().return_compatible(1,2);
  xSlice.becomeProxyCube(xSliceShape);
  xSliceMod.becomeProxyCube(xSliceModShape);
  
  semblance.init(xSliceModShape,xSliceShape,moveType,fAntialias);
  windParam windix(3,windowAxis()); windix[2].n = 1;
  
  for (ix=0;  ix<ax_x.n; ix++) {
    //get a slice from inGather
    windix[2].f = ix;
    mod.setProxyXd(&xSliceMod,windix,2);
    dat.setProxyXd(&xSlice,windix,2);
    semblance.oper(xSliceMod,xSlice,bAdj,false);
  }
  
  if (bNeedReshape) {
    //resume the shape
    mod.reshape(shpm_old); dat.reshape(shpd_old);
  }
  FUNC_TRACE_END();
}

//dat gather should be in (z,angle/offset,x) axis, the moveout is operated on the 2nd axis, mod is (zout,mu,x)
//openmp version
void Yang::calcMoveoutSemblance_omp(cubeData<float>& mod, cubeData<float>&dat, tagMoveOutType moveType, float fAntialias, bool bAdj) {
  FUNC_TRACE_BEG(Yang::calcMoveoutSemblance_omp);

  
  hypercube shpm = mod.domain(), shpd = dat.domain(), shpm_old,shpd_old;
  stdVecInt reshapeIndex(3); reshapeIndex[0]=1; reshapeIndex[1]=2;
  bool bNeedReshape = shpm.get_ndim()>3;
  if (bNeedReshape) {
    reshapeIndex[2] = shpm.get_ndim();
    shpm.reshape(reshapeIndex); shpd.reshape(reshapeIndex);
    //reshape the >2nd axes to one axis, (otherwise calcMoveoutSemblance won't work properly) before put into the semblance calculation
    mod.reshape(shpm,&shpm_old); dat.reshape(shpd,&shpd_old);
  }

  const hypercube& datShape = dat.domain();
  axis ax_x = datShape.get_axis(3);
  int ix;
  
  int nth = g_NUM_THREADS;

  //PRV(bNeedReshape); ax_x.print_info("ax_x in calcMoveoutSemb");PRV(nth);

  opMoveoutSemblance semblance[nth];
  //Get a proxy
  cubeData<float> xSlice[nth],xSliceMod[nth];
  hypercube xSliceShape = datShape.return_compatible(1,2);
  hypercube xSliceModShape = mod.domain().return_compatible(1,2);
  windParam windix[nth];
  for (int i=0; i<nth; i++) {
    xSlice[i].becomeProxyCube(xSliceShape);
    xSliceMod[i].becomeProxyCube(xSliceModShape);
    semblance[i].init(xSliceModShape,xSliceShape,moveType,fAntialias);
    windix[i].resize(3); windix[i][2].n = 1;
  }
  
  int ith;
  #pragma omp parallel default(shared) private(ith,ix)
  {
  ith = omp_get_thread_num(); 
  VERIFY(ith<nth);
  for (ix=ith;  ix<ax_x.n; ix+=nth) {
    //get a slice from inGather
    windix[ith][2].f = ix;
    mod.setProxyXd(&xSliceMod[ith],windix[ith],2);
    dat.setProxyXd(&xSlice[ith],windix[ith],2);
    semblance[ith].oper(xSliceMod[ith],xSlice[ith],bAdj,false);
  }
  } //end omp parallel
  if (bNeedReshape) {
    //resume the shape
    mod.reshape(shpm_old); dat.reshape(shpd_old);
  }
  FUNC_TRACE_END();
}

//dense matrix multiply
void opMatrixMultiply::init(const cubeData<float>& matrix) {
  const hypercube& shape = matrix.domain();
  this->setShape(shape.return_compatible(2,2),shape.return_compatible(1,1));
  matrix.copyTo(m_matCoef);
}

//Matrix mult vector
//!The matrix is stored as a column first index. 
void opMatrixMultiply::oper(cubeData<float>& mod, cubeData<float>& dat, bool bAdj, bool bAdd) {
  FUNC_TRACE_BEG(opMatrixMultiply::oper);
  addNull(mod,dat,bAdj,bAdd);
  //Do multiplication, dummy implementation
  assert(modShape.sameSize(mod.domain()));
  assert(datShape.sameSize(dat.domain()));
  int nm = modShape.get_n(1);
  int nd = datShape.get_n(1);
  
  this->oper(bAdj, bAdd, nm,nd, mod.pval,dat.pval);
  FUNC_TRACE_END();
}

void opMatrixMultiply::oper(bool bAdj, bool bAdd, int nmod, int ndat, float* mod, float* dat) {
  addNull(bAdj,bAdd,nmod,ndat,mod,dat);
  
  float ftmp = 0;
  int i,j;
  if (!bAdj) {
    for (i=0; i<ndat; i++) {
      ftmp = 0.0f;
      for (j=0; j<nmod; j++) {
        //But keep in mind in C arrays, the fastest axis comes at the end.
        ftmp += m_matCoef(j,i)*mod[j];
      }
      dat[i] += ftmp;
    }
  } else {
    for (i=0; i<ndat; i++) {
      for (j=0; j<nmod; j++) {
        mod[j] += m_matCoef(j,i)*dat[i];
      }
    }
  }
}

void Yang::deg2radianAxis(axis& ax_out, const axis& ax_i, bool bFromDegToRadian) {
  CONST_CAST_REF(ax_in,ax_i,axis);
  if (!bFromDegToRadian) {
    ax_in.unit = "rad";
    ax_out.unit= "deg";
    ax_out.o = rad2deg(ax_in.o);
    ax_out.d = rad2deg(ax_in.d);
  } else {
    ax_out.unit = "rad";
    ax_in.unit= "deg";
    ax_out.o = deg2rad(ax_in.o);
    ax_out.d = deg2rad(ax_in.d);
  }
  ax_out.n = ax_in.n;
}

void Yang::degRadianConvert(cubeDataF& in, int iAxisAng, bool bFromDegToRadian) {
  vecAxis axes = in.domain().get_axes();
  axis ax;
  deg2radianAxis(ax, axes[iAxisAng-1], bFromDegToRadian);
  axes[iAxisAng-1] = ax;
  hypercube newShape(axes);
  in.reshape(newShape,NULL);
}

