#ifndef COMMON_OPS_H
#define COMMON_OPS_H

#include "decl_fwd.h"
#include "hypercube.h"
#include "gVec.h"

#include "op.h"
#include "imaging.h"

//Common Operators
using boost::shared_ptr;

namespace Yang {
  //This shadow operator act the same as the adjoint of the input operator, useful
  //in some occasions
  class opAdj:public op {
  public:
    virtual ~opAdj() {}
    opAdj(const shared_ptr<op>& realOp) {
      m_pOp = realOp;
      this->setShape(realOp->DatShape(),realOp->ModShape());
    }
    
    virtual void oper(cubeData<float>& mod, cubeData<float>& dat, bool bAdj, bool bAdd) {
      //switch the model and data
      m_pOp->oper(dat,mod,!bAdj,bAdd);
    }
  protected:
    shared_ptr<op> m_pOp;
  };
  
//fwd: sample a bunch of points from a regular grid
//Adj: spread a bunch of points to a regular grid
class opInterpLin: public op  //Define the class of operator
{
public:
  virtual ~opInterpLin() {}

  //define forward and adjoint
  //virtual void fwd(bool add, gVecBase* mod, gVecBase* dat) {}
  //virtual void adj(bool add, gVecBase* mod, gVecBase* dat) {}
  virtual void oper(cubeData<float>& mod, cubeData<float>& dat, bool bAdj, bool bAdd);
  
  void init(const cubeData<float>& coordCube, const hypercube& regularGridShape, bool bExtend);
protected:
  cubeData<float> mCoord;
  bool m_bExtend;
  
protected:
};

//causal integration
class opCausalInt: public op{
public:
  virtual ~opCausalInt() {}
  void init(const hypercube &modShape);
  virtual void oper(cubeData<float>& mod, cubeData<float>& dat, bool bAdj, bool bAdd);
};

//mask operator
class opMask: public op{
public:
  virtual ~opMask() {}
  opMask(const hypercube &modShape) {this->setShape(modShape,modShape);}
  void maskIs(const shared_ptr<cubeDataF>& msk) {msk_=msk;}
  virtual void oper(cubeData<float>& mod, cubeData<float>& dat, bool bAdj, bool bAdd);
  
private:
  shared_ptr<cubeDataF> msk_;
};

//chain 2 Operator, C=AB
class opChain20: public op{
public:
  virtual ~opChain20() {}
  opChain20(const shared_ptr<op>& opA, const shared_ptr<op>& opB);
  virtual void oper(cubeData<float>& mod, cubeData<float>& dat, bool bAdj, bool bAdd);
  virtual void oper(bool bAdj, bool bAdd, int nmod, int ndat, float* mod, float* dat);
  
protected:
  cubeDataF tCube_;  //tmp cube stores Bm, or transp(A)d, they are same size.
  shared_ptr<op> m_opA,m_opB;
};

//Do forward and backward integration
class opBidirectInt: public op {
public:
  opBidirectInt(const hypercube& domain);
  virtual void oper(cubeData<float>& mod, cubeData<float>& dat, bool bAdj, bool bAdd);
private:
  shared_ptr<opChain20> chain_;
};

class opOff2slope: public op  //Define the class of operator
{
public:
  virtual ~opOff2slope() {}

  //define forward and adjoint
  virtual void oper(cubeData<float>& mod, cubeData<float>& dat, bool bAdj, bool bAdd);
  
  void init(const hypercube &modShape, const hypercube & datShape, bool bUseAntialias, float fAntialias=0.0f);
protected:
  bool  m_bAntialias;
  float m_fAntialias;
};

  void deg2radianAxis(axis& ax_out, const axis& ax_in,bool bFromDegToRadian);
  void degRadianConvert(cubeDataF& in, int iAxisAng,  bool bFromDegToRadian);
///mod: is one panel contains the stacking values at each depth summing from the trajectory defined by 
///     ONE moveout parameter (z,\mu);
///dat: is the input 2D gather panels(z,x). Summing is performed along a traj on (z,x) at each z.
class opMoveoutSemblance: public op{
public:
  void init(const hypercube& modShape, const hypercube& datPanelShape, tagMoveOutType moveType, float fAntiAlias);
  virtual void oper(cubeData<float>& mod, cubeData<float>& dat, bool bAdj, bool bAdd);

  inline void getMoveoutValue(float& shift, float& slope, const float& mu,
                                           const float& x, const float& z0) {
    float ftmp;
    switch (mMoType) {
    case Linear: //as in slant stack
      shift =  mu*x;
      slope = mu;
      break;
    case Quadratic:
      shift = mu*x*x;
      slope = mu*x*2.0f;
      break;
    case Hyperbolic: // z = sqrt(z0^2+x^2*s^2);
      ftmp = sqrtf(z0*z0+SQ(x*mu));
      shift = ftmp - z0;
      slope = 2*x*SQ(mu)/ftmp;
      break;
    case tangentSquare: //assume in radius
      assert(fabs(x)<PI/2);
      ftmp = tanf(x);
      shift = mu*SQ(ftmp);
      slope = 2*mu*ftmp*(1.0f/SQ(cosf(x)));
      break;
    default:
      VERIFY(0);
    }
    return;
  }
  
protected:
  tagMoveOutType mMoType;
  float mfAntiAlias;
  cubeData<float> tCube;
};

void calcMoveoutSemblance(cubeData<float>& mod, cubeData<float>&dat, tagMoveOutType moveType, float fAntialias, bool bAdj);
void calcMoveoutSemblance_omp(cubeData<float>& mod, cubeData<float>&dat, tagMoveOutType moveType, float fAntialias, bool bAdj);

//simple matrix multiplication
  class opMatrixMultiply: public op  //Define the class of operator
  {
  public:
    virtual ~opMatrixMultiply() {}

    //define forward and adjoint
    virtual void oper(cubeData<float>& mod, cubeData<float>& dat, bool bAdj, bool bAdd);
    virtual void oper(bool bAdj, bool bAdd, int nmod, int ndat, float* mod, float* dat);
    
    void init(const cubeData<float>& matrix);

  protected:
    cubeData<float> m_matCoef;
  };
  
  
} //end namespace
#endif

