#ifndef __GVEC_H_INCL__
#define __GVEC_H_INCL__

#include <list>
#include "cubeData.hpp"

namespace Yang {
class gVecBase;

typedef std::list<gVecBase*> gVecCollection;
typedef std::list<gVecBase*>::const_iterator gVecIter;

/**
 * Any object that is to be stored in the composite tree 
 * structure must implement this interface.
 * 
 * @author yang
 */
class gVecBase {
protected:
  gVecBase(){}
   virtual ~gVecBase() {}
  public:
  virtual void MemberFunction() = 0;
  virtual int get_NumberProperty() = 0;
  virtual void set_NumberProperty(int value) = 0;
  virtual int get_ChildCount() = 0;
  virtual gVecIter Children() = 0;
  virtual int AddChild(gVecBase* child) = 0;
  virtual int RemoveChild(gVecBase* child) = 0;
  
  //Basic clone functions
  virtual void copyToVec(gVecBase* pVec) const = 0;
  virtual void allocMemoryToVec(gVecBase* pVec) const = 0;
  
  //calculate Z=X+a*Y or X=X+a*Y, can reuse code
  virtual void scaleAddVec(const gVecBase* pXVec, const gVecBase* pYVec, float a) = 0;
  virtual void scaleAddVec(const gVecBase* pYVec, float a) = 0;
  
  //calc Z=X+Y or X=X+Y
  virtual void addVec(const gVecBase* pXVec, const gVecBase* pYVec) = 0;
  virtual void addVec(const gVecBase* pYVec) = 0;
  virtual void subVec(const gVecBase* pXVec, const gVecBase* pYVec) = 0;
  virtual void subVec(const gVecBase* pYVec) = 0;
  
  //calc Z=a*X or X=a*x
  virtual void scaleVec(const gVecBase* pXVec, float a) = 0;
  virtual void scaleVec(float a) = 0;
};

/**
 * This class provides a standard implementation of the Component
 * class (leaf) as part of the Composite design pattern.  Objects 
 * of this type may not contain children.
 */
class gVecComponent : public gVecBase, public cubeData<float> {

protected:
private:
  int _numberPropertyValue;
  
public:
  gVecComponent() : gVecBase(), cubeData<float>(), _numberPropertyValue(0) {}
  //need a copy constructor
  //gVecComponent(const gVecComponent& other);

public:
  // Base class overrides
  void MemberFunction();
  int get_NumberProperty();
  void set_NumberProperty(int value);
  int get_ChildCount();
  gVecIter Children();
  int AddChild(gVecBase* child);
  int RemoveChild(gVecBase* child);
  
  public:
  const cubeData<float>* getDataCube () const {return this;}
   cubeData<float>* getDataCube () {return this;}
  //float RMS() const    { return m_pCubeData->RMS();}
  //float RMS_NZ( float scale) const {return m_pCubeData->RMS_NZ(scale);}
  //float maxval() const { return m_pCubeData->maxval();}
  //float minval() const { return m_pCubeData->minval();}
  //float maxvalAbs() const { return m_pCubeData->maxvalAbs();}
  
  bool sameSizeVec(const gVecBase* pVec) const {
    //cast
    const gVecComponent* p = dynamic_cast<const gVecComponent*>(pVec);
    assert(p);
    //invoke the sameSize function of cubeData<float>
    return (p->sameSize(*this));
  }
  
  //calculate Z=X+a*Y or X=X+a*Y, can reuse code
  virtual void scaleAddVec(const gVecBase* pXVec, const gVecBase* pYVec, float a);
  virtual void scaleAddVec(const gVecBase* pYVec, float a);
  
  //calc Z=X+Y or X=X+Y
  virtual void addVec(const gVecBase* pXVec, const gVecBase* pYVec);
  virtual void addVec(const gVecBase* pYVec);
  virtual void subVec(const gVecBase* pXVec, const gVecBase* pYVec);
  virtual void subVec(const gVecBase* pYVec);
  
  //calc Z=a*X or X=a*x
  virtual void scaleVec(const gVecBase* pXVec, float a);
  virtual void scaleVec(float a);
  
  virtual void copyToVec(gVecBase* pVec) const ; 
  
  //alloc the same memory space as this
  virtual void allocMemoryToVec(gVecBase* pVec) const;
};

/**
 * This class provides a standard implementation of the 
 * Composite class (node) as part of the Composite 
 * design pattern.  A Composite node may have one or
 * many children, but a Component (leaf) may not.
 */
class gVecComposite : public gVecBase {
public:
  gVecComposite() : gVecBase() {}
   virtual ~gVecComposite() {}
private:
  gVecCollection _children;
public:
  // Base class overrides
  void  MemberFunction();
  int   get_NumberProperty();
  void  set_NumberProperty(int value);
  int   get_ChildCount();
  gVecIter Children();
  int   AddChild(gVecBase* child);
  int   RemoveChild(gVecBase* child);
};

}

#endif // __GVEC_H_INCL__

