#ifndef HYPERCUBE_H
#define HYPERCUBE_H 1
#include "axis.h"
#include <vector>

#ifdef NDEBUG
	 #undef NDEBUG
	 #include <assert.h>
	 #define HYPERCUBE_H_FLAG1
#endif

namespace Yang{

//In hypercube, the axes index are 1 based.
class hypercube{

  public:
    hypercube(): n123(0){}
    hypercube(const hypercube& hpcube);
    hypercube(const vector<axis>& axes) { 
       init_nd(axes);
    }
    hypercube(const axis& ax) { 
       init_nd(vector<axis>(1,ax));
    }
    explicit hypercube(int nz, int nx) {
	   axis ax1(nz); axis ax2(nx);
	   vector<axis> axes;
	   axes.push_back(ax1);	axes.push_back(ax2);
	   init_nd(axes);
    }
	  hypercube(int nz) {
      axis ax1(nz); vector<axis> axes;
      axes.push_back(ax1); init_nd(axes);
    }
    explicit hypercube(int nz, int nx, int ny) {
      axis ax1(nz); axis ax2(nx); axis ax3(ny);
      vector<axis> axes;
      axes.push_back(ax1); axes.push_back(ax2); axes.push_back(ax3);
      init_nd(axes);
    }

  explicit hypercube(const stdVecInt& ns) {
    axis ax(1,0.,1.);
    vector<axis> axes;
    for (size_t i=0; i<ns.size(); i++) {
      ax.n = ns[i];
      axes.push_back(ax);
    }
    init_nd(axes);
  }

    //assignment operator
    hypercube& operator= (const hypercube & hpcube);
    
    void set_axes(const std::vector<axis>& axes);
    void set_axis(int idim, axis myaxis);
    const axis& get_axis(int idim) const;
    long long get_n123() const {return n123;}
  bool isEmpty() const {
    return n123==0;
  }
  long long get_n23() const{return n123/(this->get_n(1));}
  
  long long get_n23(int iStart, int iEnd) const {
	   assert(iEnd <= int(axes.size()));
	   long long num=1;
	   for (int i=iStart; i<=iEnd; i++) {
	     num = num * axes[i-1].n;
	   }
	   return num;
	 }
	 
  std::vector<int>   get_ns() const;
  std::vector<float> get_ds() const;
  
  void deallocate(){
    axes.clear();
    n123=0;
  }               
  ~hypercube(){
    this->deallocate();
  }

  int get_ndim() const {return axes.size();}
  int get_ndim_g1 () const;
  
  std::vector<axis> return_axes(int nmax) const;
  std::vector<axis> return_axes(int iAxisBeg, int iAxisEnd) const;
  
	 //return a subcube of the first nmax axis from this cube, all iAxis is one based
	 hypercube return_compatible(int nmax) const { return hypercube(return_axes(nmax));}
	 hypercube return_compatible(int iAxisStart, int iAxisEnd) const;
  hypercube return_arbAxesShape(const vector<int>& iAxes) const;

  const std::vector<axis>& get_axes() const {return axes;}
  int get_n(int idim)   const {assert(idim <= get_ndim() && idim>0); return axes[idim-1].n;}
  float get_d(int idim) const {assert(idim <= get_ndim() && idim>0); return axes[idim-1].d;}
  float get_o(int idim) const {assert(idim <= get_ndim() && idim>0); return axes[idim-1].o;}
	 //float coord <-> integer grid coordinate
	 //Get the coord given the dimension and the offset along the grid
	 //offset is 0-based as all interfaces in this class, only except idim
	 float get_coord( int offset, int idim) const {
	   assert(idim<=get_ndim()); return axes[idim-1].d*(offset) + axes[idim-1].o; }
	 int   get_offset(float coord, int idim) const {
	   assert(idim<=get_ndim()); return int((coord-axes[idim-1].o)/axes[idim-1].d + 0.5f);
	 }
	 
	  
  bool sameSize(const hypercube& other) const;
  bool compatible(const hypercube& other) const;
  bool sameShape(const hypercube& other) const ;
  void reshape(const stdVecInt& shapeInt); //same as Transp cmd, {1,3,5} means rearrange the shape into (1),(2,3),(4,5).
    
	void info( string s) const//print out the debug information
   { 
      int i;
	  prt("%s:",s.c_str());
      for( i=0; i < get_ndim(); i++){
         //print the info for each axis
         const axis& ax = axes.at(i);
         fprintf(stderr,"axis %d, n=%d, o=%f, d=%f, label=%s, unit=%s \n",i+1, 
				 ax.n,ax.o,ax.d,ax.label.c_str(),ax.unit.c_str()); 
		 }
	  fprintf(stderr,"total Number of points = %lld\n", get_n123());
	}
	
private:
    void init_nd(const std::vector<axis>& axes);
 protected:
   long long n123;
   std::vector<axis> axes;
   
};
} //end of namespace Yang

//undo the assert effect
#ifdef HYPERCUBE_H_FLAG1
	 #define NDEBUG
	 #include <assert.h>
	 #undef HYPERCUBE_H_FLAG1
#endif

#endif

