#include "stdafx.h"

#undef DEBUG
#include "myDebug.h"

#include "op.h"
#include "cubeAlgo.h"
#include "commonOps.h"
#include "imaging.h"
#include "omp.h"
#include "misc.h"

using namespace Yang;

void Yang::getSlopeAxisFromAngle(axis& ax_p, float amax, float amin, int n) {
  VERIFY(amin<amax and amax<90.0f && amax>0.0f && n>1);
  //calc the slope p=tan(a);
  //float da = (amax-amin)/(n-1);
  float pmin,pmax;
  
  pmin = tanf(float(amin*PI/180.0f));
  pmax = tanf(float(amax*PI/180.0f));
  assert(n>1);
  ax_p.o = pmin; ax_p.d = (pmax-pmin)/(n-1); ax_p.n = n;
  ax_p.label = string("p,tan(gamma)");
  return;
}

//Use slant stack to do off<->slope(ang) transform
//if !bAdj for off to slope, otherwise from
//slope to off
//input needs to be arranged as (z,hx,x) NOT(z,x,hx)
//output is arranged as (z,px,x)
//The axis info are in these two shapes. slope is tan(\gamma) in terms of angle domain
//No antialise, just Sample the linear-interpolated fields
//!Should add antialias support!!
void Yang::off2slope2dY(cubeData<float>& out, cubeData<float>& in, bool bAdj) {
  hypercube inShape = in.domain();
  hypercube outShape = out.domain();
  axis ax_z = inShape.get_axis(1); axis ax_x = inShape.get_axis(3);
  
  assert( ax_z.same_axis(outShape.get_axis(1)));
  assert( ax_x.same_axis(outShape.get_axis(3)));
  
  axis ax_h,ax_p;

  hypercube inSliceShape,outSliceShape;
  inSliceShape = inShape.return_compatible(1,2);
  outSliceShape = outShape.return_compatible(1,2);
  
  cubeData<float> inSliceCube,outSliceCube,upSampCube;
  inSliceCube.alloc(inSliceShape);
  outSliceCube.alloc(outSliceShape);
  
  in.resetCounter();
  out.resetCounter();
  out.zero();
  //What we do is first supersample the input offset image gather using sinc interpolation (along z direction)
  //Then do slant stack along different p, according to the axis_p, use linear interp in this case
  int upSampRate=4; float p;
  if (!bAdj) {
    ax_h = inShape.get_axis(2);
    ax_p = outShape.get_axis(2);
    int n1out = (ax_z.n-1)*upSampRate+1;
    float d1out = ax_z.d / upSampRate;
    float o1out = ax_z.o;
    int lsinc = 16;
    cubeData<float> coordC, zC1d, sampC1d; //z cube 1D
    axis ax_z_new(n1out,o1out,d1out);
    zC1d.alloc(hypercube(ax_z_new)); coordC.alloc(hypercube(ax_z.n)); 
    sampC1d.alloc(coordC.domain());
    for (int ix=0; ix<ax_x.n; ix++) {
      in.seq_grab(&inSliceCube);
      seplib::interp::getInstance()->interpolate(upSampCube,inSliceCube,&n1out,&o1out,&d1out,
                                                 NULL,NULL,NULL,NULL,NULL,NULL,
                                                 seplib::sinc,0,lsinc);
      //do slant stack on upSampCube
      float zSup; float h; float fshift;
      outSliceCube.zero();
      for (int ih=0; ih<ax_h.n; ih++) {
        h = ax_h.o + ih*ax_h.d;
        //Put data into zC1d (take a line along z axis from (z,hx) cube
        for (int j=0; j<ax_z_new.n; j++) {
          zC1d[j] = upSampCube(ih,j);
        }
        for (int ip=0; ip<ax_p.n; ip++) {
          p = ax_p.o + ip*ax_p.d;
          fshift = p*h;
          for (int iz=0; iz<ax_z.n; iz++) {
            zSup = ax_z.o+iz*ax_z.d;
            coordC[iz] = zSup + fshift;
          }
          //start sampling
          irregularSamp1d(sampC1d,zC1d,coordC,false,false);
          //sum to (px,z) cube
          for (int iz=0; iz<ax_z.n; iz++) {
            outSliceCube(ip,iz) += sampC1d(iz);
          }
        }
      }
      out.seq_update(&outSliceCube);
    }// for ix
  } else { //from slope to offset domain
    //take each point on the (z,p) panel, spread it along the slant lines.
    ax_h = outShape.get_axis(2);
    ax_p = inShape.get_axis(2);
    for (int ix=0; ix<ax_x.n; ix++) {
      in.seq_grab(&inSliceCube);
      //do slant spreading into (z,hx) panel
      float z; float h, fshift;
      cubeData<float> coordC, zC1d,spreadC1d; //z cube 1D
      coordC.alloc(hypercube(ax_z.n)); zC1d.alloc(coordC.domain());
      spreadC1d.alloc(hypercube(ax_z));
      outSliceCube.zero();
      for (int ih=0; ih<ax_h.n; ih++) {
        h = ax_h.o + ih*ax_h.d;
        for (int ip=0; ip<ax_p.n; ip++) {
          p = ax_p.o + ip*ax_p.d;
          fshift = p*h;
          for (int iz=0; iz<ax_z.n; iz++) {
            z = ax_z.o+iz*ax_z.d;
            coordC[iz] = z + fshift;
            zC1d[iz] = inSliceCube(ip,iz);
          }
          //start spreading on the 1d Z line
          irregularSamp1d(spreadC1d,zC1d,coordC,true,false);
          //sum to (hx,z) cube
          for (int iz=0; iz<ax_z.n; iz++) {
            outSliceCube(ih,iz) += spreadC1d(iz);
          }
        }
      }
      out.seq_update(&outSliceCube);
    }// for ix    
  }
}

//with rectangular antialias. mod is the offset gather, dat is the angle/slope gather assume (z,hx/px,x) axis arrangement.
void Yang::off2slope2dYaa(cubeData<float>& mod, cubeData<float>& dat, bool bAdj, float fAntialias, bool bAdd, bool bTrueAmpAdj, __angle_direction_unit::_angle_direction_unit angle_unit ) {
  
  VERIFY(fAntialias < 4.0f && fAntialias >=0.f ); //should be a reasonable number
  addNull(mod,dat,bAdj,bAdd);
  if (bTrueAmpAdj) {
    VERIFY(bAdj);
  }
  
  hypercube modShape = mod.domain();
  hypercube datShape = dat.domain();
  axis ax_z = modShape.get_axis(1); axis ax_x = modShape.get_axis(3);
  
  assert( ax_z.same_axis(datShape.get_axis(1)));  assert( ax_x.same_axis(datShape.get_axis(3)));
  
  axis ax_h,ax_p;
  hypercube datSliceShape,modSliceShape;
  datSliceShape = datShape.return_compatible(1,2);
  modSliceShape = modShape.return_compatible(1,2);
  
  //Now make an openMP version
  int nth = g_NUM_THREADS, ith; //prt("in off2slope2dYaa.\t"); PRV(nth); 
  
  cubeData<float> modSliceCubes[nth],datSliceCubes[nth],inSlicets[nth];
  windParam winds[nth];
  for (ith=0;ith<nth; ith++) {
    modSliceCubes[ith].becomeProxyCube(modSliceShape);
    datSliceCubes[ith].becomeProxyCube(datSliceShape);
    winds[ith].resize(3); winds[ith][2].n=1;
    inSlicets[ith].alloc(modSliceShape);
  }
  
  mod.resetCounter(); dat.resetCounter();
  
  //Then do slant stack along different p, according to the axis_p
  float dz = ax_z.d;
  ax_h = modShape.get_axis(2);  ax_p = datShape.get_axis(2);
  //build the slope arrays from ax_p(ax_p might be degree, or angle, or just slp).
  stdVecFloat slopeVec(ax_p.n);
  int ip;
  switch(angle_unit) {
  case __angle_direction_unit::deg:
    for (ip=0;ip<ax_p.n;ip++) {
      slopeVec[ip] = tanf(deg2rad(ax_p.get_coord(ip)));
    }
    break;
  case __angle_direction_unit::rad:
    for (ip=0;ip<ax_p.n;ip++) {
      slopeVec[ip] = tanf(ax_p.get_coord(ip));
    }
    break;
  case __angle_direction_unit::slp:
    for (ip=0;ip<ax_p.n;ip++) {
      slopeVec[ip] = ax_p.get_coord(ip);
    }
    break;
  }
  
#pragma omp parallel default(shared) private(ith,ip)
{
  float p,h,ph;
  ith = omp_get_thread_num(); VERIFY(ith<nth);
  cubeDataF& modSliceCube = modSliceCubes[ith];
  cubeDataF& datSliceCube = datSliceCubes[ith];
  cubeDataF& inSlicet = inSlicets[ith];
  windParam& wind = winds[ith];
  if (!bAdj) { //convert OFF to ANG
    //mod.dump("dump1"); dat.dump("dump2");
    //PRT_LOC; cerr << "ith=" << ith << "; &modSliceCube=" << &modSliceCube << endl;
    for (int ix=ith; ix<ax_x.n; ix+=nth) {
      wind[2].f = ix;
      //Get a 2D slice from the mod/dat
      mod.setProxyXd(&modSliceCube,wind,2);
      dat.setProxyXd(&datSliceCube,wind,2);
      //PRT_LOC; cerr << "ith=" << ith << "s1" << endl;
      inSlicet.zero();
      
      //PRT_LOC; cerr << "ith=" << ith << "s2" << endl;
      //Do antiCausal integration
      //if (ix==0) {inSlicet.dump("dump4"); modSliceCube.dump("dump5");}
      causalInt(inSlicet,modSliceCube,true,false);
      //if (ix==0) {inSlicet.dump("dump3"); modSliceCube.dump("dump6");}
      float z,z1,z2; float hfpdh, hfRectWidth,amp;
      //do slant stack into (z,hx) panel
      //PRT_LOC; cerr << "ith="<< ith << "s3" << endl;
      for (ip=0; ip<ax_p.n; ip++) {
        p = slopeVec[ip];
        hfpdh = p *ax_h.d * 0.5f;
        for (int ih=0; ih<ax_h.n; ih++) {
          h = ax_h.get_coord(ih);
          ph = p*h;
          for (int iz=0; iz<ax_z.n; iz++) {
            z = ax_z.get_coord(iz);
            hfRectWidth = std::max(dz*0.5f,fAntialias*fabsf(hfpdh));
            z1 = z+ph-hfRectWidth; z2=z+ph+hfRectWidth;
            
            //The rectangular integeration should be placed between z1 and z2
            //Corresponding, the doulbets should be place at z1+0.5dz->z2+0.5dz
            amp = 0.0f;
            placeDoubletLin(amp,&inSlicet(ih,0), ax_z, z1+0.5*dz, z2+0.5*dz,true);
            datSliceCube(ip,iz) += amp*dz*0.5/hfRectWidth;
          }
        }
      }
      
    }// for ix
  } else { //from angle to offset domain, adjoint method.
    //take each point on the (z,p) panel, spread it along the slant lines.
    for (int ix=ith; ix<ax_x.n; ix+=nth) {
      wind[2].f = ix;
      mod.setProxyXd(&modSliceCube,wind,2);
      dat.setProxyXd(&datSliceCube,wind,2);
      inSlicet.zero();
      float z,z1,z2; float hfpdh, hfRectWidth, amp;
      //do slant spreading into (z,hx) panel
      for (ip=0; ip<ax_p.n; ip++) {
        p = slopeVec[ip];
        hfpdh = p *ax_h.d * 0.5f;
        for (int ih=0; ih<ax_h.n; ih++) {
          h = ax_h.get_coord(ih);
          ph = p*h;
          for (int iz=0; iz<ax_z.n; iz++) {
            z = ax_z.get_coord(iz);
            hfRectWidth = std::max(dz*0.5f,fAntialias*fabsf(hfpdh));
            z1 = z+ph-hfRectWidth; z2=z+ph+hfRectWidth;
            if (!bTrueAmpAdj) { //exact adjoint
              amp = datSliceCube(ip,iz)*dz*0.5f/hfRectWidth;
            } else { //not exact adjoint, but tries to balance amplitude
              amp = datSliceCube(ip,iz);
            }
            //The rectangular integeration should be placed between z1 and z2
            //Corresponding, the doulbets should be place at z1+0.5dz->z2+0.5dz
            placeDoubletLin(amp,&inSlicet(ih,0), ax_z, z1+0.5*dz, z2+0.5*dz,false);
          }
        }
      }
      //Do causal integration
      causalInt(inSlicet,modSliceCube,false,true);
    }// for ix    
  }
} //end omp parallel
  TRACE(("s5"));
}

/**
//with rectangular antialias. mod is the offset gather, dat is the angle/slope gather assume (z,x,hx/px) axis arrangement, openMPed version
void Yang::off2slope2dYaa_par(cubeData<float>& mod, cubeData<float>& dat, bool bAdj, float fAntialias, bool bAdd, bool bTrueAmpAdj ) {
  addNull(mod,dat,bAdj,bAdd);
  if (bTrueAmpAdj) {
    VERIFY(bAdj);
  }
  
  hypercube modShape = mod.domain();
  hypercube datShape = dat.domain();
  axis ax_z = modShape.get_axis(1); axis ax_x = modShape.get_axis(3);
  
  assert( ax_z.same_axis(datShape.get_axis(1)));
  assert( ax_x.same_axis(datShape.get_axis(3)));
  
  axis ax_h,ax_p;
  hypercube datSliceShape,modSliceShape;
  datSliceShape = datShape.return_compatible(1,2);
  modSliceShape = modShape.return_compatible(1,2);
  
#ifdef _OPENMP
	cerr<<"off2slope2dYaa_par With openMP support"<<endl;
	int nth = omp_get_max_threads();
  cubeData<float> modSliceCubes[nth],datSliceCubes[nth];
  windParam winds[nth];
  for (int i=0; i<nth; i++) {
    modSliceCubes[i].becomeProxyCube();
    datSliceCubes[i].becomeProxyCube();
    winds[i].resize(3);
    winds[i][2].n = 1;
  }
#else
	cerr<<"off2slope2dYaa_par No OpenMP support"<<endl;
  cubeData<float> modSliceCube,datSliceCube;
  modSliceCube.becomeProxyCube(modSliceShape);
  datSliceCube.becomeProxyCube(datSliceShape);
  windParam wind(3,windowAxis());
  wind[2].n=1;
#endif

  mod.resetCounter(); dat.resetCounter();
  
  //Then do slant stack along different p, according to the axis_p
  float p,h,ph,dz;
  dz = ax_z.d;
  if (!bAdj) {
    ax_h = modShape.get_axis(2);
    ax_p = datShape.get_axis(2);
    cubeData<float> inSlicet;
    
    //mod.dump("dump1"); dat.dump("dump2");
    for (int ix=0; ix<ax_x.n; ix++) {
      wind[2].f = ix;
      mod.setProxyXd(&modSliceCube,wind,2);
      dat.setProxyXd(&datSliceCube,wind,2);
      modSliceCube.allocMemoryTo(inSlicet);
      //Do antiCausal integration
      //if (ix==0) {inSlicet.dump("dump4"); modSliceCube.dump("dump5");}
      causalInt(inSlicet,modSliceCube,true,false);
      //if (ix==0) {inSlicet.dump("dump3"); modSliceCube.dump("dump6");}
      float z,z1,z2; float hfpdh, hfRectWidth,amp;
      //do slant stack into (z,hx) panel
      for (int ip=0; ip<ax_p.n; ip++) {
        p = ax_p.o + ip*ax_p.d;
        hfpdh = p *ax_h.d * 0.5f;
        for (int ih=0; ih<ax_h.n; ih++) {
          h = ax_h.get_coord(ih);
          ph = p*h;
          for (int iz=0; iz<ax_z.n; iz++) {
            z = ax_z.get_coord(iz);
            hfRectWidth = dz*0.5f + fAntialias*fabs(hfpdh);
            z1 = z+ph-hfRectWidth; z2=z+ph+hfRectWidth;
            
            //The rectangular integeration should be placed between z1 and z2
            //Corresponding, the doulbets should be place at z1+0.5dz->z2+0.5dz
            amp = 0.0f;
            placeDoubletLin(amp,&inSlicet(ih,0), ax_z, z1+0.5*dz, z2+0.5*dz,true);
            datSliceCube(ip,iz) += amp*dz*0.5/hfRectWidth;
          }
        }
      }
    }// for ix
  } else { //from slope to offset domain
    //take each point on the (z,p) panel, spread it along the slant lines.
    ax_h = modShape.get_axis(2);
    ax_p = datShape.get_axis(2);
    cubeData<float> inSlicet;
    modSliceCube.allocMemoryTo(inSlicet);
    for (int ix=0; ix<ax_x.n; ix++) {
      wind[2].f = ix;
      mod.setProxyXd(&modSliceCube,wind,2);
      dat.setProxyXd(&datSliceCube,wind,2);
      inSlicet.zero();
      float z,z1,z2; float hfpdh, hfRectWidth, amp;
      //do slant spreading into (z,hx) panel
      for (int ip=0; ip<ax_p.n; ip++) {
        p = ax_p.get_coord(ip);
        hfpdh = p *ax_h.d * 0.5f;
        for (int ih=0; ih<ax_h.n; ih++) {
          h = ax_h.get_coord(ih);
          ph = p*h;
          for (int iz=0; iz<ax_z.n; iz++) {
            z = ax_z.get_coord(iz);
            hfRectWidth = dz*0.5f + fAntialias*fabs(hfpdh);
            z1 = z+ph-hfRectWidth; z2=z+ph+hfRectWidth;
            if (!bTrueAmpAdj) { //exact adjoint
              amp = datSliceCube(ip,iz)*dz*0.5f/hfRectWidth;
            } else { //not exact adjoint, but tries to balance amplitude
              amp = datSliceCube(ip,iz);
            }
            //The rectangular integeration should be placed between z1 and z2
            //Corresponding, the doulbets should be place at z1+0.5dz->z2+0.5dz
            placeDoubletLin(amp,&inSlicet(ih,0), ax_z, z1+0.5*dz, z2+0.5*dz,false);
          }
        }
      }
      //Do causal integration
      causalInt(inSlicet,modSliceCube,false,true);
    }// for ix    
  }
}

*/

void opOff2slope::oper(cubeData<float>& mod, cubeData<float>& dat, bool bAdj, bool bAdd) {
  VERIFY(!bAdd);
  if (!m_bAntialias) {
    if (!bAdj) {
      off2slope2dY(dat,mod,bAdj);
    } else {
      off2slope2dY(mod,dat,bAdj);
    }
  } else {
    off2slope2dYaa(mod,dat,bAdj,m_fAntialias,bAdd);
  }
}

void opOff2slope::init(const hypercube &modShape, const hypercube & datShape, bool bAntialias, float fAntialias) {
  this->setShape(modShape,datShape);
  this->m_bAntialias = bAntialias;
  this->m_fAntialias = fAntialias;
}

//last two are with default paramters, this functions determines 
//the dimensions of out cube inside.
void Yang::off2Ang2d(cubeData<float>& out, cubeData<float>& in, bool bBackward, enumOff2angType off2angType, float a_h_max,
                     int n, float fAntialias, float a_h_min) {
  //set the size of out, and apply different off2ang transform
  const hypercube& inShape = in.domain();
  hypercube outShape;
  axis ax_z = inShape.get_axis(1); axis ax_x = inShape.get_axis(3);
  
  axis ax_2 = inShape.get_axis(2);
  
  axis ax_h_p;
  if (bBackward) {
    ax_h_p.label = string("subOffsetX");
  } else {
    ax_h_p.label = string("angleX");
  }

  //process the default param
  if (n == DEF_VAL_OFF2ANG2D_N) {
    //set n to be the same size as input
    n = ax_2.n;
  } else {
    VERIFY( (off2angType == slantstackYang) || (off2angType == slantstackYangAA) ); //The other types does not have such flexibility
  }
  
  if (equalf(a_h_min,DEF_VAL_OFF2ANG2D_HMIN)) {
    a_h_min = -a_h_max;
  } else {
    VERIFY( off2angType == slantstackYang || (off2angType == slantstackYangAA) ); //The other types does not have such flexibility
  }
  
  
  if (!bBackward && ((off2angType == slantstackYang)||(off2angType==slantstackYangAA))) {
    //need slope axis
    getSlopeAxisFromAngle(ax_h_p,a_h_max,a_h_min,n);
  } else { //common case
    ax_h_p.n = n; ax_h_p.o = a_h_min; ax_h_p.d = (a_h_max-a_h_min)/(ax_h_p.n-1);
  }
  
  if (off2angType == slantstackYang || slantstackYangAA) { //need alloc the output myself 
    vecAxis axes; 
    axes.push_back(ax_z); axes.push_back(ax_h_p);
    axes.push_back(ax_x);
    outShape.set_axes(axes);
    out.alloc(outShape);
  }
  
  fAntialias = 1.0f;
  //PRV(fAntialias);
  switch (off2angType) {
  case Yang::seplibModified:
    seplib::off2ang(out,in,bBackward,a_h_max);
    break;
  case Yang::slantstackYangAA:
    if (!bBackward) {
      off2slope2dYaa(in,out,bBackward,fAntialias,false);
    } else {
      bool bTrueAmpAdj = true;
      off2slope2dYaa(out,in,bBackward,fAntialias,false,bTrueAmpAdj);
    }
    // LOGC(o2a2);  in.dump("dump1");  out.dump("dump2");    exit(0);
    break;
  case Yang::slantstackYang:
    off2slope2dY(out,in,bBackward);
    break;
  default:
    VERIFY(0);
  }
  return;
}

//only works with degrees, not radians or slopes!, and only use slant-stack.
void Yang::off2Ang2d_new(cubeData<float>& out, const cubeData<float>& in, bool bBackward, enumOff2angType off2angType, const axis& ax_h_or_gamma,
                     float fAntialias) {
  //set the size of out, and apply different off2ang transform
  const hypercube& inShape = in.domain();
  hypercube outShape;
  axis ax_z = inShape.get_axis(1); axis ax_x = inShape.get_axis(3);
  //axis ax_2 = inShape.get_axis(2);
  
  shared_ptr<axis> ax_c;
  if (bBackward) {
    ax_c.reset(new axis(ax_h_or_gamma)); //offset axis
  } else {
    ax_c.reset(new axis_deg(ax_h_or_gamma)); //need to degree
  }
  VERIFY((off2angType == slantstackYang) || (off2angType==slantstackYangAA));
  
  //need alloc the output myself 
  vecAxis axes; 
  axes.push_back(ax_z); axes.push_back(*ax_c);
  axes.push_back(ax_x);
  outShape.set_axes(axes);
  out.alloc(outShape);
  
  outShape.info("o1");
  
  fAntialias = 1.0f;
  cubeDataF& _in = const_cast<cubeDataF&>(in);
  bool bTrueAmpAdj;
  if (!bBackward) { //from off 2 angle.
    bTrueAmpAdj = false;
    off2slope2dYaa(_in,out,bBackward,fAntialias,false,bTrueAmpAdj,__angle_direction_unit::deg);
  } else { //from ang to off
    bTrueAmpAdj = false;
    off2slope2dYaa(out,_in,bBackward,fAntialias,false,bTrueAmpAdj,__angle_direction_unit::deg);
  }
  return;
}

//here model is offset gather and dat is angle gather.
void OpOff2Ang2d::oper(cubeData<float>& mod, cubeData<float>& dat, bool bAdj, bool bAdd) {
  addNull(mod,dat,bAdj,bAdd);
  //check if shape is good.
  axis ax_g = dat.domain().get_axis(2);
  axis ax_h = mod.domain().get_axis(2);
  
  float fAntialias = 1.0f;
  if (bAdj) {
    off2slope2dYaa(mod,dat,bAdj,fAntialias,bAdd,false,__angle_direction_unit::deg);
  } else {
    off2slope2dYaa(mod,dat,bAdj,fAntialias,bAdd,false,__angle_direction_unit::deg);
  }
}
