
/* Solver function for iterative least-squares optimization. Originally tinysolver.h in RSF package*/
/*
  Copyright (C) 2006 University of Texas at Austin
  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "stdafx.h"
#include "cubeData.hpp"
#include "dmRsfDump.h"
#include "op.h"
#include "cdstep.h"

#include "linSolver.h"

#undef DEBUG
#include "myDebug.h"
//#include "mkl_cblas.h"
#include "cblas_RSF.h"
using namespace Yang;
using namespace std;

//Solve Fm ~= d
int Yang::simpLinSolver ( op& Fop, /* linear operator */
		    cdstepper& stepper, /* stepping function */
		    int nm, int nd,                /* size of model and data */ 
		    float* m,              /* estimated model */
		    const float* m0,       /* starting model */
		    const float* d,        /* data */
		    int niter,            /* iterations */
        stdVecFloat& errVec          ) //record all error residuals, size of niter+1
/*< Generic linear solver. Solves oper{x} =~ dat >*/
{
  int i, iter;
  valarray<float> g(nm), rr(nd), gg(nd);

  for (i=0; i < nd; i++) {rr[i] = - d[i];}

  if (NULL==m0) {
    myMemset(m,0.f,nm);
  } else {
       for (i=0; i < nm; i++) {m[i] = m0[i];}
       Fop.oper(false, true, nm, nd, m, va_ptr(rr)); //(adj,add,nmod,ndat,mod,dat)
  }
  
  errVec.clear();
  //calc the first residuals cblas_dsdot( ny, gg, 1, gg, 1)
  float ferr = cblas_dsdot( nd, va_ptr(rr), 1, va_ptr(rr), 1);
  errVec.push_back(ferr);
  
  int ret_step;
  for (iter=0; iter < niter; iter++) {
    Fop.oper(true, false, nm, nd,va_ptr(g), va_ptr(rr)); //compute gradient dm
    Fop.oper(false,false, nm, nd,va_ptr(g), va_ptr(gg)); //compute residual incremental dr
    ret_step = stepper.step(true,   nm, nd, m, va_ptr(g), va_ptr(rr), va_ptr(gg));

    ferr = cblas_dsdot( nd, va_ptr(rr), 1, va_ptr(rr), 1);
    //prt("ferr=%4.12g \n",ferr);
    errVec.push_back(ferr);
    
    if (ret_step < 0) {
      prt("stepper quit at iter=%d \n", iter);
      break;
    }
  } //end for iter
  return 0;
}

//Add data weight and mask weight
//Compute D(FW(Jm'+(I-J)m0') -d) ~= 0, D is the data weight, and W is the model weight, J is the mask operator
//since we set m=m0 to start with, and the gradient only updates the Jm part. so Jm+(I-J)m0 will always be m in our computation.
//There is a implicit precondition,because WJm' = JWm' = Jm, and W(I-J)m0' = (I-J)m0.
int Yang::solver(op& Foper,   //linear operator
                  cdstepper& stepper, //stepping function
                  int nm,             //size of model
                  int nd,             // size of dat
                  float* x,           // estimated model
                  const float* dat,   // data 
                  int niter,          // number of iterations
                  //rests are optional arguments
                  float* wt, //data wt array, it might change during the inversion
                  FPcalcWeight wtGen, //recalculate data wt
                  const float* x0, //initial model
                  op*   nloper   , //nonline Operator !
                  const float* mwt, //model weight/mask array
                  int   verb, //verbosity flag
                  const float* mmask, //model mask
                  const int*  p_nmem, //iteration memory
                  const int*  nfreq , //periodic restart
                  dmRsf_ocDump<float>* xmov, //model iterations
                  dmRsf_ocDump<float>* rmov, //data iterations
                  stdVecFloat*  err, //final fitting error for each iterations
                  cubeDataF*    res) //final residual
//Generic linear solver.
//---
//Solves
//oper{x}    =~ dat
//---
//The last parameter in the call to this function should be "end".
//Example:
//---
//sf_solver (oper_lop,sf_cgstep,nx,ny,x,y,100);
{
  float dpr, dpg, dpr0, dpg0;
  int i, iter; 
  bool forget = true;
  float *wht = NULL;
  valarray<float> g(nm),rr(nd),gg(nd),td, wt2,xp(nm);//xp is the preconditioned x
  
  cubeDataF cub_m,cub_d;
  hypercube shpm(nm),shpd(nd);
  cub_m.becomeProxyCube(shpm); cub_d.becomeProxyCube(shpd);
  
  LOG("g1");
  if (wt != NULL || wtGen != NULL) {
    td.resize(nd);
    if (wt != NULL) {
      wht = wt;
    } else {
      wt2.resize(nd); wt2=1.0f;
      wht = va_ptr(wt2);
    }
  }
  LOG("g2");
  if (mwt!=NULL) {
    VERIFY(*(std::min_element(mwt,mwt+nm))>1e-4); //model weight(used as preconditioning) cannot be too small, because we need to divide it later.
  }
  //set initial data fitting error
  for (i=0; i < nd; i++) rr[i] = -dat[i];
  //arrayStat("rrafter assign",va_ptr(rr),nd);
  
  LOG("g3");
  //set initial model
  if (x0 != NULL) {
    for (i=0; i < nm; i++) {
      x[i]  = x0[i];        //this x0 is m0, not m0'.
      xp[i] = x[i];
    }
    if (mwt!=NULL) {
      xp[i] = x0[i]/mwt[i];
    }
    
    if (nloper != NULL) {
      nloper->oper(false, true, nm,nd, x, va_ptr(rr));
    } else {
      Foper.oper  (false, true, nm,nd, x, va_ptr(rr));
    }
  } else { //just set initial model to zero
    for (i=0; i < nm; i++) {
       x[i]   = 0.0f;
       xp[i]  = 0.0f;
    }
  } //end if set_initial model
  LOG("g4");
  
  //Now rr = FWm-d; Need to multiple data weight D, and x = W m0
  int nmem = (p_nmem==NULL ? -1:*p_nmem);
  if (NULL != err) err->resize(niter);
  
  //start the iteration
  for (iter=0; iter<niter; iter++) {
    if ( nmem >= 0) {  // restart
      forget = (iter >= nmem);
    }
    //regenerate the data weighting function
    if (wtGen != NULL && forget) {
       wtGen (nd, va_ptr(rr), wht);
    }
    
    //compute the gradient g.
    if (wht != NULL) {
       for (i=0; i < nd; i++) {
           rr[i] *= wht[i]; //rr = L(FW m0 - d) = r0
           td[i] = rr[i]*wht[i];
       }
       //arrayStat("rdat",va_ptr(rr),nd); arrayStat("dat",dat,nd);
       if (iter==0) {
         dpr0 = cblas_snrm2(nd, va_ptr(rr), 1);  dpg0 = 1.; //calculate the initial error ||r0||
       }
       Foper.oper(true, false, nm,nd, va_ptr(g), va_ptr(td)); // g = F'Lr, where grad=W'F'Lr
    } else {
       Foper.oper(true, false, nm,nd, va_ptr(g), va_ptr(rr));
    }
    
    //arrayStat("rdat",va_ptr(rr),nd); arrayStat("dat",dat,nd);
    //arrayStat("grad",va_ptr(g),nm); arrayStat("dmsk",wht,nd);
    VERIFY(!mwt);VERIFY(!mmask);
    //apply model weight and model mask
    if (mwt != NULL) {
      for (i=0; i < nm; i++) g[i] *= mwt[i]; // g = W'F'Lr == grad.
    }
    if (mmask != NULL) {
      for (i=0; i < nm; i++) {
        if (mmask[i]<0.5f) g[i] = 0.0f;     // apply mask to g, allow solver to updates only a selcted part of model space.
      }
    }
    
    //use model gradeint grad to generate data gradient "gg", gg = LFW grad
    if (mwt != NULL) {
       for (i=0; i < nm; i++) g[i] *= mwt[i]; 
    }
    Foper.oper(false,false, nm,nd, va_ptr(g), va_ptr(gg)); //gg = FW*grad
    
    if (wht != NULL) {
       for (i=0; i < nd; i++) gg[i] *= wht[i]; //gg = LFW*grad == ggrad
    }
    
    //flag forget=true every a few iterations.
    if (forget && nfreq != NULL) { /* periodic restart */
       forget = (0 == (iter)%(*nfreq)); 
    }
    
    if (iter == 0) {
      dpg0  = cblas_snrm2(nm, va_ptr(g), 1); PRV(dpg0);
      dpr   = 1.;
      dpg   = 1.;
    } else {
      dpr = cblas_snrm2 (nd, va_ptr(rr), 1)/dpr0;
      dpg = cblas_snrm2 (nm, va_ptr(g) , 1)/dpg0;
    }
    
    if (verb) {
      prt("Iter %4d res(d_wted) %f mod %f grad %f",
          iter+1, dpr, cblas_snrm2 (nm, x, 1), dpg);
    }
    
    //if converged, because of residual 'rr' or gradient 'g' becomes very small.
    if (dpr < epsilon || dpg < epsilon) {
      if (verb) {
        prt("convergence in %d iterations",iter+1);
      }
      if (mwt != NULL) {  //convert m' to m
        for (i=0; i < nm; i++) x[i] = xp[i]*mwt[i]; 
      }
      break;
    }
    
    PRV(forget);
    //simplified to the simply case g=dm, gg=F_*dm', rr= F_*m' - d_ ;
    int ierr;
    ierr = stepper.step(forget, nm, nd, va_ptr(xp), va_ptr(g), va_ptr(rr), va_ptr(gg));
    PRV(ierr); VERIFY(ierr>=0);
    
    forget = false;
    //now recalculate the data fitting residual for next iterations.
    //x = Wm' = m.
    if (mwt != NULL) {
      for (i=0; i < nm; i++)   x[i] = xp[i]*mwt[i];
    } else {
      for (i=0; i < nm; i++)   x[i] = xp[i];
    }
    
    //if operator is non-linear, we have to reapply the operator
    if (nloper != NULL) { 
      for (i=0; i < nd; i++) {
        rr[i] = -dat[i]; 
      }
      nloper->oper(false, true, nm, nd, x, va_ptr(rr));
    } else if (wht != NULL) { //if there is data weighting, need to reapply the opeator to compute the actual data fitting, rather than the weighted one.
      for (i=0; i < nd; i++) {
        rr[i] = -dat[i]; 
      }
      Foper.oper(false, true, nm, nd, x, va_ptr(rr)); //r = Fm-d, multiply L is done at the begining of next iteration
    } else if ((xmov != NULL || iter == niter-1)) {
    }
    
    if (xmov != NULL) {
      cub_m.proxyPtrIs(x); xmov->dumpFrame(cub_m);
    }
    if (rmov != NULL) {
      cub_d.proxyPtrIs(va_ptr(rr)); rmov->dumpFrame(cub_d);
    }
    if (err != NULL) err->operator[](iter) = cblas_snrm2(nd,va_ptr(rr),1); //This is outputing data matching error without taking account of the data weighting.
  } //end for iterations
  
  if (res != NULL) {
    for (i=0; i<nd; i++) res[i] = rr[i];
  }
  return 0;
}

