#ifndef MY_TYPES_H
#define MY_TYPES_H

#include <iostream>
#include <sstream>
#include <limits.h>
#include <float.h>

namespace fwk{
  
template <class UnitType, class RepType>
class Nominal{
public:
  //Constructor, explicit so that no type conversion is taken automatically
  explicit Nominal( RepType v): value_(v) {}

  int operator == (const Nominal<UnitType, RepType>& v) const {
    return value_ = v.value_;
  }
  int operator != (const Nominal<UnitType, RepType>& v) const {
    return value_ != v.value_;
  }
  //The Default copy constructor should be able to do correct things, thus omit it.
  
  //The assignment operator
  const Nominal<UnitType, RepType>& operator= (const Nominal<UnitType,RepType>& v) {
    value_ = v.value_;
    return *this;
  }
  
  //Automatic conversion to RepType, but not the otherway around.
  operator RepType() const {
    return value_;
  }
  
  RepType value() const{
    return value_;
  }
  
  RepType* addr() {
    return &value_;
  }
  
private:
  RepType value_;
};

template <class UnitType, class RepType, RepType const & _min, RepType const & _max>
//The value Range is constrained
class Nominal_valRangeConsF: public Nominal<UnitType, RepType> {
public:
  explicit Nominal_valRangeConsF(RepType pin = _min): Nominal<UnitType,RepType>(pin){
    //Add value Range Verification
    if (pin < _min || pin > _max) {
      std::ostringstream msgStrm;
      msgStrm<< "!illegal Nominal_valRangeCons value:" << RepType(*this) << endl;
      //throw msgStrm.str();
      cerr << msgStrm.str();
      VERIFY(0);
    }
  }

  //The assignment operator
  const Nominal_valRangeConsF<UnitType, RepType,_min,_max>& operator= (RepType v) {
    *this = Nominal_valRangeConsF<UnitType, RepType,_min,_max>(v); //this should call the base assignment operator
    return *this;
  }
};

template <class UnitType, class RepType, int _min, int _max>
//The value Range is constrained
class Nominal_valRangeConsI: public Nominal<UnitType, RepType> {
public:
  explicit Nominal_valRangeConsI(RepType pin = _min): Nominal<UnitType,RepType>(pin){
    //Add value Range Verification
    if (pin < RepType(_min) || pin > RepType(_max)) {
      std::ostringstream msgStrm;
      msgStrm<< "!illegal Nominal_valRangeCons value:" << RepType(*this) << endl;
      cerr << msgStrm.str();
      //throw msgStrm.str();
      VERIFY(0);
    }
  }
  
  //The assignment operator
  const Nominal_valRangeConsI<UnitType, RepType,_min,_max>& operator= (RepType v) {
    //call base assignment operator
    Nominal<UnitType, RepType>::operator =(Nominal_valRangeConsI<UnitType, RepType,_min,_max>(v));
    return *this;
  }
};

class _NumSubOffset {};
class _PosInt   {};
class _NnegInt  {};
class _Int{};

class OddInt: public Nominal<_Int,int> {
public:
    explicit OddInt(int pin = 1): Nominal<_Int,int>(pin){
    //Add value Range Verification
    if (pin%2!=1) {
      std::ostringstream msgStrm;
      msgStrm<< "!illegal OddInt value:" << this->value() << endl;
      cerr << msgStrm.str();
      //throw msgStrm.str();
      VERIFY(0);
    }
  }
  
  const OddInt& operator = (int v) {
    //call base assignment operator
    Nominal<_Int, int>::operator =(OddInt(v));
    return *this;
  }
};

class EvenInt: public Nominal<_Int,int> {
public:
    explicit EvenInt(int pin = 0): Nominal<_Int,int>(pin){
    //Add value Range Verification
    if (pin%2!=0) {
      std::ostringstream msgStrm;
      msgStrm<< "!illegal OddInt value:" << this->value() << endl;
      cerr << msgStrm.str();
      //throw msgStrm.str();
      VERIFY(0);
    }
  }
  
  const EvenInt& operator = (int v) {
    //call base assignment operator
    Nominal<_Int, int>::operator =(EvenInt(v));
    return *this;
  }
};

typedef Nominal_valRangeConsI<_NumSubOffset, int, 1, 1024> NumSubOffset;
typedef Nominal_valRangeConsI<_PosInt, int, 1, INT_MAX> PosInt;
typedef Nominal_valRangeConsI<_NnegInt,int, 0, INT_MAX> NnegInt;
typedef Nominal_valRangeConsI<_PosInt,int, 1, 2> NormPower12;

extern float const cgb_flt_max;
extern float const cgb_flt_posMin;
extern float const cgb_flt_zero;
extern float const cgb_flt_360;
extern float const cgb_flt_Neg360;
extern float const cgb_flt_2PI;
extern float const cgb_flt_Neg2PI;

class _Float{};
class _PosFloat{};
class _NnegFloat{};
typedef Nominal_valRangeConsF<_PosFloat,  float, cgb_flt_posMin, cgb_flt_max> PosFloat;
typedef Nominal_valRangeConsF<_NnegFloat, float, cgb_flt_zero,   cgb_flt_max> NnegFloat;

typedef Nominal_valRangeConsF<_Float,  float, cgb_flt_Neg360, cgb_flt_360> degree;
typedef Nominal_valRangeConsF<_Float,  float, cgb_flt_Neg2PI, cgb_flt_2PI> radian;

//find the name of the type in type parameters in template
template <typename T>
struct TypeName {
public:
  static const char* Get() {
    return typeid(T).name(); //however typeid gives are non-readable.
  }  
};
//add some specialziation
template <>
struct TypeName<degree> {
  static const char* Get() {return "degree";}
};
template <>
struct TypeName<float> {
  static const char* Get() {return "float";}
};
template <>
struct TypeName<radian> {
  static const char* Get() {return "radian";}
}; 

} //end namespace

using namespace fwk;

#include "axis.h"

//axis with unit
template <class UnitType>
class axis_w_unit: public axis {
public:
  //Constructor, explicit so that no type conversion is taken automatically
  explicit axis_w_unit(): axis() {}
  explicit axis_w_unit(const axis& _ax) {
    if (!_ax.unit.empty() || _ax.unit.compare("Undefined")==0) {
      this->set(_ax.n,UnitType(_ax.o),UnitType(_ax.d),_ax.label,_ax.unit);
    } else {
      this->set(_ax.n,UnitType(_ax.o),UnitType(_ax.d),_ax.label,string(TypeName<UnitType>::Get()));
    }
  }
  //The Default copy constructor should be able to do correct things, thus omit it.
  //The default assignment operator also should do the work.
};

class axis_deg: public axis_w_unit<degree> {
public:
  explicit axis_deg(const axis& _ax): axis_w_unit<degree> (_ax) {
    //check the value of axis_deg, should be resonable value.
    VERIFY(this->d > 0.099f); VERIFY(this->d < 30.1f); 
  }
};
class axis_rad: public axis_w_unit<radian> {
public:
  explicit axis_rad(const axis& _ax): axis_w_unit<radian> (_ax) {
    //check the value of axis_deg, should be resonable value.
    VERIFY(this->d > deg2rad(0.099f)); VERIFY(this->d < deg2rad(30.1f)); 
  }
};


#endif

