#ifndef __SOLVERNONLINEARSEARCH_H_INCL__
#define __SOLVERNONLINEARSEARCH_H_INCL__

#include "stdafx.h"
#include "dmRsfDump.h"
#include "cubeData.hpp"
#include "decl_fwd.h"
#include "gVec.h"
#include "objFunc.h"
#include "nonLinSolver.param.h"

using boost::shared_ptr;

namespace Yang {

//Free the ptr memories in a vector
template <typename T>
void deleteVecPointers(vector<T*>& ptrVec) {
  for (size_t i=0; i<ptrVec.size(); i++) {
    delete(ptrVec[i]);
  }
  ptrVec.clear();
}

template <typename T>
void deleteVecPointers(vector<T*>& ptrVec, vector<T*>& ptrVec2) {
  deleteVecPointers(ptrVec); 
  deleteVecPointers(ptrVec2);
  }

template <typename T>
void deleteVecPointers(vector<T*>& ptrVec, vector<T*>&ptrVec2, vector<T*>&ptrVec3, vector<T*>&ptrVec4) {
  deleteVecPointers(ptrVec,ptrVec2); 
  deleteVecPointers(ptrVec3,ptrVec4);
}

//Dump the itermediate models
inline void dumpVecModels(vector<gVecComponent*>& modelptrVec, const string& tag) {
  size_t vecSize = modelptrVec.size();
  if (vecSize>0) {
    dmRsf_ocDump<float> dumpModels(tag,tag);
    dumpModels.init(&(modelptrVec[0]->domain()));
    for (size_t i=0; i<vecSize; i++) {
      dumpModels.dumpFrame(*modelptrVec[i]);
    }
  }
}

inline void dumpVecModels(vector<gVecComponent*>& modelptrVec, dmRsf_ocDump<float>& ocDump) {
  size_t vecSize = modelptrVec.size();
  if (vecSize>0) {
    for (size_t i=0; i<vecSize; i++) {
      ocDump.dumpFrame(*modelptrVec[i]);
    }
  }
}

/** 
 * solver non-linear, perform a line search along the update 
 * direction 
 */

class smoother_intf;

struct solverParam {
  //For clipping the model
  float fmaxValAllow, fminValAllow;
  
  int maxIter;
  float errTol; //error tolerance for termination of iteration
  float fMaxPerturbScale, fcurMaxPerturbScale;
  bool bDecreaseMaxStepLengthWithIteration;
  //The formula is stepLenMax = (base1 + (1-base1)*(fdecay^(iter)))*fMaxPerturbScale
  float fbase1,fdecay;
  
  typedef enum tagSmoothType {
    sm_spline,
    sm_rect,
    sm_alongDip
  } smoothType;
  smoothType eSmooth;
  
  bool bSmoothSearchDir,bChangeSmoothParamWithIteration;
  int   nSpl[3], nSpl_end[3]; //number of points per axis at first iteration and last iteration.
  float fsSpl[3]; //scale nxspl by these numbers every iteration, should be a number bigger than 1.0.
  float curnspl_f[3];
  //From coarse grid to finer grid.
  //int rect[3]; 
  int repeat;
  
  bool      bMaskGradient; //if mask gradient
  cubeDataF gradMask;

  enum checkptResumeStat{rsm_grad,rsm_none};
  checkptResumeStat rsmStat_;

  bool bRecalcObjFuncEachIter;
  solverParam(): fmaxValAllow(0.0f), fminValAllow(0.0f),
    maxIter(0), errTol(0.0f), fMaxPerturbScale(0.01f), 
    bDecreaseMaxStepLengthWithIteration(false), fbase1(1.f),fdecay(0.9f),
    bSmoothSearchDir(true),bChangeSmoothParamWithIteration(true) {
    nSpl[0] =nSpl[1]  =nSpl[2]  =0; myMemset(nSpl_end,0,3);
    fsSpl[0]=fsSpl[1] =fsSpl[2] =0.0f; myMemset(curnspl_f,0.0f,3);
    eSmooth = sm_spline; bRecalcObjFuncEachIter=false;
  }
  
};

//input the current search direction, it will output the adjusted "conjugate" direction.
class nonLinSearchDirection{
public:
  typedef enum method__{FletcherR,PolakR,HestenesS,steepest} dirMethod;
  
  nonLinSearchDirection(dirMethod method, const hypercube& modShp): prevSearchDir_(new gVecComponent()),prevObjNegGradient_(new gVecComponent()), method_(method) {
    prevSearchDir_->alloc(modShp); prevObjNegGradient_->alloc(modShp);
  }
   
  //The input is the negative gradient of the minimizing objective function, the program will give the final search direction.
  int searchDirection(gVecComponent& dirOut, const gVecComponent& dir, bool forget=false);
protected:
  shared_ptr<gVecComponent> prevSearchDir_,prevObjNegGradient_;
  dirMethod method_; float beta_;
};

class solverNonlinearSearch {
private:
  shared_ptr<objFunc>            m_pObjFunc;
  shared_ptr<smoother_intf>      m_pSmoother;
  solverParam*        m_pSolverPar;
  shared_ptr<nonLinSearchDirection> directionPredictor_;
  float refStepSize_; //start as 1.0f, and will adjust within iteration based on the shape of the fitted parabola.
  
protected:
  float m_curObjFuncValue;//Prev Obj Function Value
  float m_RMS_initModel;  //Initial Model's Root-mean-square, used as a scaling reference
  
  shared_ptr<gVecComponent> m_apCurModel;
  shared_ptr<gVecComponent> m_apTrialModel; //A temp variable used in updateModel;
public:
  // Constructor
  solverNonlinearSearch(const shared_ptr<objFunc>& pObj, const gVecComponent* pInitModel, solverParam* pSolParam, const shared_ptr<nonLinSearchDirection>& searchDir);
  void  init(const shared_ptr<objFunc>& pObj, const gVecComponent* pInitModel, solverParam* pSolParam);

  // Destructor
  virtual ~solverNonlinearSearch();
  
  //update the model given a search direction
  bool updateModel( const gVecComponent& searchDir, float maxPerturbScale, float curStepSize, float* pNewStepSize=NULL);
  
  const gVecComponent& getCurModel() const {
    return *m_apCurModel;
  }
   
  //return the iteration-complete code
  int solve( gVecComponent& finalModel, stdVecFloat& finalErrorVec, int& finalIter,
             ocDumpF* pVecModels,ocDumpF* pVecGrads,ocDumpF* pVecDatas, 
             bool bDataInCore = true);

  virtual void smoothGradient(const gVecComponent& inputSearchDir, gVecComponent& searchDir);
protected:
  void setSmoothParam(int nSpl[3]);
  void storeDataVec(int ind);
  void loadDataVec(int ind);
    
  stdVecInt mCur_nspl; //Spline nodes amount in current iterations.

  
  cubeData<float> calDataCubes[3];
  string          calDataFns[3]; 
  bool bDataInCore;
    
private:
  
  // Copy constructor
  // TODO: Uncomment the copy constructor when you need it.
  solverNonlinearSearch(const solverNonlinearSearch& src);
  
  // Assignment operator
  // TODO: Uncomment the assignment operator when you need it.
  solverNonlinearSearch& operator=(const solverNonlinearSearch& src);
};

// TODO: Uncomment the copy constructor when you need it.
//inline solverNonlinearSearch::solverNonlinearSearch(const solverNonlinearSearch& src)
//{
//   // TODO: copy
//}

// TODO: Uncomment the assignment operator when you need it.
//inline solverNonlinearSearch& solverNonlinearSearch::operator=(const solverNonlinearSearch& rhs)
//{
//   if (this == &rhs) {
//      return *this;
//   }
//
//   // TODO: assignment
//
//   return *this;
//}

}

#endif // __SOLVERNONLINEARSEARCH_H_INCL__
