//#undef NDEBUG

#include "stdafx.h"

//enable assertion
#undef NDEBUG
#include <assert.h>

#include "cubeData.hpp"
#include "util_seplib.h"
#include "point2d.hpp"
#include "param_func.h"

using namespace seplib;
using namespace Yang;

void AGC_Param::readParam(const param_func* pars, const hypercube& shp0, const string& prefix) {
  const axis& ax_1 = shp0.get_axis(1);
  const axis& ax_2 = shp0.get_axis(2);
  std::stringstream ss;
  ss.str(""); ss << prefix << "nAGCwnd1";
  int n1def,n2def; n1def = ax_1.n/2; n2def = ax_2.n/10;
  this->nw1_    = pars->get_int(ss.str(), n1def);
  ss.str(""); ss << prefix << "nAGCwnd2";
  this->nw2_    = pars->get_int(ss.str(), n2def);
  ss.str(""); ss << prefix << "nAGCwnd3";
  this->nw3_    = pars->get_int(ss.str(), 1);
  VERIFY(this->nw3_==1);
  if (this->nw1_ == n1def && this->nw2_ == n2def) {
    PRT_LOC; prt("\n !Warning, AGC param for %s cannot be fetched from input!\n", prefix.c_str());
  }
  this->normPow_= PosInt(pars->get_int("normAGCpow",2));
}

//notice that accE is of size (nz+1)*(nx+1), the left and top is padded 1 layer of zeros for convenience.
void GenAccumulateImage(cubeDataF& accE, const cubeDataF& _in, const NormPower12& pow) {
  //first compute each col sum, like causal integration.
  const hypercube& shp = _in.domain();
  int n1 = shp.get_n(1), n2 = shp.get_n(2);
  hypercube shp2(n1+1,n2+1);
  accE.alloc(shp2); //init to zero.
  int i1,i2;
  bool bL2 = pow==2; float tmp;
  for (i2=1; i2<n2+1; i2++) {
    accE(i2,1) = _in(i2-1,0);
    for (i1=2; i1<n1+1; i1++) {
      if (bL2) { tmp = SQ(_in(i2-1,i1-1));}
      else {tmp = fabsf(_in(i2-1,i1-1));}
      accE(i2,i1) = accE(i2,i1-1) + tmp;
    }
  }
  //the first two cols is already done
  for (i2=2;i2<n2+1;i2++) {
    for (i1=1;i1<n1+1;i1++){
      accE(i2,i1) = accE(i2-1,i1) + accE(i2,i1);
    }
  }
  return;
}

//Automatic gain control, can do up to 2-D.
int seplib::AGC(const cubeDataF& _in, cubeDataF& _out, const AGC_Param* par, NnegInt divideTimes) {
  int ix,iz, nt, nx, nz;
  double lpnorm, lptemp, oldnorm;
  const float *x,*xp,*start,*end, *tail, *head;
  float *y, *yp;

  const hypercube& shp = _in.domain();
  _out.alloc(shp);
  axis ax_z = shp.get_axis(1);
  axis ax_x = shp.get_axis(2);
  
  nz = nt = ax_z.n;
  nx = shp.get_n23();
  
  VERIFY(divideTimes <= 2);
  bool bBoostByDivide = false;
  if (divideTimes == 2) { bBoostByDivide = true;}
  
  int window1 = par->nw1_; int window2 = par->nw2_;
  int normpow = par->normPow_;
  if(normpow !=1 && normpow !=2) {seperr("normpow must be 1 or 2.\n");}

  bool bOnly1D  = (window2==1);
  int hfWind1   = (window1+1)/2;
  
  if (bOnly1D) {
    // gain each trace by l1 norm of moving window 
    for (ix = 0; ix<nx; ix++) {
      x = &(_in  (ix,0)); xp = x;
      y = &(_out (ix,0)); yp = y;
      
      // untaper beginning of trace, to counter the boundary effect.
      start=xp; end=xp+window1;
      for (lpnorm=0.f; xp<end; ) {
        if( normpow == 2) {
          lpnorm += (*xp) * (*xp); xp++;
        } else { //norm = 1;
          if (*xp>0.) lpnorm += *xp++;
          else lpnorm -= *xp++;
        }
      }
      if (normpow == 2) lptemp = sqrt(lpnorm);  else lptemp = lpnorm;
      if (!bBoostByDivide) {
        for (xp=start, end=start+hfWind1; xp<end; ) *yp++ = *xp++/float(lptemp); //compensate the fir st window1/2 samples.
      } else {
        for (xp=start, end=start+hfWind1; xp<end; ) *yp++ = *xp++/SQ(float(lptemp)); //compensate more 
      }


      // core routine, xp now is at location (window1)/2.
      for(tail=start,head=start+window1,oldnorm=lpnorm; head<x+nt; head++,tail++) {
        //calculate the current norm_sum within the sliding window
        if(normpow == 2) { //L2 norm
            lpnorm += ((*head) * (*head)); 
            lpnorm -= ((*tail) * (*tail));
            if (lpnorm <= 0.0f) lpnorm = oldnorm;
            oldnorm = lpnorm;
            lptemp = sqrt(lpnorm);
        } else { //L1 norm
          if (*head>0.) lpnorm += *head;  else lpnorm -= *head;
          if (*tail>0.) lpnorm -= *tail;  else lpnorm += *tail;
          if (lpnorm <= 0.) lpnorm = oldnorm;
          oldnorm = lpnorm;
          lptemp = lpnorm;
        }
        if (!bBoostByDivide) {
          *yp++ = *xp++ / lptemp;
        } else {
          *yp++ = *xp++ / SQ(lptemp);
        }
      }
      // untaper end of trace, compensate edge effect
      if (!bBoostByDivide) {
        for (end = start+nz;xp<end;xp++,yp++) *yp = *xp/lptemp;
      } else {
        for (end = start+nz;xp<end;xp++,yp++) *yp = *xp/SQ(lptemp);
      }
    } //for each trace
  } else { //do 2-D as well
    //The  fast algorithm relies on quick norm summation of the windowed sub image.
    //build the temp image that contains the accumulated sum from origin (0,0).
    cubeDataF accE; GenAccumulateImage(accE,_in, NormPower12(normpow));
    double ratioN = double(window2*window1), ratioSqrtN = sqrt(ratioN);
    //accE.dump("dump1"); cubeDataF deNom(shp);
    float fmval = _in.maxvalAbs();
    bool bL2Norm = normpow==2; PRV(normpow); PRV(fmval);
    nx = ax_x.n; point2dI p1,p2,p3,p4;
    int hfWind2 = window2/2;
    for (ix=0;ix<nx;ix++) {
      for (iz=0;iz<nz;iz++) {
        //find the four corners of the windowing domain.
        p1.sz = iz-hfWind1; p1.sx = ix-hfWind2;
        p2    = p1 + point2dI(window1,0);
        p3    = p1 + point2dI(window1,window2);
        p4    = p1 + point2dI(0,window2);
        //determine the cases.
        bool bPt1zInRange = (p1.sz>=0); bool bPt1xInRange = (p1.sx>=0);
        bool bPt3zInRange = (p3.sz<=nz); bool bPt3xInRange = (p3.sx<=nx);
        if (bPt1zInRange && bPt1xInRange && bPt3zInRange && bPt3xInRange) { //the window is entirely within domain
          lptemp = accE(p3.sx,p3.sz) + accE(p1.sx,p1.sz) - (accE(p2.sx,p2.sz)+accE(p4.sx,p4.sz));
        } else if (!bPt1zInRange && bPt1xInRange && bPt3zInRange && bPt3xInRange)   { //top
          lptemp = accE(p3.sx,window1) - accE(p2.sx,window1);
        } else if (!bPt1zInRange && !bPt1xInRange && bPt3zInRange && bPt3xInRange)  { //corner 1 
          lptemp = accE(window2,window1);
        } else if (bPt1zInRange && !bPt1xInRange && bPt3zInRange && bPt3xInRange)   { //left
          lptemp = accE(window2,p3.sz) - accE(window2,p4.sz);
        } else if (bPt1zInRange && !bPt1xInRange && !bPt3zInRange && bPt3xInRange)  { //corner 2
          lptemp = accE(window2,nz) - accE(window2,nz-window1);
        } else if (bPt1zInRange && bPt1xInRange && !bPt3zInRange && bPt3xInRange)   { //bottom
          lptemp = (accE(p3.sx,nz) + accE(p1.sx,nz-window1))-(accE(p2.sx,nz)+accE(p4.sx,nz-window1));
        } else if (bPt1zInRange && bPt1xInRange && !bPt3zInRange && !bPt3xInRange)  { //corner 3
          lptemp = (accE(nx,nz)+accE(nx-window2,nz-window1))-(accE(nx-window2,nz)+accE(nx,nz-window1));
        } else if (bPt1zInRange && bPt1xInRange && bPt3zInRange && !bPt3xInRange)   { //right
          lptemp = (accE(nx,p3.sz)+accE(nx-window2,p1.sz))-(accE(nx-window2,p2.sz)+accE(nx,p4.sz));
        } else if (!bPt1zInRange && bPt1xInRange && bPt3zInRange && !bPt3xInRange)   { //corner 4
          lptemp = accE(nx,window1) - accE(nx-window2,window1);
        } else {
          VERIFY(0); //should not end up here.
        }

        if(lptemp<-fmval*1e-2f) {
          PRV2(lptemp,fmval); PRV2(ix,iz);
          PRV4(bPt1zInRange,bPt1xInRange,bPt3zInRange,bPt3xInRange);
        }
        assert(lptemp>-fmval*1e-2f); //sometime small negative values can happen because of numerical error
        if (bL2Norm) {lptemp=sqrt(fabs(lptemp));}
        if(lptemp < 1e-1*fmval) { //avoid divide by zero case.
          //PRT_LOC; prt("!!The dividing factor is too small!, f=%f; Consider using larger AGC window size\n", lptemp);
          lptemp = 1e-1*fmval;
        }
        if (bL2Norm) {
          _out(ix,iz) = _in(ix,iz)/float(lptemp)*ratioSqrtN; //divide the operator
        } else {
          _out(ix,iz) = _in(ix,iz)/float(lptemp)*ratioN; //divide the operator.
        }
        if (bBoostByDivide) {
          _out(ix,iz) = _out(ix,iz)/float(lptemp); //divide the operator again
        }
        //deNom(ix,iz) = lptemp;
      }
    } //end for ix.
    //deNom.dump("dump2");
  } //end if 2D case
  LOG("finish AGC");
  return 0;
}

int seplib::AGCdivide(const Yang::cubeDataF& _in, const Yang::cubeDataF& _inDenom, const Yang::cubeDataF* pMask, Yang::cubeDataF& _out, const AGC_Param* par, NnegInt divideTimes) {
  int ix,iz, nt, nx, nz;
  double lpnorm, lptemp, oldnorm;
  const float *x,*xp,*start,*end, *tail, *head;
  float *y, *yp;

  const hypercube& shp = _in.domain();
  _out.alloc(shp);
  axis ax_z = shp.get_axis(1);
  axis ax_x = shp.get_axis(2);

  nz = nt = ax_z.n;
  nx = shp.get_n23();

  VERIFY(divideTimes <= 2);
  bool bBoostByDivide = false;
  if (divideTimes == 2) { bBoostByDivide = true;}

  int window1 = par->nw1_; int window2 = par->nw2_;
  int normpow = par->normPow_;
  if(normpow !=1) {PRT_LOC; seperr("normpow must be 1 in seplib::AGCdivide \n");}
  
  bool bMask = pMask!=NULL;
  //The  fast algorithm relies on quick norm summation of the windowed sub image.
  //build the temp image that contains the accumulated sum from origin (0,0).
  cubeDataF accE; GenAccumulateImage(accE,_inDenom, NormPower12(normpow));
  //double ratioN = double(window2*window1), ratioSqrtN = sqrt(ratioN);
  //accE.dump("dump1"); cubeDataF deNom(shp);
  float fmval = _inDenom.maxvalAbs();
  nx = ax_x.n; point2dI p1,p2,p3,p4;
  int hfWind1 = window1/2; int hfWind2 = window2/2;
  for (ix=0;ix<nx;ix++) {
    for (iz=0;iz<nz;iz++) {
      if (bMask && (*pMask)(ix,iz)<0.5f) {
        _out(ix,iz) = 0.0f; continue;
      } //skip it

      //find the four corners of the windowing domain.
      p1.sz = iz-hfWind1; p1.sx = ix-hfWind2;
      p2    = p1 + point2dI(window1,0);
      p3    = p1 + point2dI(window1,window2);
      p4    = p1 + point2dI(0,window2);
      //determine the cases.
      bool bPt1zInRange = (p1.sz>=0);   bool bPt1xInRange = (p1.sx>=0);
      bool bPt3zInRange = (p3.sz<=nz);  bool bPt3xInRange = (p3.sx<=nx);
      if (bPt1zInRange && bPt1xInRange && bPt3zInRange && bPt3xInRange) { //the window is entirely within domain
        lptemp = accE(p3.sx,p3.sz) + accE(p1.sx,p1.sz) - (accE(p2.sx,p2.sz)+accE(p4.sx,p4.sz));
      } else if (!bPt1zInRange && bPt1xInRange && bPt3zInRange && bPt3xInRange)   { //top
        lptemp = accE(p3.sx,window1) - accE(p2.sx,window1);
      } else if (!bPt1zInRange && !bPt1xInRange && bPt3zInRange && bPt3xInRange)  { //corner 1 
        lptemp = accE(window2,window1);
      } else if (bPt1zInRange && !bPt1xInRange && bPt3zInRange && bPt3xInRange)   { //left
        lptemp = accE(window2,p3.sz) - accE(window2,p4.sz);
      } else if (bPt1zInRange && !bPt1xInRange && !bPt3zInRange && bPt3xInRange)  { //corner 2
        lptemp = accE(window2,nz) - accE(window2,nz-window1);
      } else if (bPt1zInRange && bPt1xInRange && !bPt3zInRange && bPt3xInRange)   { //bottom
        lptemp = (accE(p3.sx,nz) + accE(p1.sx,nz-window1))-(accE(p2.sx,nz)+accE(p4.sx,nz-window1));
      } else if (bPt1zInRange && bPt1xInRange && !bPt3zInRange && !bPt3xInRange)  { //corner 3
        lptemp = (accE(nx,nz)+accE(nx-window2,nz-window1))-(accE(nx-window2,nz)+accE(nx,nz-window1));
      } else if (bPt1zInRange && bPt1xInRange && bPt3zInRange && !bPt3xInRange)   { //right
        lptemp = (accE(nx,p3.sz)+accE(nx-window2,p1.sz))-(accE(nx-window2,p2.sz)+accE(nx,p4.sz));
      } else if (!bPt1zInRange && bPt1xInRange && bPt3zInRange && !bPt3xInRange)   { //corner 4
        lptemp = accE(nx,window1) - accE(nx-window2,window1);
      } else {
        VERIFY(0); //should not end up here.
      }
      assert(lptemp>-fmval*1e-2f); //sometime small negative values can happen because of numerical error
      if(lptemp < 1e-1*fmval) { //avoid divide by zero case.
        //PRT_LOC; prt("!!The dividing factor is too small!, f=%f; Consider using larger AGC window size\n", lptemp);
        lptemp = 1e-1*fmval;
      }
      _out(ix,iz) = _in(ix,iz)/(lptemp); //divide
      if (bBoostByDivide) {
        _out(ix,iz) = _out(ix,iz)/float(lptemp); //divide the operator again
      }
      //deNom(ix,iz) = lptemp;

  }} //end for ix,iz;
  //deNom.dump("dump2");
  LOG("finish agc_divide");
  return 0;
}

//Divide by Tiles (Tiles sum of the inDenom cube), using the window size given in AGC_Param. Currently the tiles do not overlap.
int seplib::divideByTiles(const cubeDataF& _in, const cubeDataF& _inDenom, cubeDataF& _out, const AGC_Param* par, NnegInt divideTimes) {
  int ix,iz, nt, nx, nz;
  double lpnorm, lptemp, oldnorm;
  const float *x,*xp,*start,*end, *tail, *head;
  float *y, *yp;

  const hypercube& shp = _in.domain();
  _out.alloc(shp);
  axis ax_z = shp.get_axis(1);
  axis ax_x = shp.get_axis(2);
  
  nz = nt = ax_z.n;
  nx = shp.get_n23();
  
  VERIFY(divideTimes <= 2);
  bool bBoostByDivide = false;
  if (divideTimes == 2) { bBoostByDivide = true;}
  
  int window1 = par->nw1_; int window2 = par->nw2_;
  int normpow = par->normPow_;
  VERIFY(normpow == 1);

  //bool bOnly1D  = (window2==1);
  int hfWind1   = (window1+1)/2;
  int hfWind2 = window2/2;
  VERIFY(window2 > 0);
  float fmval = _inDenom.maxvalAbs();

  //Start the 2-D tile calculation
  {
    //do 2-D as well
    //The  fast algorithm relies on quick norm summation of the windowed sub image.
    //build the temp image that contains the accumulated sum from origin (0,0).
    cubeDataF accE; GenAccumulateImage(accE,_inDenom, NormPower12(1));
    //accE.dump("dump1"); cubeDataF deNom(shp);
    nx = ax_x.n; point2dI p1,p2,p3,p4;
    
    int IX,IZ; int IZmax,IXmax;
    IZmax = (ax_z.n-1) / window1; IXmax = (ax_x.n-1) / window2;
    //Generate Tiles array
    hypercube shpTile(IZmax+1,IXmax+1);
    cubeDataF tileCb(shpTile);
    for (IX=0;IX<IXmax+1;IX++) {
      for (IZ=0;IZ<IZmax+1;IZ++) {
        //decide which tile am in.
        IZ = iz/window1; IX = ix/window2;
        //find the four corners of the tile
        p1.sz = IZ*window1; p1.sx = IX*window2;
        p2    = p1 + point2dI(window1,0);
        p3    = p1 + point2dI(window1,window2);
        p4    = p1 + point2dI(0,window2);
        //determine the cases.
        //bool bPt1zInRange = (IZ<IZmax); bool bPt1xInRange = (IX<IXmax);
        bool bPt3zInRange = (p3.sz<=nz); bool bPt3xInRange = (p3.sx<=nx);
        if (bPt3zInRange && bPt3xInRange) { //the window is entirely within domain
          lptemp = accE(p3.sx,p3.sz) + accE(p1.sx,p1.sz) - (accE(p2.sx,p2.sz)+accE(p4.sx,p4.sz));
        } else if (!bPt3zInRange && bPt3xInRange)   { //bottom
          lptemp = (accE(p3.sx,nz) + accE(p1.sx,nz-window1))-(accE(p2.sx,nz)+accE(p4.sx,nz-window1));
        } else if (!bPt3zInRange && !bPt3xInRange)  { //corner 3
          lptemp = (accE(nx,nz)+accE(nx-window2,nz-window1))-(accE(nx-window2,nz)+accE(nx,nz-window1));
        } else if (bPt3zInRange && !bPt3xInRange)   { //right
          lptemp = (accE(nx,p3.sz)+accE(nx-window2,p1.sz))-(accE(nx-window2,p2.sz)+accE(nx,p4.sz));
        } else {
          VERIFY(0); //should not end up here.
        }
        assert(lptemp>-fmval*1e-2f); //sometime small negative values can happen because of numerical error
        if(lptemp < 1e-2*fmval) { //avoid divide by zero case.
          //PRT_LOC; prt("!!The dividing factor is too small!, f=%f; Consider using larger AGC window size\n", lptemp);
          lptemp = 1e-2*fmval;
          VERIFY(0);
        }
        tileCb(IX,IZ) = lptemp;
      }
    }
    
    for (ix=0;ix<nx;ix++) {
      for (iz=0;iz<nz;iz++) {
        //decide which tile am in.
        IZ = iz/window1; IX = ix/window2;
        _out(ix,iz) = _in(ix,iz)/tileCb(IX,IZ); //divide the operator
      }
    } //end for ix.
  } //end if 2D case
  LOG("finish DivideByTiles");
  return 0;
}

