#ifndef WAVEOPPARAMS_H
#define WAVEOPPARAMS_H 1

#include "decl_fwd.h"
#include "hypercube.h"
#include "cubeData.hpp"

using boost::shared_ptr;
namespace Yang {

class wavFieldGroup{
public:
   cubeData<ETYPE>      p_cb[3]; //acoustic wavefield pressure, cube data
   cubeData<ETYPE>    * pp_cb[3];//pointers for array p
private:
   wavFieldGroup( const wavFieldGroup& wfg);
   
public: //method
   wavFieldGroup() {
	  pp_cb[0] = NULL;
	  pp_cb[1] = NULL;
	  pp_cb[2] = NULL;
   }
   void switchWavfieldPtrs();
   void getPtrs(cubeData<ETYPE> **ppp0, cubeData<ETYPE> **ppp1, cubeData<ETYPE> **ppp2);
   void reset();
   void alloc(const hypercube& shape);
   void dealloc();
};

inline void wavFieldGroup::switchWavfieldPtrs()
{
   cubeData<ETYPE> *p_tmp;
   p_tmp = pp_cb[0];
   pp_cb[0] = pp_cb[1];
   pp_cb[1] = pp_cb[2];
   pp_cb[2] = p_tmp;
}

inline void wavFieldGroup::reset()
{
   pp_cb[0]->zero();
   pp_cb[1]->zero();
   pp_cb[2]->zero();
}

inline void wavFieldGroup::alloc(const hypercube& shape)
{
   p_cb[0].alloc(shape);
   p_cb[1].alloc(shape);
   p_cb[2].alloc(shape);   
   pp_cb[0] = &(p_cb[0]);
   pp_cb[1] = &(p_cb[1]);
   pp_cb[2] = &(p_cb[2]);
}

inline void wavFieldGroup::dealloc()
{
   p_cb[0].dealloc();
   p_cb[1].dealloc();
   p_cb[2].dealloc();
   pp_cb[0] = NULL;
   pp_cb[1] = NULL;
   pp_cb[2] = NULL;   
}

//cannotmake virtual because the argument type is different from 3D case
inline void wavFieldGroup::getPtrs(cubeData<ETYPE> **ppp0, cubeData<ETYPE> **ppp1, cubeData<ETYPE> **ppp2)
{
   *ppp0 = pp_cb[0];
   *ppp1 = pp_cb[1];
   *ppp2 = pp_cb[2];
}

//Params of doing prestack imaging
//Measured by grid points in wavfield grid pts.
struct offsetImagingParams {
   bool bDoOffsetGather;
   int ohx,dhx;
   int ohy,dhy;
   int ohz,dhz;
   NumSubOffset nhx,nhy,nhz;
	int ot,dt,nt;
   unsigned int AxesUsage; //Use the least siginifcant bits to indicate the which axis/axes(hx,hy,hz,t) is used
   offsetImagingParams(): bDoOffsetGather(false),ohx(0),dhx(0), ohy(0),dhy(0),
	  ohz(0),dhz(0), AxesUsage(0) {}
};

//The axes are (x,y,z,hx [,hy] [,hz], [t])
//Only deal with sptial axes currently
//If input wfld, then output wfld shape slices //For Image output
//If input vel grid, then ouptut vel grid shape slices.//For image input
inline void get_prestack_img_cube_shape(hypercube& newShape, const offsetImagingParams& off_img_param, const hypercube& wfldShape, float dtComp = 0.0f) {
   assert(off_img_param.bDoOffsetGather);
   vecAxis axess = wfldShape.get_axes();
   unsigned int axesUsage = off_img_param.AxesUsage;
   axis ax, ax_s; float d;
   if (axesUsage & 0x1) {//Use hx axis, remember that x axis is axis 2(TWO) in wfldShape
	  ax.label = string("hx");
	  ax_s = wfldShape.get_axis(2); d=ax_s.d;
	  ax.o = off_img_param.ohx * d;
	  ax.d = ax_s.d * off_img_param.dhx;
	  ax.n = off_img_param.nhx;
	  axess.push_back(ax);
   }
   if (axesUsage & 0x2) {//Use hy axis
	  ax.label = string("hy");
	  ax_s = wfldShape.get_axis(3); d=ax_s.d;
	  ax.o = off_img_param.ohy * d;
	  ax.d = ax_s.d * off_img_param.dhy;
	  ax.n = off_img_param.nhy;
	  axess.push_back(ax);
   }
   if (axesUsage & 0x4) {//Use hz axis
	  ax.label = string("hz");
	  ax_s = wfldShape.get_axis(1); d=ax_s.d;
	  ax.o = off_img_param.ohz * d;
	  ax.d = ax_s.d * off_img_param.dhz;
	  ax.n = off_img_param.nhz;
	  axess.push_back(ax);
   }      
	newShape.set_axes(axess);
	return;
}

enum impose_receiver_record_type
{
   AsForce,
   AsBoundary
};
enum boundaryType 
{
   abs_null, //do nothing
   abs_ord1, //absorb 1st order
   abs_ord2, //absorb 2nd order
   abs_Dave,  //Improved 1st order absorption method by Dave Hale, used in SU model
   abs_ord1par //absorb 1st order, but only first order accuracy in FD scheme, but easy to parallel.
};

class vel;
class srcWav;

//Wave propagate params
struct propagateParams {
public:  
  //3D field for wavefield snap could be an out of core object or incore.
  dmRsf<float> *pSrcSnapshot;
  dmRsf<float> *pRecvSnapshot;
  
  dmRsf<float> *pSrcSnapshot2;
  dmRsf<float> *pRecvSnapshot2;
  
  //FOR SIMPLICITY, ALL QUANTITIES ARE OF WAVEFIELD SIZE
  //All For zero offset
  dmRsf<float> *pImages; //stores RTM_image every shot, for zero offset
  cubeData<float> *pFinalImage; //Image summing up all shots, also as the input refl model for adjRTM
  
  cubeData<float> *pFinalDeltaS;  //This is always incore, input for born Modeling, for zero-offset
  cubeData<float>*  pDeltaSs; //offset Image gathers summing all shots.  
  
  //the seismic record, input for Imaging, output of modeling
  dmRsf<float> *pRsfSeisRec; 	 
  
  //For snap/Dump params
  int   j_shot_snap, f_shot_snap, n_shot_snap;
  bool  bSaveWfldSnap;
  
  impose_receiver_record_type mImposeRecvType;
  
  //velocity model grid shape, global vel Model
  hypercube velModelShape;  //The entire vel model
  hypercube finalImageShape;  //The entire image
  hypercube finalOffImageShape;  //The entire offset image (if there is).
  
  //For each shot
  hypercube mShapeWavfield; //The dimension of wavefield
  hypercube mImageShape;    //The size of image per shot, HERE I remove the boundary and the padding zone
  
public:
  propagateParams():	bSaveWfldSnap(false) {
    //Put default values here.
    pSrcSnapshot 	= NULL;
    pRecvSnapshot	= NULL;
    pSrcSnapshot2 	= NULL;
    pRecvSnapshot2	= NULL;
    pImages		= NULL;
    pFinalImage = NULL;
    pDeltaSs = pFinalDeltaS = NULL;
  }
  virtual ~propagateParams() {}
private:
  //Make copy constructors forbidden
  propagateParams(const propagateParams& other);
};

struct FDModelingTimeParams {
  PosFloat  totalTime,dtComp,dtRecord;
  PosFloat  dtSnap,dtImaging;
  PosInt  Nt,nSnap,nImaging,nRec;
  PosInt  jSnap,jRec,jImaging;
  //default copy constructor is OK, no pointers
};

struct OnewayOpPar{
 string csouFileName,crecFileName;
 string sz_additional_mig_par;
 //int ntpx; //taper points
 //int nws;  //number of requencies.
 //string pdip,oper;
 //int nxPadOneway,nyPadOneway;
}; //specifying the parameter about the oneway operator's implementation.

class ImgPar {
public:
  bool  bGenOffsetGather() const { return offImgParams_.bDoOffsetGather;}

  const offsetImagingParams& offImgPars() const {return offImgParams_;}
  void  offImgParsIs( const offsetImagingParams& offImageParam);

  const hypercube& finalImageShp() const {return finalImageShape_;}
  void  finalImageShpIs(const hypercube& finalImgShp);
  
  void  readParam(param_func* pars, const AcqPar& acqPar);
  
  const hypercube& finalSubOffImageShp() const {return finalSubOffsetImageShape_;}
private:
  hypercube finalImageShape_,finalSubOffsetImageShape_;
  //For prestack image with subsurface offset, for simplicity, these quantities are all of wavefield size grids.
  offsetImagingParams offImgParams_;
  //dmRsf<float>*		pOffImages; 	//offset Image gathers for each shot
  //cubeData<float>*  pFinalOffImage;//summing all shots
  bool bIllumBalancing_;
};

class AcqPar {
public:
  const hypercube& vmodShape() const {return vmodShape_;}
  int readParam(param_func* pars, const hypercube& vmodShape);

  enum SRType {tagAlong, splitSpread}; //source receiver type
  SRType  srType() const {return srType_;}
  enum srcType {STpoint,STplane,STrandom}; //different src type, compoisite sources except the first one.
  srcType sourceType() const {return srcType_;}
  
  bool b3D() const {return b3D_;}
  void getSrcRecvInfoX(PosInt& nShotx, NnegInt& shot_ox, PosInt& shot_dx, PosInt& nRecvx,  int& recv_ox, PosInt& recv_dx) const {
    nShotx = m_nShots;  shot_ox = mShot1_ix; shot_dx = mShotIncX;
    nRecvx = m_nTraces; recv_ox = mTrace1ix; recv_dx = mTraceIncX;
  }
  void getSrcRecvInfoY(PosInt& nShoty, NnegInt& shot_oy, PosInt& shot_dy, PosInt& nRecvy,  int& recv_oy, PosInt& recv_dy) const {
    VERIFY(b3D_);
    nShoty = m_nShotsY;  shot_oy = mShot1_iy; shot_dy = mShotIncY;
    nRecvy = m_nTracesY; recv_oy = mTrace1iy; recv_dy = mTraceIncY;
  }
  const PosInt& nShots() {return m_nShots;}
  NnegInt shotDepth() const {return mShot1_iz;}
  NnegInt recvDepth() const {return NnegInt(mTrace1iz);}
  
  bool bSrcRecvLocCovered(const float* coordS, const float* coordR);
  bool bSrcRecvLocCovered(const stdVecFloat& SrcRay, const stdVecFloat& RecvRay);
  
private:
  hypercube vmodShape_;
  bool    b3D_;
  PosInt  m_nShots  ,m_nTraces; //number of sources and receivers along the inline direction
  PosInt  m_nShotsY ,m_nTracesY; //number of sources and receivers cross-line direction
  NnegInt mShot1_ix,mShot1_iz,mShot1_iy;
  int mShotIncX,mShotIncY;
  int mTrace1ix,mTrace1iz,mTrace1iy;
  int mTraceIncX,mTraceIncY;

  SRType srType_; //tag along means marine acquisition, second means land acquisition.
  srcType srcType_;
};

//For compatibility, this structure contain both oneway and two-way parameters.
class waveOpParams: public propagateParams{
public:
  waveOpParams(): propagateParams(), bUse4thOrder(false), bTake2ndTimeDerivForRTM(false),offset_imaging_gpu_kernel_type(0),
    bUseOnewayPropagator_(false),pImageMask(NULL) {}
  virtual ~waveOpParams() {}
  int   readParam(param_func* pars);
  int   read_twoway_prop_param(param_func* pars);
  int   read_oneway_prop_param(param_func* pars);
  bool  bOneway() const {return bUseOnewayPropagator_;}

  int calcTimeParamsFor2wayModeling(float maxFreq, float vmax, float vmin, const hypercube& propGridShape ); //propagation model grid.
  
  const OnewayOpPar* onewayPar() const {return &onewayPar_;}
  
  //I use it to simulate sparse sampling of the image gather, temporary solution here
  cubeData<float>* pImageMask; 
  string fn_imageMask;

  //For MPI
  string suf_MPI_; //file suffix for MPI purpose.
private:
  OnewayOpPar onewayPar_;
  FDModelingTimeParams timeParam_;
  bool bUseOnewayPropagator_;
  int stencilWidth;
  bool bFreeSurfaceOnTop;
  bool bUse4thOrder;

  int offx,offz,offy; //off between vel model cube and wavefield cube
  //padding layer related
  int m_nPad;
  //boundary layer
  boundaryType m_bdType;
	
  bool bTake2ndTimeDerivForRTM;
  //For GPU
  int offset_imaging_gpu_kernel_type;
};

//There will be one-way Wave Equation based Modeling Params.
inline void setFDwavfieldShape(hypercube& wavfieldShape, int stenWid, const hypercube& velGridShape) {
   int ndim = velGridShape.get_ndim();
   vector<axis> axes;
   axis ax;
   for (int i=1;i<=ndim;i++) //get_axis is 1 based.
   {
      ax = velGridShape.get_axis(i);
      ax.n = ax.n+2*stenWid;
      ax.o = ax.o - stenWid*ax.d;
      axes.push_back(ax);
   }
   wavfieldShape.set_axes(axes);
}

inline void checkSnapRecordParams(propagateParams* par, AcqPar* acqPar) {
  int o,j;
  o=par->f_shot_snap;   j=par->j_shot_snap;
  int nShots = acqPar->nShots();
  
  VERIFYC((o) >= 0);
  VERIFYC((j) >= 1);
  if (par->n_shot_snap == 0)	{ //record all shots
         par->n_shot_snap = (nShots-1-o)/j + 1;
  } else {
         VERIFYC(par->n_shot_snap >= 1);
         VERIFYC(par->f_shot_snap + (par->n_shot_snap-1)*(par->j_shot_snap) < nShots);
  }
}

} //end namespace
#endif
