#include "stdafx.h"
#include "cubeData.hpp"
#include "velFD2d.h"
#include "WEMVA.h"
#include "waveOps.h"
#include "dampingFunc.h"
#include "misc.h"
#include "cubeAlgo2.h"
#include "point2d.hpp"
#include "commonOps.h"

using namespace Yang;

void WEMVA_FD_beamFocus::flipGradient(cubeData<float>& gradCube) {
   gradCube.scale(-1.0f); //This -1 belongs to the negative sign for the backprojImage.
}

//virtual function, compute the objective function
float WEMVA_FD_beamFocus::calcObjFunc() {
  FUNC_TRACE_BEG(WEMVA_FD_beamFocus::calcObjFunc);
  LOG(calcObjFunc::WEMVA_FD_beamFocus);
  float costFunc = this->stackPowerObjFunc();
  prt("finish calc Obj Func in WEMVA_FD_beamFocus, obj = %f\n", costFunc);
  FUNC_TRACE_END(costFunc);
}
  
///convert between (z,x,hx) offset image and (z,angle,x) angel domain image.
///In 3d, we have z,angle,azim,x,y as the angle domain image.
void WEMVA_FD::convertAngleOffImage(cubeData<float>& img_ax_t, cubeData<float>& img_hx, bool bOff2Ang ) {
  FUNC_TRACE_BEG(WEMVA_FD::convertAngleOffImage);
  const hypercube& shapeOffImage = img_hx.domain();
  VERIFY(shapeOffImage.get_ndim()==3);
  
  vecAxis axes; axes.push_back(shapeOffImage.get_axis(1));
  axes.push_back(shapeOffImage.get_axis(3)); axes.push_back(shapeOffImage.get_axis(2));
  hypercube img_hx_t_shape(axes);
  cubeData<float> img_hx_t; img_hx_t.alloc(img_hx_t_shape);
  axis offAxis = shapeOffImage.get_axis(3);
  const waveOpParams* waveOpPar = this->waveOp()->waveOpPars().get();
  
  bool bOneWay = waveOpPar->bOneway();
  prt("In WEMVA_FD::convertAngleOffImage,");PRV(m_pWEMVAparams->off2angType); 
  if (bOff2Ang) {
    //First transpose
    //Assume 2D tranpose from z,x,hx -> z,hx,x, thus plane=23
    seplib::callSepTool(img_hx_t,img_hx,"Transp","plane=23");
    //Then convert to angle domain
    //need to stuff dummy axis for OFF2ANG program, from z,hx,x --> z,hx,hy,x,y
    //img_hx_t.info("img_hx_t");
    Yang::off2Ang2d(img_ax_t,img_hx_t,false,
                    m_pWEMVAparams->off2angType,m_pWEMVAparams->amax,m_pWEMVAparams->ax_ang_n);
    if (bOneWay) { //divide the angle gather by cos^2(gamma) to balance the amplitude.
      axis ax_p = img_ax_t.domain().get_axis(2);
      float fang;
      cubeData<float> cosr(ax_p);
      for (int ia=0; ia<ax_p.n; ia++) {
        fang = atanf(ax_p.get_coord(ia));
        cosr(ia) = (SQ(1.0/cosf(fang)));
      }
      //cosr.dump("dump10"); cosr.info("cosr");
      multiplySubCube(img_ax_t,img_ax_t,cosr,2,2);
    }
  } else {
    float hmax = -1.0f * offAxis.o;
    Yang::off2Ang2d(img_hx_t,img_ax_t,true,
                    m_pWEMVAparams->off2angType,hmax,offAxis.n);
    assert(img_hx_t_shape.sameShape(img_hx_t.domain()));
    //Transp
    seplib::callSepTool(img_hx,img_hx_t,"Transp","plane=23");
  }
  FUNC_TRACE_END();
}

//Here dJdu is of the shape (azim,x,z) for 3d, and (x,z) for 2d, sembStack is (z,gamma,azim,x,y) or (z,gamma,x)
void WEMVA_FD::findSmoothDJDUfromSemb(cubeDataF& dJdu, cubeDataF& sembStack, int iSembAxis) {
  //Then either smooth the envelope along z and angle direction and take derivative
  //Or use Gaussian Derivative

  const velAnaParam* pVelAnaParam = this->velAnaPar();
  //Then find the derivative, use 2 pts center difference.
  const axis &ax_semb = pVelAnaParam->ax_semb;
  int iax0            = ax_semb.get_offset(0.0f);
  float ax_semb_d     = ax_semb.d;
  bool bUseGaussianDeriv  = pVelAnaParam->bUseGuassianDerivForDJDU;
  float fSmoothWindow     = pVelAnaParam->fSembSmoothWindow;
  if (!bUseGaussianDeriv) {
    char smoothPar[128]; int smWnd_n = int(fSmoothWindow*0.5f/ax_semb_d);
    assert(smWnd_n < ax_semb.n);
    sprintf(smoothPar,"rect1=%d rect2=%d repeat=%d",5,smWnd_n,2); 
    //should not smooth too much, but here have to use heavy smooth to get rid of the edge effects for c1 case, instrinsically, it is caused by the fact that the dJ/du is not in proportion to the picked mu value itself.
    cubeDataF smoothSemb;
    seplib::callSepTool(smoothSemb,sembStack,"Smooth",string(smoothPar),true);
    smoothSemb.dump("dump10");
    deriv1stAtIndex(dJdu,smoothSemb,iSembAxis,iax0,true);
  } else { //Apply gaussian derivative, this case no need to alloc smoothSemb param
    cubeDataF gaussDeriv;
    Yang::getGaussfunc(gaussDeriv,ax_semb,fSmoothWindow,0.0f,1);
    dotProdAlongAxis(dJdu,sembStack,gaussDeriv,iSembAxis);
    gaussDeriv.dump("dump10");
    dJdu.scale(-1.0f); //doprod + scale -1.0 equals to convolve the Guassian Deriv
  }
}

void WEMVA_FD_eventBased::calcBackProjData() {
  WEMVA_Param* par = m_pWEMVAparams;
  this->postCalcData(); //calc the angle gather at this step
  const AcqPar* acqPar = waveOp()->acqPars().get();
  if (!acqPar->b3D()) {
    this->calcBackProjData2d();
  } else { //3D case
    this->calcBackProjData3d();
  }
}

// calc du/ds (u is the curvature) by inverting the 2X2 hessian matrix with the curvature and static shift moveout parameters (the direct operator).
void WEMVA_FD_beamFocus::calcBackProjData2d() {
  FUNC_TRACE_BEG(WEMVA_FD_beamFocus::calcBackProjData2d);
  
  //The Image0 is been calculated already.
  WEMVA_Param* pWEMVAparam  = m_pWEMVAparams;
  velAnaParam* pVelAnaParam = velAnaPar();
  
  //get angle image in shape (z,ax,x), has been calculated in postCalcData
  //convertAngleOffImage(pWEMVAparam->Img0_ax_t, pWEMVAparam->Img0_hx,true);
  
  const hypercube& shapeOffImage = pWEMVAparam->Img0_hx.domain();
  
  //LOG(afert off2Ang image transform)
  pWEMVAparam->Img0_ax_t.dump("angCube");
  
  //Calculate the dJ/du term first
  cubeData<float>& img_ax_t = pWEMVAparam->Img0_ax_t;
  this->calcDJDU();
  cubeData<float>& dJduCube = pVelAnaParam->dJdbCube; //the dJ/du term
  dJduCube.dump("dJdu");
  
  if (pWEMVAparam->bSingleEvent) {
    if (pWEMVAparam->bUseSingleMoveoutParam) {
      //A cube, for single event, is a function of x,zw,just simply set zw to entire z axis,
      //otherwise the size is of (x,z,zw)
      cubeData<float> Acube;
      hypercube shapeAcube = shapeOffImage.return_compatible(1,2);
      Acube.alloc(shapeAcube); 
      
      cubeData<float>& Ecube = pWEMVAparam->Ecube; //the divider
      hypercube shapeEcube = shapeOffImage.return_compatible(2,2);
      Ecube.alloc(shapeEcube); //one for each x
      
      cubeData<float> Img0_dz(img_ax_t.domain());
      timeDeriv1st(Img0_dz,img_ax_t,true); //Take derivative along z direction
      //Img0_dz.dump("dump3");
      
      cubeData<float> frCube; axis axis_angle = img_ax_t.domain().get_axis(2);
      hypercube frShape(vecAxis(1,axis_angle)); frCube.alloc(frShape);
      this->getfrCube(frCube);
        
      LOGC(Calc E cube (denominator));
      cubeData<float> fr2cube; fr2cube.alloc(frShape);
      fr2cube.multiply(frCube,frCube);
      
      cubeData<float> Img0_dz2(img_ax_t.domain());
      timeDeriv2nd(Img0_dz2,img_ax_t,true);
      Img0_dz2.multiply(img_ax_t);
      Acube.zero(); multAndStackAlongAxis2(Acube,Img0_dz2,&fr2cube);
      sumAlongTimeTrc(Ecube,Acube);
      
      Ecube.dump("Ecube");
      
      Ecube.allocMemoryTo(pWEMVAparam->finalWtCube);
      //Since we need to divide Ecube. It is necessary to build a mask before division based on Ecube's value
      float qt_mask = pWEMVAparam->thresMaxValPercent;
      PRV(pWEMVAparam->thresMaxValPercent);
      
      divideWithCaveat(pWEMVAparam->finalWtCube, dJduCube, Ecube,
                       qt_mask, qt_mask, pVelAnaParam->bDivideEnergyTermUseMask, &(pWEMVAparam->thresMaskCube),true);
      
      pWEMVAparam->thresMaskCube.dump("Emask");
      pWEMVAparam->finalWtCube.dump("finalWt");
      
      pWEMVAparam->finalWtCube.normalize();
      
      
      img_ax_t.allocMemoryTo(m_bpjImg_ax_t);
      m_bpjImg_ax_t = Img0_dz;
    
      multiplySubCube(m_bpjImg_ax_t,m_bpjImg_ax_t,frCube,2,2);
      //m_bpjImg_ax_t.dump("dump10");
      scaleSlice(m_bpjImg_ax_t,m_bpjImg_ax_t,pWEMVAparam->finalWtCube); //multiply on x axis
      
    } else { //Use two parameter fitting
      verbPrt("Use two moveout parameters");
      //In this case we need to do the inversion one by one x location.
      //First calculate the E matrix.
      
      
      cubeData<float>& EmatCube = pWEMVAparam->Ecube; //the divider
      vecAxis axes = shapeOffImage.return_axes(2,2);
      hypercube shapeEslice(axes); //one for each x
      axis ax_tt(3); axes.push_back(ax_tt); //Three planes for E11,E12 and E22.
      hypercube shapeEcube(axes);
  
      EmatCube.alloc(shapeEcube); //one for each x
      cubeData<float> EmatSlice; EmatSlice.becomeProxyCube(shapeEslice);
      
      
      //Get first/2nd derivative
      cubeData<float> Img0_dz(img_ax_t.domain()),Img0_dz2(img_ax_t.domain());
      timeDeriv1st(Img0_dz,img_ax_t,true); //Take derivative along z direction
      timeDeriv2nd(Img0_dz2,img_ax_t,true);
  
      cubeData<float> frCube,grCube; axis axis_angle = img_ax_t.domain().get_axis(2);
      hypercube frShape(vecAxis(1,axis_angle)); 
      frCube.alloc(frShape); grCube.alloc(frShape); //grCube is the constant shift amount.
      this->getfrCube(frCube); //get fr
      velAna_FD::setMoveoutDeriv(grCube,Yang::constant);
      cubeData<float> fr2cube(frShape),gr2cube(frShape),fgrcube(frShape);
      fr2cube.multiply(frCube,frCube); fgrcube.multiply(frCube,grCube); gr2cube.multiply(grCube,grCube);
  
      LOGC(Calc E matrix (denominator));
      Img0_dz2.multiply(img_ax_t);
      
      //calc E11
      windParam windE(2,windowAxis());
      windE[1].n = 1; windE[1].f = 0;
      EmatCube.setProxyXd(&EmatSlice,windE,1);
      
      hypercube shapeAcube = shapeOffImage.return_compatible(1,2);
      cubeData<float> Acube(shapeAcube);
      Acube.zero(); multAndStackAlongAxis2(Acube,Img0_dz2,&fr2cube);
      sumAlongTimeTrc(EmatSlice,Acube);
      
      //build a mask based on EmatSlice value.
      float qt_mask = pWEMVAparam->thresMaxValPercent;
      float vth = qt_mask * EmatSlice.maxvalAbs();
      //PRV(pWEMVAparam->thresMaxValPercent);
      EmatSlice.allocMemoryTo(pWEMVAparam->thresMaskCube);
      buildThresholdMask(pWEMVAparam->thresMaskCube,EmatSlice,vth,true);
      
      //LOG(calc E12);
      windE[1].f = 1;
      EmatCube.setProxyXd(&EmatSlice,windE,1);
      Acube.zero(); multAndStackAlongAxis2(Acube,Img0_dz2,&fgrcube);
      sumAlongTimeTrc(EmatSlice,Acube);
      //LOG(calc E22);
      windE[1].f = 2;
      EmatCube.setProxyXd(&EmatSlice,windE,1);
      Acube.zero(); multAndStackAlongAxis2(Acube,Img0_dz2,&gr2cube);
      sumAlongTimeTrc(EmatSlice,Acube);
      
      EmatCube.dump("Ecube");
      
      LOG("invert E mat");
      cubeData<float> Fcube; EmatCube.allocMemoryTo(Fcube);
      //Now do the back projection part
      axis ax_x = EmatCube.domain().get_axis(1);
      point2dF pt2d; float ptmod;
      for (int ix = 0; ix<ax_x.n; ix++) {
        if (pWEMVAparam->thresMaskCube(ix)<0.5f) {
          continue;
        } else {
          //First solve the F matrix.
          float tmpF[4]; float tmpE[4]; float determ, cond;
          tmpE[0] = EmatCube(0,ix); tmpE[1]=tmpE[2]=EmatCube(1,ix); tmpE[3] = EmatCube(2,ix);
          cond = invert2X2(tmpF,tmpE,determ);
          if (cond > 1e3) {
            verbPrt("\n cond number is very big %f in beamFocus2!! E=%.3e,%.3e,%.3e,%.3e", 
                    cond, tmpE[0], tmpE[1], tmpE[1], tmpE[3]);
          }
          pt2d.x = tmpF[0]; pt2d.y = tmpF[1];
          ptmod = 1.0f;
          if (pWEMVAparam->bBalanceEnumeratorWt) {
            ptmod = 1.0f/pt2d.module();
            pt2d *= ptmod;
          }
          Fcube(0,ix) = pt2d.x; Fcube(1,ix) = pt2d.y; Fcube(2,ix) = tmpF[3]*ptmod; //will not use the 3rd element.
        }
      }
      Fcube.allocMemoryTo(pWEMVAparam->finalWtCube);
      multiplySlice(pWEMVAparam->finalWtCube,Fcube,dJduCube);
      
      Fcube.dump("Fcube");
      pWEMVAparam->thresMaskCube.dump("Emask");
      pWEMVAparam->finalWtCube.dump("finalWt");
      pWEMVAparam->finalWtCube.normalize();
      
      //I_z(z,ax,x), need to add two terms.
      img_ax_t.allocMemoryTo(m_bpjImg_ax_t);
      
      //LOG(assemble the bpj angle image gather);
      //Take F11 slice
      windE[1].f = 0; pWEMVAparam->finalWtCube.setProxyXd(&EmatSlice,windE,1);
      multiplySubCube(m_bpjImg_ax_t,Img0_dz,frCube,2,2);
      scaleSlice(m_bpjImg_ax_t,m_bpjImg_ax_t,EmatSlice);
      //Take F12 slice
      windE[1].f = 1; pWEMVAparam->finalWtCube.setProxyXd(&EmatSlice,windE,1);
      cubeData<float> img_ax_t2(m_bpjImg_ax_t.domain());
      multiplySubCube(img_ax_t2,Img0_dz,grCube,2,2);
      scaleSlice(img_ax_t2,img_ax_t2,EmatSlice);
      m_bpjImg_ax_t.add(img_ax_t2);
    }
  }else { //If Multiple events
    axis ax_beam_z = m_pWEMVAparams->ax_beam_z; //For dividing windows along z directions
    axis  ax_z = shapeOffImage.get_axis(1); int nz = ax_z.n;
    float hfBeam_z_window = m_pWEMVAparams->beam_z_window*0.5f;
    axis ax_x = shapeOffImage.get_axis(2);
    if (pWEMVAparam->bUseSingleMoveoutParam) {
      VERIFY(0);
    } else { //Use two parameter fitting
      verbPrt("\nUse two moveout parameters");
      
      //In this case we need to do the inversion one by one on (x,z_b) location.
      //First calculate the E matrix.
      cubeData<float>& EmatCube = pWEMVAparam->Ecube; //the divider
      vecAxis axes; axes.push_back(ax_beam_z);axes.push_back(ax_x);
      hypercube shapeEslice(axes);
      axis ax_tt(3); axes.push_back(ax_tt); //Three planes for E11,E12 and E22.
      hypercube shapeEcube(axes);

      EmatCube.alloc(shapeEcube);
      cubeData<float> EmatSlice; EmatSlice.becomeProxyCube(shapeEslice);

      //Get first/2nd derivative
      cubeData<float> Img0_dz(img_ax_t.domain()),Img0_dz2(img_ax_t.domain());
      timeDeriv1st(Img0_dz,img_ax_t,true); //Take derivative along z direction
      timeDeriv2nd(Img0_dz2,img_ax_t,true);

      cubeData<float> frCube,grCube; axis axis_angle = img_ax_t.domain().get_axis(2);
      hypercube frShape(vecAxis(1,axis_angle)); 
      frCube.alloc(frShape); grCube.alloc(frShape); //grCube is the constant shift amount.
      this->getfrCube(frCube); //get fr
      velAna_FD::setMoveoutDeriv(grCube,Yang::constant);
      cubeData<float> fr2cube(frShape),gr2cube(frShape),fgrcube(frShape);
      fr2cube.multiply(frCube,frCube); fgrcube.multiply(frCube,grCube); gr2cube.multiply(grCube,grCube);

      LOGC(Calc E matrix (denominator) in multiple events case);
      Img0_dz2.multiply(img_ax_t);

      hypercube shapeAcube = shapeOffImage.return_compatible(1,2);
      cubeData<float> Acube(shapeAcube),Bcube(shapeAcube),Ccube(shapeAcube);
      //stack over angles.
      Acube.zero(); multAndStackAlongAxis2(Acube,Img0_dz2,&fr2cube);
      Bcube.zero(); multAndStackAlongAxis2(Bcube,Img0_dz2,&fgrcube);
      Ccube.zero(); multAndStackAlongAxis2(Ccube,Img0_dz2,&gr2cube);
      
      LOG(calc E matrix);
      int izwbeg,izwend;
      for (int ix=0; ix<ax_x.n; ix++) {
        for (int izb=0; izb<ax_beam_z.n; izb++) {
          izwbeg = ax_z.get_offset(ax_beam_z.get_coord(izb)-hfBeam_z_window);
          izwend = ax_z.get_offset(ax_beam_z.get_coord(izb)+hfBeam_z_window);
          if (izwbeg>=0 && izwbeg<=nz-1 && izwend>=0 && izwend<=nz-1) {
            //Sum along the small window along z
            for (int iiz=izwbeg; iiz<=izwend; iiz++) {
              EmatCube(0,ix,izb) +=  Acube(ix,iiz);
              EmatCube(1,ix,izb) +=  Bcube(ix,iiz);
              EmatCube(2,ix,izb) +=  Ccube(ix,iiz);
            }
          }
        }
      }
      //take E11 for mask building.
      windParam windE(3,windowAxis());
      windE[2].n = 1; windE[2].f = 0;
      EmatCube.setProxyXd(&EmatSlice,windE,2);
      
      //build a mask based on EmatSlice value, !!Maybe we should use different threshold value for different the events.
      float qt_mask = pWEMVAparam->thresMaxValPercent;
      float vth = qt_mask * EmatSlice.maxvalAbs();
      PRV(pWEMVAparam->thresMaxValPercent);
      EmatSlice.allocMemoryTo(pWEMVAparam->thresMaskCube);
      buildThresholdMask(pWEMVAparam->thresMaskCube,EmatSlice,vth,true);

      EmatCube.dump("Ecube");

      LOG(invert E mat in multiple event case);
      cubeData<float> Fcube; EmatCube.allocMemoryTo(Fcube);
      //Now do the back projection part
      //First solve the F matrix.
      point2dF pt2d; float ptmod;
      float tmpF[4]; float tmpE[4]; float determ, cond;
      for (int ix = 0; ix<ax_x.n; ix++) { 
        for (int izb=0; izb<ax_beam_z.n; izb++) {
        if (pWEMVAparam->thresMaskCube(ix,izb)<0.5f) { //don't operate on mask=0
          continue;
        } else {
          tmpE[0] = EmatCube(0,ix,izb); tmpE[1]=tmpE[2]=EmatCube(1,ix,izb); tmpE[3] = EmatCube(2,ix,izb);
          cond = invert2X2(tmpF,tmpE,determ);
          if (cond > 1000.0f) {
            verbPrt("\n cond number is very big %f in beamFocus2!! E=%.3e,%.3e,%.3e,%.3e F=%.3e,%.3e,%.3e", 
                    cond, tmpE[0], tmpE[1], tmpE[2], tmpE[3], tmpF[0],tmpF[1],tmpF[3]);
          }
          pt2d.x = tmpF[0]; pt2d.y = tmpF[1];
          ptmod = 1.0f;
          if (pWEMVAparam->bBalanceEnumeratorWt) {
            ptmod = 1.0f/pt2d.module();
            pt2d *= ptmod;
          }
          //will not use the 3rd element.
          Fcube(0,ix,izb) = pt2d.x; Fcube(1,ix,izb) = pt2d.y; Fcube(2,ix,izb) = tmpF[3]*ptmod; 
        }
        }
      }
      
      
      //I_z(z,ax,x), need to add two terms., define wt(ax,zb,x);
      cubeData<float> wt2[2];
      axes.resize(3); axes[0] = axis_angle; axes[1] = ax_beam_z; axes[2]=ax_x;
      hypercube wtShape(axes);
      wt2[0].alloc(wtShape); wt2[1].alloc(wtShape);
      
      
      LOG(comp the F11* + F12* term);
      //(Take F11(z_b,x) * fr(ax) + F12(z_b ,x) * gr(ax)) * Iz(zw+z_b,ax;x,z_b);
      //wt2[0].info("wt2[0]"); EmatSlice.info("EmatSliceinfo");
      windE[2].f = 0;
      Fcube.setProxyXd(&EmatSlice,windE,2);
      duplicateAlongAxis(wt2[0],EmatSlice,1);
      multiplySlice(wt2[0],wt2[0],frCube);

      windE[2].f = 1;
      Fcube.setProxyXd(&EmatSlice,windE,2);
      duplicateAlongAxis(wt2[1],EmatSlice,1);
      multiplySlice(wt2[1],wt2[1],grCube);
      wt2[0].add(wt2[1]);
      
      Fcube.dump("Fcube"); wt2[0].dump("dump5"); wt2[1].dump("dump6");
      wt2[0].allocMemoryTo(pWEMVAparam->finalWtCube);
      scaleSlice(pWEMVAparam->finalWtCube,wt2[0],dJduCube);
      pWEMVAparam->thresMaskCube.dump("Emask");
      pWEMVAparam->finalWtCube.dump("finalWt");
      pWEMVAparam->finalWtCube.normalize();

      img_ax_t.allocMemoryTo(m_bpjImg_ax_t);
      cubeDataF& imgax = m_bpjImg_ax_t;
      cubeDataF scaleCube; //size of (z,ax,x), a multiplier at each image location
      imgax.allocMemoryTo(scaleCube);

      spreadWtFromCoarseToFineGrid(scaleCube, pWEMVAparam->finalWtCube);
      scaleCube.dump("wtAll");
      imgax.multiply(Img0_dz,scaleCube);
    } //end if use single moveout param
  } //end if bSingle event
  
  m_bpjImg_ax_t.dump("angCubeBp");
  //convert it back to offset image
  //LOG(Convert ax_t image back to offset image);
  pWEMVAparam->backProjImg_hx.alloc(shapeOffImage);
  convertAngleOffImage(m_bpjImg_ax_t,pWEMVAparam->backProjImg_hx,false);

  FUNC_TRACE_END();
}

void WEMVA_FD_beamFocus::calcDJDU() {
  WEMVA_Param* par = m_pWEMVAparams;
  const AcqPar* acqPar = waveOp()->acqPars().get();
  if (!acqPar->b3D()) {
    this->calcDJDU2d();
  } else { //3D case
    this->calcDJDU3d();
  }
}

void WEMVA_FD_beamFocus::calcDJDU2d() {
  assert(m_bCalcAngleGather);
  //Assume the angle image should be ready
  assert(!m_pWEMVAparams->Img0_ax_t.domain().isEmpty());
  cubeData<float>& img_ax_t = m_pWEMVAparams->Img0_ax_t;
  
  velAnaParam* pVelAnaParam = velAnaPar();
  
  cubeData<float>& dJduCube = pVelAnaParam->dJdbCube; //the dJ/du term
  hypercube shapeOffImage = m_pWEMVAparams->Img0_hx.domain();
  
  if (m_pWEMVAparams->bSingleEvent) {
    hypercube shapeDJDU = shapeOffImage.return_compatible(2,2);
    dJduCube.alloc(shapeDJDU); //one for each x
    if (m_pWEMVAparams->bOnlyCalcSensitivityKernel) {
      axis axis_x = shapeDJDU.get_axis(1);

      //Put 6 spikes spike for sensitivity kernel
      /**
      int dnw = axis_x.n/8;
      for (int i=0; i<7; i++) {
        dJduCube(axis_x.n/2+(i-3)*dnw) = 1.0f;
      }
      **/
      dJduCube(axis_x.n/2) = 1.0f;
    } else {
      //Use the analytic form
      if (!pVelAnaParam->bUseSemblanceToCalcDJdb) {
        //Try to calculate dJ over moveout parms.
        cubeData<float> Acube,Bcube;
        //A cube, for single event, is a function of x,zw,just simply set zw to entire z axis,
        //otherwise the size is of (x,z,zw)
        hypercube shapeABcube = shapeOffImage.return_compatible(1,2);
        Acube.alloc(shapeABcube); Bcube.alloc(shapeABcube);
        
        //LOG(calc A cube);
        multAndStackAlongAxis2(Acube, img_ax_t,NULL);
        Acube.dump("Acube");
        
        LOG(calc B cube);
        cubeData<float> Img0_dz(img_ax_t.domain());
        timeDeriv1st(Img0_dz,img_ax_t,true); //Take derivative along z direction
        
        cubeData<float> frCube; axis axis_angle = img_ax_t.domain().get_axis(2);
        hypercube frShape(vecAxis(1,axis_angle)); frCube.alloc(frShape);
        this->getfrCube(frCube);
        
        multAndStackAlongAxis2(Bcube,Img0_dz,&frCube);
        Bcube.dump("Bcube"); //PRV(Bcube.domain().get_n23());
        Bcube.multiply(Acube);
        
        dJduCube.zero(); sumAlongTimeTrc(dJduCube,Bcube);
      }
    }
  } else { 
    //Do multiple events. Now the DJDU term is of (x_b,z_b), this grid might be coarser than the image grid.
    axis ax_beam_z = m_pWEMVAparams->ax_beam_z; //For dividing windows along z directions
    axis ax_z = shapeOffImage.get_axis(1); int nz=ax_z.n;
    float hfBeam_z_window = m_pWEMVAparams->beam_z_window*0.5f;
    axis ax_x = shapeOffImage.get_axis(2);
    vecAxis axes; axes.push_back(ax_beam_z); axes.push_back(ax_x);
    hypercube shapeDJDU(axes);
    dJduCube.alloc(shapeDJDU); //one for each (x_b,z_b);
    if (m_pWEMVAparams->bOnlyCalcSensitivityKernel) {
      axis axis_x = shapeDJDU.get_axis(2);
      int dnw = axis_x.n/4;
      //Put a series of spikes in the center x for sensitivity kernel
      for (int i=0; i<3; i++) {
      for (int izb=5; izb<ax_beam_z.n; izb+=6) {
        dJduCube(ax_x.n/2+(i-1)*dnw,izb) = 1.0f;
      }}
    } else {
      if (!pVelAnaParam->bUseSemblanceToCalcDJdb) {
        //Try to calculate dJ over moveout parms
        cubeData<float> Acube,Bcube;
        //A cube, for multiple event, is a function of x,z,zw, but here 
        //use it to save the angle stack (z,x)
        axes.resize(2); axes[0] = ax_z; axes[1] = ax_x;
        hypercube shapeABcube(axes);
        Acube.alloc(shapeABcube); Bcube.alloc(shapeABcube);
  
        LOG(Do angle stack to get A cube); 
        multAndStackAlongAxis2(Acube, img_ax_t,NULL);
        LOG(Do angle stack to get B cube);
        cubeData<float> Img0_dz(img_ax_t.domain());
        timeDeriv1st(Img0_dz,img_ax_t,true); //Take derivative along z direction
  
        cubeData<float> frCube; axis axis_angle = img_ax_t.domain().get_axis(2);
        hypercube frShape(vecAxis(1,axis_angle)); frCube.alloc(frShape);
        this->getfrCube(frCube);
        multAndStackAlongAxis2(Bcube,Img0_dz,&frCube);
        // Bcube.dump("dump2");
        Bcube.multiply(Acube);
        // Acube.dump("dump3"); Bcube.dump("ABcube"); 
        //PRV(Bcube.domain().get_n23());
        LOGC(now start calc ing the dJdU);
        dJduCube.zero();
        int izwbeg,izwend;
        for (int ix=0; ix<ax_x.n; ix++) {
          for (int izb=0; izb<ax_beam_z.n; izb++) {
            izwbeg = ax_z.get_offset(ax_beam_z.get_coord(izb)-hfBeam_z_window);
            izwend = ax_z.get_offset(ax_beam_z.get_coord(izb)+hfBeam_z_window);
            if (izwbeg>=0 && izwbeg<=nz-1 && izwend>=0 && izwend<=nz-1) {
              //Sum along the small window along z
              for (int iiz=izwbeg; iiz<=izwend; iiz++) {
                dJduCube(ix,izb) = dJduCube(ix,izb) + Bcube(ix,iiz);
              }
            }
          }
        }
      }
    }
  }
  
  //If use semblance panel to determine the moveout parameters
  if (!m_pWEMVAparams->bOnlyCalcSensitivityKernel && pVelAnaParam->bUseSemblanceToCalcDJdb)   {
    LOG(create semblance panel);
    //set the semblance panel size.
    cubeDataF& semb = pVelAnaParam->semblanceCube;
    cubeDataF envSemb;
    vecAxis axes = img_ax_t.domain().get_axes();
    axes[1] = pVelAnaParam->ax_semb;
    semb.alloc(axes);
    float fAntiAlias=1.0f;
    //First Calculate the semblance panel, based on the formula given in SimonLuo's abstract, basically stack over parabolas then smooth along t.
    calcMoveoutSemblance(semb, img_ax_t, Yang::Quadratic,fAntiAlias, true);
    //seplib::callSepTool(envSemb,semb,"Envelope","",true);
    //instead, just square the semblance
    semb.allocMemoryTo(envSemb);
    envSemb.multiply(semb,semb);
    seplib::callSepTool(envSemb,envSemb,"Smooth"," repeat=2 rect1=3 rect2=1 ",false); 
    bool bNormalizeSemb = pVelAnaParam->bNormalizeSemb; //Divide by the sum of squares over angles, not square of sums over angles.
    //envSemb.dump("dump7");
    if (bNormalizeSemb) {
      LOG(calc the normalized semblance);
      //Calculate the denominator
      cubeDataF img_ax_t_sq, semb2;
      img_ax_t.copyTo(img_ax_t_sq); img_ax_t_sq.multiply(img_ax_t);
      semb.copyTo(semb2);
      
      //This way to calc denom ??? is it the best way ???????
      calcMoveoutSemblance(semb2, img_ax_t_sq, Yang::Quadratic,fAntiAlias, true);
      //seplib::callSepTool(semb2,semb2,"Envelope","",true); //since we applied envelope above
      //Smooth the panel and add epsilon
      seplib::callSepTool(semb2,semb2,"Smooth"," repeat=2 rect1=3 rect2=1 ",false);
      float fRMS = semb2.RMS_NZ(1e-4); //PRV(fRMS);
      semb2.add(fRMS*0.05);
      //envSemb.dump("dump9");  img_ax_t_sq.dump("dump10");
      envSemb.divide(semb2); envSemb.scale(1.0f/(pVelAnaParam->ax_semb.n));
    }
    
    //Do slightly smoothing in this case
    //seplib::callSepTool(envSemb,envSemb,"Smooth","rect1=4 rect2=3 repeat=2 ",true);
    //semb.dump("dump3"); 
    envSemb.dump("sembS");
    
    cubeDataF sembZstack;
    bool bPickMoveoutPar = pVelAnaParam->bPickMoveoutPar;
    PRV(bPickMoveoutPar);
    //The put it to dJ over dU cube
    if (m_pWEMVAparams->bSingleEvent) { //if single events
      hypercube shapeDJDU = shapeOffImage.return_compatible(2,2);
      dJduCube.alloc(shapeDJDU); //one for each x
      //First stack the semblance along the z window, the shape is (mu,x)
      sembZstack.alloc(envSemb.domain().return_compatible(2,3));
      sumAlongTimeTrc(sembZstack, envSemb);
      if (!bPickMoveoutPar) { //Take derivative
        this->findSmoothDJDUfromSemb(dJduCube,sembZstack,1);        
      } else {
        findMaxValTimePerTrc(dJduCube,sembZstack);
      }
    } else { //multiple events.
      axis ax_beam_z = m_pWEMVAparams->ax_beam_z; //For dividing windows along z directions
      axis ax_z = shapeOffImage.get_axis(1); //int nz=ax_z.n;
      float hfBeam_z_window = m_pWEMVAparams->beam_z_window*0.5f;
      axis ax_x = shapeOffImage.get_axis(2);
      vecAxis axes; axes.push_back(ax_beam_z); axes.push_back(ax_x);
      hypercube shapeDJDU(axes);
      dJduCube.alloc(shapeDJDU); //one for each (x_b,z_b);      
      int izwbeg,izwend;
      //Here the way I implement is summing the values in the smoothed envelope as in a beam window and then take derivatives or find maximum values
      //determine the stacked cube shape the shape is (mu,z_b,x)
      axis ax_mu = envSemb.domain().get_axis(2);
      axes.resize(3); axes[0] = ax_mu; axes[1]= ax_beam_z; axes[2]=ax_x;
      sembZstack.alloc(axes);      
      
      int nz = ax_z.n;
      for (int ix=0; ix<ax_x.n; ix++) {
        for (int izb=0; izb<ax_beam_z.n; izb++) {
          izwbeg = ax_z.get_offset(ax_beam_z.get_coord(izb)-hfBeam_z_window);
          izwend = ax_z.get_offset(ax_beam_z.get_coord(izb)+hfBeam_z_window);
          assert (izwbeg>=0 && izwbeg<=nz-1 && izwend>=0 && izwend<=nz-1);
          //Sum along the small window along z
          for (int imu=0; imu<ax_mu.n; imu++) {
          for (int iiz=izwbeg; iiz<=izwend; iiz++) {
            sembZstack(ix,izb,imu) += envSemb(ix,imu,iiz);
          }
          }
        }
      }
      //sembZstack.dump("dump5");
      if (!bPickMoveoutPar) { //Take derivative
        this->findSmoothDJDUfromSemb(dJduCube,sembZstack,1);
      } else {
        findMaxValTimePerTrc(dJduCube,sembZstack);
      }
    }
  }
  return;
}

//Use tan(\gamma)^2
void WEMVA_FD_beamFocus::getfrCube(cubeData<float>& frCube) {
  if (m_pWEMVAparams->off2angType == slantstackYang || m_pWEMVAparams->off2angType == slantstackYangAA ) {
    velAna_FD::setMoveoutDeriv(frCube, Quadratic);
  } else {
    //assume the unit is in radians, not degree
    VERIFY(frCube.domain().get_d(1) < 0.9f); //if degree, then it is likely to be bigger than 0.9deg
    velAna_FD::setMoveoutDeriv(frCube, tangentSquare);
  }
}

//Calculate Gradient, calc du/ds using least square fitting projection (the indirect operator).
void WEMVA_FD_beamFocus_m::calcBackProjData2d() {
  FUNC_TRACE_BEG(WEMVA_FD_beamFocus::calcBackProjData);
  
  //The Back Proj data is the Image I(x,h)h^2
  //The Image0 is been calculated already.
  WEMVA_Param* pWEMVAparam  = m_pWEMVAparams;
  velAnaParam* pVelAnaParam = velAnaPar();
  
  //get angle image in shape (z,ax,x)
  //convertAngleOffImage(pWEMVAparam->Img0_ax_t, pWEMVAparam->Img0_hx,true);
  const hypercube& shapeOffImage = pWEMVAparam->Img0_hx.domain();
  
  //LOG(afert off2Ang image transform)
  pWEMVAparam->Img0_ax_t.dump("angCube");
  
  //Calculate the dJ/du term first
  cubeData<float>& img_ax_t = pWEMVAparam->Img0_ax_t;
  axis ax_x = img_ax_t.domain().get_axis(3);
  
  this->calcDJDU();
  cubeData<float>& dJduCube = pVelAnaParam->dJdbCube; //the dJ/du term
  dJduCube.dump("dJdu");
  
  //After get DJDU, we need to backproject DJDU (delta u) to the individual shift at each angle
  cubeData<float> Img0_dz(img_ax_t.domain());
  timeDeriv1st(Img0_dz,img_ax_t,true); //Take derivative along z direction
  cubeData<float> Img0_dz2(img_ax_t.domain());
  timeDeriv2nd(Img0_dz2,img_ax_t,true);
  
  cubeData<float>& Ecube = pWEMVAparam->Ecube; //the divider, now E is of (ax,x);
  
  Img0_dz2.multiply(img_ax_t);
  VERIFY(pWEMVAparam->bUseSingleMoveoutParam);
  
  cubeData<float> frCube; 
  axis axis_angle = img_ax_t.domain().get_axis(2); int na = axis_angle.n;
  hypercube frShape(vecAxis(1,axis_angle)); frCube.alloc(frShape);
  this->getfrCube(frCube);

  //problematic, here Ecube is not energy term !!!! Should not use Ecube as weight, especially in the case of multiple events.
  int i;
  if (pWEMVAparam->bUseSingleMoveoutParam) {
    if (pWEMVAparam->bSingleEvent) {
      hypercube shapeEcube = img_ax_t.domain().return_compatible(2,3);
      Ecube.alloc(shapeEcube); //one for each (ax,x);
      sumAlongTimeTrc(Ecube,Img0_dz2);
      cubeData<float> crCube(frCube.domain());
      cubeData<float> cr_x_cube(img_ax_t.domain().return_compatible(2,3));
      windParam cr_x_wind(2);  cr_x_wind[1].n = 1;
      
      if (!pWEMVAparam->bUseWeightedLSfit) {
        float r2mean = 0.0f;
        if (pWEMVAparam->bNoLSfit) {
          r2mean = 0.0f;
        } else {
          r2mean = frCube.sum() / float(na);
        }
        float fDenom = 0.0f;
        for (i=0; i<axis_angle.n; i++) {
          crCube(i) = frCube(i) - r2mean;
          fDenom += SQ(crCube(i));
        }
        crCube.scale(1.0f/fDenom);
        crCube.dump("dump4");
       
        //Calc the final Wt, wt Cube is of size (ax,x)
        cubeData<float>& wtCube = pWEMVAparam->finalWtCube;
        Ecube.allocMemoryTo(wtCube);
        wtCube = 1.0f;
        scaleSlice(wtCube, wtCube, dJduCube);
        multiplySlice(wtCube, wtCube, crCube);
        
        LOG(apply mask);
        //Since we need to divide Ecube. It is necessary to build a mask before division based on Ecube's value
        float qt_mask = pWEMVAparam->thresMaxValPercent;
        PRV(pWEMVAparam->thresMaxValPercent);
        bool bMaskResultToo = true;
        divideWithCaveat(wtCube, wtCube, Ecube,
                         qt_mask, qt_mask, pVelAnaParam->bDivideEnergyTermUseMask, 
                         &(pWEMVAparam->thresMaskCube),bMaskResultToo);
        pWEMVAparam->thresMaskCube.dump("Emask");
      } else { //The weighted sum like r2 mean will be weighted by the energy.
        LOG(use weighted LSQ fitting for delta z);
        float r2sum, r4sum, wt1sum, fdenom, ftmp;
        cubeDataF denomCube(ax_x.n),r2sumCube(ax_x.n),r4sumCube(ax_x.n),wt1sumCube(ax_x.n);
        //Do it one by one x location, assume Ecube holds the enery weight w^2
        for (int ix=0; ix<ax_x.n; ix++) {
          //First calc the average terms.
          r2sum = r4sum = wt1sum = 0.0f;
          for (i=0; i<axis_angle.n; i++) {
            ftmp = fabs(Ecube(ix,i));
            wt1sum  += ftmp;
            r2sum   += frCube(i)*ftmp;
            r4sum   += SQ(frCube(i))*ftmp;
          }
          fdenom = wt1sum*r4sum - SQ(r2sum);
          denomCube(ix) = fdenom;
          r2sumCube(ix) = r2sum; r4sumCube(ix) = r4sum; wt1sumCube(ix) = wt1sum;
        }
        //Build a mask on denom.
        float fvThres  = denomCube.maxvalAbs() * pWEMVAparam->thresMaxValPercent;
        for (int ix=0; ix<ax_x.n; ix++) {
          fdenom = denomCube(ix);
          r2sum = r2sumCube(ix); r4sum = r4sumCube(ix); wt1sum = wt1sumCube(ix);
          
          if (fdenom > fvThres) {
            for (i=0; i<axis_angle.n; i++) {
              crCube(i) = (wt1sum*frCube(i) - r2sum)/fdenom ;
            }
            //Should put crCube into a (ax,x) cube and multiply with dJdu to get the final wt.
            //No need to divide the E term, becauses it been cancelled off from the frac expression.
            cr_x_wind[1].f = ix;
            cr_x_cube.update1d(&crCube,cr_x_wind);
          }
        }
        cr_x_cube.dump("Bcube");
        
        //Calc the final Wt, wt Cube is of size (ax,x)
        cubeData<float>& wtCube = pWEMVAparam->finalWtCube;
        Ecube.allocMemoryTo(wtCube);
        wtCube = 1.0f;
        scaleSlice(wtCube, wtCube, dJduCube);
        wtCube.multiply(cr_x_cube);
        wtCube.scale(-1.0f); //because I did not actually divide C cube.
      }
      pWEMVAparam->finalWtCube.dump("finalWt");
      pWEMVAparam->finalWtCube.normalize();
      
      LOG(generate final angle image);
      img_ax_t.allocMemoryTo(m_bpjImg_ax_t);
      m_bpjImg_ax_t = Img0_dz;
      //scale on (ax,x) axis, the last two axes.
      scaleSlice(m_bpjImg_ax_t,m_bpjImg_ax_t,pWEMVAparam->finalWtCube); 
    } else { //if do multiple event
      axis& ax_beam_z = m_pWEMVAparams->ax_beam_z; //For dividing windows along z directions
      axis ax_z = shapeOffImage.get_axis(1); int nz = ax_z.n;
      float hfBeam_z_window = m_pWEMVAparams->beam_z_window*0.5f;
           
      vecAxis axes(3); axes[0]=axis_angle;axes[1]=ax_beam_z;axes[2]=ax_x;
      hypercube shapeEcube(axes); // (angle,z_b,x)
      Ecube.alloc(shapeEcube);

      cubeData<float> crCube(frCube.domain());
      cubeData<float> cr_x_cube(shapeEcube); //(angle,z_b,x);
      windParam cr_x_wind(3);  cr_x_wind[1].n = 1; cr_x_wind[2].n = 1;
      
      int izwbeg,izwend;
      //Calc E cube
      //Do it one by one beam (z_b,x) location, assume Ecube holds the enery weight w^2
      for (int ix=0; ix<ax_x.n; ix++) {
      for (int izb=0; izb<ax_beam_z.n; izb++) {
        izwbeg = ax_z.get_offset(ax_beam_z.get_coord(izb)-hfBeam_z_window);
        izwend = ax_z.get_offset(ax_beam_z.get_coord(izb)+hfBeam_z_window);
        assert (izwbeg>=0 && izwbeg<=nz-1 && izwend>=0 && izwend<=nz-1);
        for (i=0; i<axis_angle.n; i++) {
          for (int iiz=izwbeg; iiz<=izwend; iiz++) {
            Ecube(ix,izb,i) += Img0_dz2(ix,i,iiz);  
          }
        }
      }}
      if (!pWEMVAparam->bUseWeightedLSfit) {
        //Just use deltab = delta alpha * tan^2(gamma);
        float fDenom = 0.0f;
        for (i=0; i<axis_angle.n; i++) {
          crCube(i) = frCube(i);
          fDenom += SQ(crCube(i));
        }
        crCube.scale(1.0f/fDenom);
        for (int ix=0; ix<ax_x.n; ix++) {
          for (int izb=0; izb<ax_beam_z.n; izb++) {
            for (int iang=0; iang<axis_angle.n;iang++) {
              cr_x_cube(ix,izb,iang) = crCube(iang);
            }
          }
        }
      } else { //The weighted sum like r2 mean will be weighted by the energy.
        LOG(use weighted LSQ fitting for delta z in Multiple events case);
        float r2sum, r4sum, wt1sum, fdenom, ftmp;
        hypercube shape_zb_x(ax_beam_z.n,ax_x.n);
        cubeDataF denomCube(shape_zb_x),r2sumCube(shape_zb_x),r4sumCube(shape_zb_x),wt1sumCube(shape_zb_x);
        //Do it one by one beam (z_b,x) location, assume Ecube holds the enery weight w^2
        for (int ix=0; ix<ax_x.n; ix++) {
          for (int izb=0; izb<ax_beam_z.n; izb++) {              
            //First calc the average terms.
            r2sum = r4sum = wt1sum = 0.0f;
            for (i=0; i<axis_angle.n; i++) {
              ftmp = fabs(Ecube(ix,izb,i));
              wt1sum  += ftmp;
              r2sum   += frCube(i)*ftmp;
              r4sum   += SQ(frCube(i))*ftmp;
            }
            fdenom = wt1sum*r4sum - SQ(r2sum);
            denomCube(ix,izb) = fdenom; r2sumCube(ix,izb) = r2sum; 
            r4sumCube(ix,izb) = r4sum; wt1sumCube(ix,izb) = wt1sum;
          }
        }
        //Build a mask on denom.
        float fvThres  = denomCube.maxvalAbs() * pWEMVAparam->thresMaxValPercent;
        for (int ix=0; ix<ax_x.n; ix++) {
          cr_x_wind[2].f = ix;
          for (int izb=0; izb<ax_beam_z.n; izb++) {
            cr_x_wind[1].f = izb;
            izwbeg = ax_z.get_offset(ax_beam_z.get_coord(izb)-hfBeam_z_window);
            izwend = ax_z.get_offset(ax_beam_z.get_coord(izb)+hfBeam_z_window);
            if (izwbeg>=0 && izwbeg<=nz-1 && izwend>=0 && izwend<=nz-1) {
              fdenom = denomCube(ix,izb);
              r2sum = r2sumCube(ix,izb); r4sum = r4sumCube(ix,izb); wt1sum = wt1sumCube(ix,izb);
              if (fdenom > fvThres) {
                for (i=0; i<axis_angle.n; i++) {
                  crCube(i) = (wt1sum*frCube(i) - r2sum)/fdenom ;
                }
                cr_x_wind[1].f = izb;
                cr_x_cube.update1d(&crCube,cr_x_wind);
              }
            }
          }
        }
      }
      cr_x_cube.dump("Bcube");
      //Should put crCube into a (ax,z_b,x) cube and multiply with dJdu to get the final wt.
      
      cubeData<float>& wtCube = pWEMVAparam->finalWtCube;
      img_ax_t.allocMemoryTo(wtCube);

      cubeData<float> scaleCube; //(angle,zb,x)
      Ecube.allocMemoryTo(scaleCube);scaleCube = 1.0f;
      scaleSlice(scaleCube, scaleCube, dJduCube);
      scaleCube.multiply(cr_x_cube);
      if (pWEMVAparam->bUseWeightedLSfit) {
        scaleCube.scale(-1.0f); //because I did not actually divide C cube, there is a negative sign implied.
      } else {
        //divide the E cube.
        float qt_mask = pWEMVAparam->thresMaxValPercent;
        divideWithCaveat(scaleCube,scaleCube,Ecube,qt_mask,qt_mask,true,&(pWEMVAparam->thresMaskCube),true);
      }
      this->spreadWtFromCoarseToFineGrid(wtCube,scaleCube);
      wtCube.dump("finalWt");
      pWEMVAparam->finalWtCube.normalize();
      LOG(generate final angle image);
      img_ax_t.allocMemoryTo(m_bpjImg_ax_t);
      m_bpjImg_ax_t = Img0_dz;
      m_bpjImg_ax_t.multiply(pWEMVAparam->finalWtCube);
    }
  } else { //use two moveout param
    VERIFY(0); //this case we only need one moveout params.
  }//end if use single moveout par
  Ecube.dump("Ecube");
  m_bpjImg_ax_t.dump("angCubeBp");
  //convert it back to offset image
  LOG(m8);
  pWEMVAparam->backProjImg_hx.alloc(shapeOffImage);
  convertAngleOffImage(m_bpjImg_ax_t,pWEMVAparam->backProjImg_hx,false);

  //Put negative sign on it
  pWEMVAparam->backProjImg_hx.scale(-1.0f);
  FUNC_TRACE_END()
}

//Assume the input (angle,z_b,x), the output (z,angle,x)
void WEMVA_FD::spreadWtFromCoarseToFineGrid(cubeDataF& wtCube,const cubeDataF& coarseCube) {
  FUNC_TRACE_BEG(spreadWtFromCoarseToFineGrid);
  VERIFY(!m_pWEMVAparams->bSingleEvent);
  axis& ax_beam_z   = m_pWEMVAparams->ax_beam_z; //For dividing windows along z directions
  axis  ax_z        = wtCube.domain().get_axis(1); int nz = ax_z.n;
  axis  ax_x        = wtCube.domain().get_axis(3);
  axis  ax_angle    = coarseCube.domain().get_axis(1);
  
  float hfBeam_z_window = m_pWEMVAparams->beam_z_window*0.5f;
  wtCube.zero();
  int izwbeg,izwend;
  for (int ix=0; ix<ax_x.n; ix++) {
    for (int izb=0; izb<ax_beam_z.n; izb++) {
      izwbeg = ax_z.get_offset(ax_beam_z.get_coord(izb)-hfBeam_z_window);
      izwend = ax_z.get_offset(ax_beam_z.get_coord(izb)+hfBeam_z_window);
      if (izwbeg>=0 && izwbeg<=nz-1 && izwend>=0 && izwend<=nz-1) {
        for (int i=0; i<ax_angle.n; i++) {
          for (int iiz=izwbeg; iiz<=izwend; iiz++) {
            wtCube(ix,i,iiz) += coarseCube(ix,izb,i);
          }
        }
      }
    }
  }
  FUNC_TRACE_END();
}

