#include "stdafx.h"
#include "dmRsf.hpp"

#include "srcWav.h"
#include "velFD2d.h"
#include "point3d.hpp"

#include "modelingFD2d_op.h"
#include "modelingFDgpu.h"

#include "SR_manager.h"
#include "paddingFD.h"
#include "paddingFD_damping.h"
#include "boundaryFD.h"

#include "param_entry.h"
#include "myOMP.h"

#include "WEMVA.h"
#include "stackTraceC.h"
//#include "../misc.h"


/**
 * For beam focusing, Wave Equation Tomography, gradient 
 * calculation, CUDA enabled. Reflection case
 */


using namespace Yang;
using namespace std;
   
//Doing beam wave-equation tomography for reflection case
/// The input of the program: initial velocity model vel
/// observed data data_obs along with its geometry

int main (int argc, char *argv[])
{
   initCallStackTrace();
   //Read params, params need contain geometry info that is exact the same as the input data. In
   //synthetic this is easy to do.
   auto_ptr<sep_params> autoPars(new sep_params(argc,argv));
   sep_params* pars = autoPars.get();

   auto_ptr<vel2d> velObj(new vel2d(csz_tag_vel,dm::IN,"vel"));
   auto_ptr<srcWav> waveletObj(new srcWav1d(csz_tag_wavelet,dm::IN,"wavelet"));

   //Read in wavelet and initial velocity, input 2nd derivative of Gaussian wavelet 
   waveletObj->loadWavelet();
   velObj->loadVel();
   
   //Create the modeling Obj for modeling and also RTM(for calculating the gradient)
   auto_ptr<modelingFDgpu2d_op> modelingObj(new modelingFDgpu2d_op());
   int gpuID = pars->get_int(csz_gpuID,1);
   modelingObj->setGPU_ID(gpuID);
   
   //Create the padding and boundary processor
   auto_ptr<SR_managerFD2d>  SR_managerObj(new SR_managerFD2d(waveletObj.get()));
   
   //chain two padding method together
   //float dampingExtent = 0.5f;
   //auto_ptr<paddingFD2dDamping> paddingDampingObj ( new paddingFD2dDamping(NULL,dampingExtent));
   auto_ptr<paddingFD2d> paddingObj (new paddingFD2d(NULL));
   //cosDampingFunc dampFunc;
   //paddingDampingObj->m_pDampingFunc = &dampFunc;
		
   //boundary condition
   auto_ptr<boundaryFD2d> boundaryObj(new boundaryFD2d());

   //assemble
   modelingObj->assemble(SR_managerObj.get(), boundaryObj.get(),paddingObj.get());

   //Read in the parameters for the modeling Operator
   FDModelingParams params;
   velAnaParam velAnaParams;
   WEMVA_Param WEMVAparam;
   //load the observed data
   velAnaParams.pDataObs = new cubeData<float>(csz_tag_dataObs,dm::IN,"data observed");
   velAnaParams.pDataObs->init(NULL);
   velAnaParams.pDataObs ->domain().info("Data obs info");

   float maxFreq,vmax,vmin;
   params.pVel = velObj.get();
   params.pWavelet = waveletObj.get();
   //Read in the peak/center freq of the seismic wavelet
   maxFreq = waveletObj->get_float("peakFreq");
   waveletObj->setPeakFreq(maxFreq);
   vmax = velObj->maxval();
   vmin = velObj->minval();

   //The size of whole velocity model
   params.velModelShape = velObj->domain();   
   //set size of mShapeWavfieeld, just simple case, where migration aperture doesnot change
   params.modelGridShape = params.velModelShape;

   //also get the time param from the params
   params.bFreeSurfaceOnTop = pars->get_bool(csz_freeSurface);
   params.stencilWidth = pars->get_int(csz_stenWidth);
   params.bUse4thOrder= pars->get_bool(csz_use4thOrderInTime);
   

   params.ndim = params.velModelShape.get_ndim();
   params.offx = params.stencilWidth;params.offz = params.stencilWidth;
   params.m_nShots 	= pars->get_int(csz_nShots);
   params.mShot1_ix = pars->get_int(csz_shot1_ix);
   params.mShot1_iz = pars->get_int(csz_shot1_iz);
   params.mShotIncX = pars->get_int(csz_shotIncX);
   params.mTrace1ix = pars->get_int(csz_trace1_ix);
   params.mTrace1iz = pars->get_int(csz_trace1_iz);
   params.mTraceIncX = pars->get_int(csz_traceIncX);
   params.m_nTraces = pars->get_int(csz_nTraces);
   params.m_nPad = pars->get_int(csz_npad);

   //Recording Snapshot params
   params.f_shot_snap = pars->get_int(csz_snap_f,0);
   params.j_shot_snap = pars->get_int(csz_snap_j,1);
   params.n_shot_snap = pars->get_int(csz_snap_n,0);
   checkSnapRecordParams(&params);

   //Read in Absorption type
   int itype = pars->get_int(csz_bdType);
   switch (itype)
   {
      case 1:
         params.m_bdType = abs_ord1;
         break;
      case 2:
         params.m_bdType = abs_ord2;
         break;
      case 3:
         params.m_bdType = abs_Dave;
         break;
      default:
         seperr("!unknown boundary condition type!\n");
   }
   
   //Here I set the data rsf objects
   params.mImposeRecvType = AsForce;

   params.pSrcSnapshot 	= new cubeData<float>(csz_tag_srcSnapshot,dm::OUT,"src snapshot");
   params.pRecvSnapshot	= new cubeData<float>(csz_tag_recvSnapshot,dm::OUT,"recv snapshot");
   params.pSrcSnapshot2 	= new cubeData<float>(csz_tag_srcSnapshot2,dm::OUT,"src snapshot");
   params.pRecvSnapshot2	= new cubeData<float>(csz_tag_recvSnapshot2,dm::OUT,"recv snapshot");

   //The input for DSO gradient
   params.pFinalImage	= new cubeData<float>("",dm::INOUT,"");
   params.pFinalOffImage	= new cubeData<float>("",dm::INOUT,""); //!! Memory Leak here
   params.pImages	= new cubeData<float>("",dm::INOUT,"");
   params.pOffImages= new cubeData<float>("",dm::INOUT,""); //!! Memory Leak here

   //The gradient of the local objective function, the possible output
   params.pDeltaSs		= new cubeData<float>("",dm::INOUT,"Initial Image");
   params.pFinalDeltaS	= new cubeData<float>("",dm::INOUT,"Initial Image");
   
   //calc time parameters
   params.timeParam.dtComp = pars->get_float(csz_dtComp);
   params.timeParam.dtSnap = pars->get_float(csz_dtSnap);
   params.timeParam.dtImaging = pars->get_float(csz_dtImaging,params.timeParam.dtRecord);
      
   //Always assume the recording starts from zero time. !! 
   params.timeParam.dtRecord = velAnaParams.pDataObs->domain().get_d(1);
   params.timeParam.totalTime= (velAnaParams.pDataObs->domain().get_n(1)-1) * params.timeParam.dtRecord;
   
   params.timeParam.dtComp = pars->get_float(csz_dtComp);
   params.timeParam.dtSnap = pars->get_float(csz_dtSnap);
   params.timeParam.dtImaging = pars->get_float(csz_dtImaging,params.timeParam.dtRecord);
   
   //calc time parameters
   calcTimeParamsFor2wayModeling(&(params.timeParam),maxFreq,vmax,vmin,params.modelGridShape);
  
   //Load the observed data
   velAnaParams.dObsCube.alloc(velAnaParams.pDataObs->domain());
   velAnaParams.pDataObs->seq_grabAll(&(velAnaParams.dObsCube));
   params.pRsfSeisRec = &velAnaParams.dObsCube;
   
   //Load the off Image parameters.
   bool bDoOffsetGather = pars->get_bool(csz_do_offset,false);
   params.mOffImgParams.bDoOffsetGather = bDoOffsetGather;
   if (bDoOffsetGather) {
	  params.mOffImgParams.dhx = pars->get_int(csz_img_dhx);
	  params.mOffImgParams.nhx = pars->get_int(csz_img_nhx);
	  params.mOffImgParams.ohx = pars->get_int(csz_img_ohx);
	  params.mOffImgParams.AxesUsage = 0x1;
   }

   WEMVAparam.off_img_param = params.mOffImgParams;
   
   int objFuncIndex = pars->get_int("objFunc",1);
   int moveOutFuncType = pars->get_int("moveOutType",2);
   switch(moveOutFuncType) {
   case 1: 
	  velAnaParams.moveOutType = velAnaParam::Linear;
	  break;
   case 2:
	  velAnaParams.moveOutType = velAnaParam::Quadratic;
	  break;
   default:
	  seperr("wrong moveOutType for beam focusing");
	  break;
   }
   
   //Create FWIObj
   float v0,vTrue;
   v0 = pars->get_float("v0",-1);
   vTrue = pars->get_float("vTrue",-1);
   WEMVA_FD* pWEMVAobj = NULL;
   bool bTaperEdge = false;
   
   switch(objFuncIndex) {
   case 6: 
	  prt("using DSO Refl obj Function");
	  velAnaParams.objFuncType = velAnaParam::classic; 
	  pWEMVAobj = new WEMVA_FD_DSO(&velAnaParams, &params, modelingObj.get(), &WEMVAparam);
	  break;
   case 2: 	  break;
   case 5:	  
	  prt("using img-space beam focusing");
	  velAnaParams.objFuncType = velAnaParam::beamFocusing1; 
	  pWEMVAobj = new WEMVA_FD_beamFocus(&velAnaParams, &params, modelingObj.get(), &WEMVAparam);
      WEMVAparam.thresMaxValPercent = pars->get_float("thresMax",0.01);
   case 9: break;
   case 10: break;
   default:
	  seperr("wrong  objFunction index");
	  exit(-1);
   }
   
   //Calculate the data(prestack RTM image)

   //pWEMVAobj->calcData();
   //For Convenience, Just Load the inital Image
   dmRsf_oc<float> Img0_dm("inputImg",dm::IN,"Initial RTM Image");
   Img0_dm.init(NULL,true);
   
   WEMVAparam.Img0_hx.alloc(Img0_dm.domain());
   Img0_dm.seq_grab(&WEMVAparam.Img0_hx);
   Img0_dm.unInit();
   
   pWEMVAobj->calcObjFunc();

   //calculate the bacjProject data
   pWEMVAobj->calcBackProjData();
   //Image Residue
   WEMVAparam.backProjImg_hx.dump("backProjImage");
   
   //calculate gradient this time
   //velAnaParams.velGradCube.alloc(params.velModelShape);
   //!!Temporary
   velAnaParams.velGradCube.alloc(params.velModelShape);
   pWEMVAobj->calcSlowGradient(velAnaParams.velGradCube);
   
   //Dump the src and recv wavefield.
   params.pSrcSnapshot->unInit();
   params.pSrcSnapshot2->unInit();
   params.pRecvSnapshot->unInit();
   params.pRecvSnapshot2->unInit();
   
   //Save the gradient to see
   velAnaParams.velGradCube.dump("grad");
   
   //dump RTM image by each shot
   params.pDeltaSs->dump("grads");
   params.pFinalDeltaS->dump("out");
	
   //clean up
   velAnaParams.pDataObs->unInit();
   SAFE_DELETE(pWEMVAobj);
  
   //We use auto_ptr now, no need to delete those ptrs.
   return 0;
}
