#include "stdafx.h"
#include "dmRsf.hpp"

#include "srcWav.h"
#include "velFD2d.h"
#include "point3d.hpp"

#include "modelingFD2d_op.h"
#include "modelingFDgpu.h"
#include "waveOpsOneway.h"
#include "SR_manager.h"
#include "paddingFD.h"
#include "paddingFD_damping.h"
#include "boundaryFD.h"

#include "param_entry.h"
#include "myOMP.h"

#include "FWI_FD.h"
#include "WEMVA.h"


#include "stackTraceC.h"
#include "FWI_objFunc.h"
#include "solverNonlinearSearch.h"
#include "commonOps.h"

#include "FWI_FD.h"
#include "FWI_FD_beamFocus.h"
#include "FWI_FD_DSO.h"
#include "WEMVA.h"
///Read parameters
#include "nonLinSolver.param.h"
#include "wavePropagate.param.h"
#include "velAna.param.h"
#include "off2ang3d.param.h"

#include "misc.h"

/**
 * For beam focusing, Wave Equation Tomography, Just do the 
 * non-linear line search, in the 3D case, where everything is out-of-core, i.e. stored in disk file not in memory.
 * in reflection case.
 */

using namespace Yang;
using namespace std;
   
//Doing beam wave-equation tomography for reflection case.
//Use cuda Wave modeling engine
/// The input of the program
/// initial velocity model vel
/// observed data data_obs along with its geometry
int main (int argc, char *argv[])
{
   initCallStackTrace();

   init_omp(100);

   //Read params, params need contain geometry info that is exact the same as the input data. In
   //synthetic this is easy to do.
   auto_ptr<sep_params> autoPars(new sep_params(argc,argv));
   sep_params* pars = autoPars.get();

   auto_ptr<vel2d> velObj(new vel2d(csz_tag_vel,dm::IN,"vel"));
   auto_ptr<srcWav> waveletObj(new srcWav1d(csz_tag_wavelet,dm::IN,"wavelet"));

   //Read in wavelet and initial velocity, input 2nd derivative of Gaussian wavelet 
   waveletObj->loadWavelet();
   int ndim_3D = 3;
   velObj->loadVel(ndim_3D); //load velocity model, this can be in memory, read in as a 3D cube.
   velObj->info("Initial velocity model");
   
   waveOps *modelingObj = NULL;
   //Read in the parameters for the modeling Operator
   waveOpParams params;
   velAnaParam velAnaParams;
   WEMVA_Param WEMVAparam;

   bool bUseOneWay = pars->get_bool("bUseOneway",false);
   params.bUseOnewayPropagator = bUseOneWay;
   //might use one-way
   if (bUseOneWay) {
     modelingObj = new waveOpsOneway();
     //Read in the oneway file tags
     string sz_csou = "csou";
     VERIFY(tagExist(sz_csou));
     //params.oper = pars->get_string("OneWayOper","ssfpi");
     //params.pdip = pars->get_string("pdip","45");
     //params.nws  = pars->get_int("nws");
     params.crecFileName = pars->get_string(csz_tag_dataObs); //!assume we use the same data every time
     params.csouFileName = pars->get_string("csou");
     //params.nxPadOneway  = pars->get_int(csz_npad);
     //params.nyPadOneway  = params.nxPadOneway;
     velAnaParams.dObsFilename = pars->get_string(csz_tag_dataObs); //For Imaging, observed data tag
     
     params.sz_additional_mig_par = pars->get_string("add_migpar_1w");
     PRV(params.sz_additional_mig_par);
   } else {
     VERIFY(0); //not ready for 3D case.
     //Create the modeling Obj for modeling and also RTM(for calculating the gradient)
     modelingFDgpu2d_op* modelingObj_ptr = new modelingFDgpu2d_op();
     int gpuID = pars->get_int(csz_gpuID,1);
     modelingObj_ptr->setGPU_ID(gpuID);
     //Create the padding and boundary processor
     SR_managerFD2d*  SR_managerObj = new SR_managerFD2d(waveletObj.get());
     //auto_ptr<paddingFD2dDamping> paddingDampingObj ( new paddingFD2dDamping(NULL,dampingExtent));
     paddingFD2d*     paddingObj  = new paddingFD2d(NULL);
     //boundary condition
     boundaryFD2d*    boundaryObj = new boundaryFD2d();
     //assemble
     modelingObj_ptr->assemble(SR_managerObj, boundaryObj,paddingObj);
     modelingObj = modelingObj_ptr;
   }

  params.pVel = velObj.get();
  params.pWavelet = waveletObj.get();
  
  params.modelGridShape = velObj->domain();
  params.velModelShape = velObj->domain();
  params.ndim = params.velModelShape.get_ndim();
  
  read_twoway_prop_param(&params,pars);
  
  //calc time parameters
  float maxFreq,vmax,vmin;
  //Read in the peak/center freq of the seismic wavelet
  maxFreq = waveletObj->get_float("peakFreq");
  waveletObj->setPeakFreq(maxFreq);
  vmax = velObj->maxval();
  vmin = velObj->minval();
  calcTimeParamsFor2wayModeling(&(params.timeParam),maxFreq,vmax,vmin,params.velModelShape);
   
  read_migration_param(&params,pars);
  //The final image shape can be different from the velocity model, because the velocity model would be padded.

  hypercube imgOffShape;
  get_prestack_img_cube_shape(imgOffShape, params.mOffImgParams, params.finalImageShape); 
  params.finalOffImageShape = imgOffShape;

  //Here, need to read in the mask as the random sparse sampling of the migrated gather.
  cubeData<float>* pImageMask = NULL;
  if (tagExist(csz_tag_img_rand_samp_mask)) {
    pImageMask = new cubeData<float>(csz_tag_img_rand_samp_mask,dm::IN,"image random sample mask");
    LOG(Enable Random Sparse Sampling of the migrated Image.);
    pImageMask->init(NULL); //read the mask in.
  }
  params.pImageMask = pImageMask;

  if (!bUseOneWay) { //if not use oneway, then read the recorded data
    VERIFY(0); //not finished yet.
    //load the observed data
    velAnaParams.pDataObs = new cubeData<float>(csz_tag_dataObs,dm::IN,"data observed");
    velAnaParams.pDataObs->init(NULL);
    velAnaParams.pDataObs ->domain().info("Data obs info");
    velAnaParams.dObsCube.alloc(velAnaParams.pDataObs->domain());
    velAnaParams.pDataObs->seq_grabAll(&(velAnaParams.dObsCube));
  } //otherwise it is one-way, let Yaxun's program to take care of it.
  
  WEMVAparam.off_img_param  = params.mOffImgParams;
  WEMVAparam.imageShape     = params.finalImageShape;
  WEMVAparam.b3D            = params.b3D;
  read_velAna_param(&velAnaParams,pars);

  WEMVA_FD* pWEMVAobj = NULL;
  waveOps* pModeling = modelingObj;
  //read_WEMVA_param(&WEMVAparam,pars) is included in the create_WEMVA_obj function
  pWEMVAobj = create_WEMVA_obj(&velAnaParams, &params, pModeling, &WEMVAparam,pars);

  //set off2ang3d params
  WEMVAparam.par_o2a3d = new param_off2ang3d(); WEMVAparam.par_a2o3d = new param_off2ang3d();
  hypercube inshapeOff,inshapeAng,shapezxy;
  WEMVAparam.imageShape.info("WEMVAparam.imageShape");
  shapezxy = WEMVAparam.imageShape;
  get_prestack_img_cube_shape(inshapeOff, WEMVAparam.off_img_param, shapezxy); 
  inshapeOff.info("inshapeOff");
  //convert z,x,y,hx,hy to hx,hy,kz,kmx,kmy
  vecAxis axes = inshapeOff.get_axes(); inshapeOff.set_axis(1,axes[3]); inshapeOff.set_axis(2,axes[4]);
  axis ax_kz,ax_kx,ax_ky;

  int nkz,nkx,nky;
  nkz = pars->get_int(csz_nz_pad);
  nkx = pars->get_int(csz_nx_pad);
  nky = pars->get_int(csz_ny_pad);
  
  get_axis_inv(ax_kz, nkz, axes[0]);  inshapeOff.set_axis(3,ax_kz);
  get_axis_inv(ax_kx, nkx, axes[1]);  inshapeOff.set_axis(4,ax_kx);
  get_axis_inv(ax_ky, nky, axes[2]);  inshapeOff.set_axis(5,ax_ky);
  
  read_param_off2ang3d(WEMVAparam.par_o2a3d, pars, inshapeOff, shapezxy,true); 

  //now we know the dimension of the angle domain shape in par_a2o3d;
  inshapeAng = WEMVAparam.par_o2a3d->shapeAng;
  read_param_off2ang3d(WEMVAparam.par_a2o3d, pars, inshapeAng, shapezxy,false); 
  
  WEMVAparam.par_a2o3d->shapeOff.info("par_a2o3d->shapeOff");

  solverParam solvPar;
  read_nonLinSolver_par(&solvPar,pars);

  //Create the solver
  gVecComponent initSlowModelVec, finalModelVec;
  
  params.pVel->copyTo(initSlowModelVec);
  initSlowModelVec.reciprocal();
  params.pVel->allocMemoryTo(finalModelVec);
  solverNonlinearSearch nonLinSolver(pWEMVAobj, &initSlowModelVec,&solvPar);
  
  stdVecFloat finalErrorVec; int finalIter;
  //no need to generate gradient mask, beacause the deltaS is generated after being Windowed by the apertures of each shot respectively.
  vector<gVecComponent*> modelsVec, gradsVec, datsVec; //The vectors are all in-core, this is not OK for 3D case.
  bool bIsDataInCore = !params.b3D;
  
  typedef enum _enumsolver_act_type {
    solver_inv,solver_calcData,solver_grad,solver_objval
  } solver_act_type;
  
  solver_act_type action = solver_inv;
  if (action == solver_inv) {
    nonLinSolver.solve( finalModelVec, finalErrorVec, finalIter, 
                       &modelsVec, &gradsVec, NULL, bIsDataInCore);
    
    LOG(m4);
    //Save result
    finalModelVec.dump("out");
    finalModelVec.unInit();
    
    //dump error
    int iSize = finalErrorVec.size(); 
    PRV(iSize);
    cubeData<float> finalErrCube(iSize);
    myMemcpy(finalErrCube.pval,&finalErrorVec[0],iSize);
    finalErrCube.dump("error");
    LOG(m5);
    //dump models and grads
    PRV(modelsVec.size()); PRV(gradsVec.size());
    dumpVecModels(modelsVec,"mods");
    dumpVecModels(gradsVec,"grads");
    LOG(deleteVecModels);
    deleteVecPointers(modelsVec);
    deleteVecPointers(gradsVec);
    LOG(dump final image);
  
    //Dump the final data
    if (bIsDataInCore) {
      const cubeData<float>& finalImage = pWEMVAobj->getCalcData();
      finalImage.dump("fImage");
    } else { //just specify the file name
      string fimg_fn = pars->get_string("fImage");
      char cmd[1024];
      sprintf(cmd,"Mv %s %s", pWEMVAobj->getCalcData_fn().c_str(), fimg_fn.c_str());
      VERIFY(system(cmd)==0);
    }
  } else if (action == solver_calcData) {
    //pWEMVAobj->calcData();
    pWEMVAobj->postCalcData();
  } else if (action == solver_grad) {
    pWEMVAobj->m_bAngleGatherCalculated = true;
    gVecComponent   cube_gradient;
    cube_gradient.alloc(velObj->domain());
    pWEMVAobj->calcSearchDir(cube_gradient);
    //pWEMVAobj->calcSlowGradient(cube_gradient);
    cube_gradient.dump("grads");
  } else if (action == solver_objval) {
    //assume the image has been calculated. I just call the obj function.
    pWEMVAobj->m_bAngleGatherCalculated = true;
    pWEMVAobj->calcObjFunc();
  }
  
  LOG("final clean up");
  //clean up
  if (!bUseOneWay) {
    velAnaParams.pDataObs->unInit();
  }
  SAFE_DELETE(pWEMVAobj);
  return 0;
}

