#include "stdafx.h"
#include "sep_params.h"
#include "srcWav.h"
#include "velFD2d.h"
#include "point3d.hpp"
#include "../param_entry.h"

#include "boundaryFD.h"
#include "paddingFD.h"
#include "waveOps.h"
#include "modelingFDgpu.h"
#include "SR_manager.h"
#include "stackTraceC.h"

using namespace Yang;
using namespace std;

//CUDA modeling 2d case, can do nonlinear modeling or born modeling (Adjoint RTM)
int main (int argc, char *argv[])
{
   initCallStackTrace();
   //Read params, params need contain geometry info that is exact the same as the input data. In
   //synthetic this is easy to do.
   auto_ptr<sep_params> autoPars(new sep_params(argc,argv));
   sep_params* pars = autoPars.get();

   auto_ptr<vel2d> velObj(new vel2d(csz_tag_vel,dm::IN,"velocity"));

   auto_ptr<srcWav> waveletObj(new srcWav1d(csz_tag_wavelet,dm::IN,"wavelet"));

   //Read in wavelet and initial velocity, usually Ricker
   //derivate to create the ricker wavelet.
   waveletObj->loadWavelet();
   velObj->loadVel();

   velObj->domain().info("vel");
   
   //Create the modeling Obj for modeling and also RTM(for calculating the gradient)
   auto_ptr<modelingFDgpu2d_op> modelingObj(new modelingFDgpu2d_op());

   modelingObj->m_dampingCoef = 0.0f;
   
   auto_ptr<SR_managerFD2d>  SR_managerObj(new SR_managerFD2d(waveletObj.get()));
   
   //Create the padding and boundary processor, dummy objects
   auto_ptr<paddingFD2d>  paddingObj(new paddingFD2d(NULL)); //dummy object
   //boundary condition dummy
   auto_ptr<boundaryFD2d> boundaryObj(new boundaryFD2d());

   //assemble
   modelingObj->assemble(SR_managerObj.get(), boundaryObj.get(),paddingObj.get());

   //read and set the parameters
   waveOpParams params;
   
   float maxFreq,vmax,vmin;
   params.pVel = velObj.get();
   //Read in the peak/center freq of the seismic wavelet
   maxFreq = waveletObj->get_float("peakFreq");
   waveletObj->setPeakFreq(maxFreq);
   
   vmax = velObj->maxval();
   vmin = velObj->minval();
   params.velModelShape = velObj->domain();
   params.modelGridShape = params.velModelShape;
   //also get the time param from the params
   params.bFreeSurfaceOnTop = pars->get_bool(csz_freeSurface);
   params.stencilWidth = pars->get_int(csz_stenWidth);
   VERIFY(params.stencilWidth == 4);
   
   params.bUse4thOrder= pars->get_bool(csz_use4thOrderInTime);
   
   params.ndim = params.velModelShape.get_ndim();
   params.offx = params.stencilWidth;params.offz = params.stencilWidth;
   
   params.m_nShots 	= pars->get_int(csz_nShots);
   params.mShot1_ix = pars->get_int(csz_shot1_ix);
   params.mShot1_iz = pars->get_int(csz_shot1_iz);
   params.mShotIncX = pars->get_int(csz_shotIncX);

   params.mTrace1ix = pars->get_int(csz_trace1_ix);
   params.mTrace1iz = pars->get_int(csz_trace1_iz);
   params.mTraceIncX = pars->get_int(csz_traceIncX);
   params.m_nTraces = pars->get_int(csz_nTraces);
   
   //Recording Snapshot params
   params.f_shot_snap = pars->get_int(csz_snap_f,0);
   params.j_shot_snap = pars->get_int(csz_snap_j);
   params.n_shot_snap = pars->get_int(csz_snap_n,0);
   checkSnapRecordParams(&params);
   
   params.m_nPad = pars->get_int(csz_npad);

   bool bOffset = pars->get_bool(
     csz_do_offset,false);
   params.mOffImgParams.bDoOffsetGather = bOffset;
   params.bTake2ndTimeDerivForRTM = pars->get_bool(csz_take2ndTimeDerivRTM,false);
   //Read in Absorption type
   int itype = pars->get_int(csz_bdType);
   switch (itype)
   {
   case 1:
	  params.m_bdType = abs_ord1;
	  break;
   case 2:
	  params.m_bdType = abs_ord2;
	  break;
   case 3:
	  params.m_bdType = abs_Dave;
	  break;
   case 4:
	  params.m_bdType = abs_ord1par; break;
   case 0:
	  params.m_bdType = abs_null;
      break;
      default:
         seperr("!unknown boundary condition type!\n");
   }
   //calc time parameters
   params.timeParam.totalTime = pars->get_float(csz_totalTime);
   params.timeParam.dtComp = pars->get_float(csz_dtComp);
   params.timeParam.dtRecord = pars->get_float(csz_dtRec);
   params.timeParam.dtSnap = pars->get_float(csz_dtSnap);
   params.timeParam.dtImaging = pars->get_float(csz_dtImaging,params.timeParam.dtRecord);
   
   calcTimeParamsFor2wayModeling(&(params.timeParam),maxFreq,vmax,vmin,params.modelGridShape);
   
   params.pSrcSnapshot 	= new cubeData<float>(csz_tag_srcSnapshot,dm::OUT,"src snapshot");   

   bool bLinearModeling = pars->get_bool(csz_bLinearModeling,false);
   PRV(bLinearModeling);
   
   cubeData<float> refl;
   if (bLinearModeling) {
	  params.pSrcSnapshot2 = new cubeData<float>(csz_tag_srcSnapshot2,dm::OUT,"src snapshot2");
		 dmRsf_oc<float> rsfRefl(csz_tag_reflModel,dm::IN,"Refl Model");
     rsfRefl.init(NULL,true);
		 refl.alloc(rsfRefl.domain());
     rsfRefl.seq_grabAll(&refl);
     rsfRefl.unInit();
   }
   
   //Init the output   
   cubeData<float> rsfSeisRec(csz_tag_seisRec,dm::OUT,"seis-record");
   //We need to figure out the size of recorded data ourself
   SR_managerObj->init(&params);
   hypercube seisRecShape = SR_managerObj->calcModelingDataSize(&params);
   rsfSeisRec.init(&seisRecShape);
   
   //Init the operator and do the modeling
   if (bLinearModeling) {
	  modelingObj->initOperatorEx(&params, acBornModeling);
	  modelingObj->bornModeling_op(&refl,&rsfSeisRec);
    rsfSeisRec.info("rsfSeisRec");
   } else {
	  modelingObj->initOperatorEx(&params, acModeling);
	  modelingObj->modeling_op(NULL,&rsfSeisRec);
   }

   //then save the necessary wavefields
   //After modeling, save the snapshot and seisrecord
   rsfSeisRec.unInit();
   params.pSrcSnapshot->unInit();
   if (bLinearModeling) {
     params.pSrcSnapshot2->unInit();
   }
   
   //pWav->getWavelet()->dump("dump");
   return 0;
}
