#include "stdafx.h"
#include "sep_params.h"
#include "srcWav.h"
#include "velFD2d.h"
#include "point3d.hpp"
#include "../param_entry.h"

#include "boundaryFD.h"
#include "paddingFD.h"
#include "waveOps.h"
#include "waveOpsOneway.h"
#include "modelingFDgpu.h"
#include "SR_manager.h"
#include "stackTraceC.h"

#include "wavePropagate.param.h"

using namespace Yang;
using namespace std;

//CUDA modeling 2d case, do RTM, can do offset Image gather
int main (int argc, char *argv[])
{
   initCallStackTrace();
   //Read params, params need contain geometry info that is exact the same as the input data. In
   //synthetic this is easy to do.
   auto_ptr<sep_params> autoPars(new sep_params(argc,argv));
   sep_params* pars = autoPars.get();

   auto_ptr<vel2d> velObj(new vel2d(csz_tag_vel,dm::IN,"velocity"));
   auto_ptr<srcWav> waveletObj(new srcWav1d(csz_tag_wavelet,dm::IN,"wavelet"));

   //Read in wavelet and initial velocity, just input a Guassian wavelet, beacause it will take 2nd 
   //derivate to create the ricker wavelet.
   waveletObj->loadWavelet();
   velObj->loadVel();

   velObj->domain().info("vel");
   
   waveOps *modelingObj = NULL;
   bool bUseOneWay = pars->get_bool("bUseOneway",false);
   //Read in the parameters for the modeling Operator
   waveOpParams params;   
   //might use one-way
   if (bUseOneWay) {
     modelingObj = new waveOpsOneway();
     //Read in the oneway params
     params.sz_additional_mig_par = pars->get_string("add_migpar_1w");
     params.crecFileName = pars->get_string(csz_tag_dataObs); //!assume we use the same data every time
     string sz_csou = "csou";
     VERIFY(tagExist(sz_csou));
     params.csouFileName = pars->get_string("csou");
   } else {
     //Create the modeling Obj for modeling and also RTM(for calculating the gradient)
     modelingFDgpu2d_op* modelingObj_ptr = new modelingFDgpu2d_op();
  
     int gpuID = pars->get_int(csz_gpuID,1);
     modelingObj_ptr->setGPU_ID(gpuID);
  
     //Create the padding and boundary processor
     SR_managerFD2d*  SR_managerObj = new SR_managerFD2d(waveletObj.get());
     
     //chain two padding method together
     //float dampingExtent = 0.5f;
     //auto_ptr<paddingFD2dDamping> paddingDampingObj ( new paddingFD2dDamping(NULL,dampingExtent));
     paddingFD2d*     paddingObj  = new paddingFD2d(NULL);
     //cosDampingFunc dampFunc;
     //paddingDampingObj->m_pDampingFunc = &dampFunc;
      
     //boundary condition
     boundaryFD2d*    boundaryObj = new boundaryFD2d();
  
     //assemble
     modelingObj_ptr->assemble(SR_managerObj, boundaryObj,paddingObj);
     modelingObj = modelingObj_ptr;
   }
   
   params.pVel = velObj.get();
   params.pWavelet = waveletObj.get();
   params.velModelShape   = velObj->domain();
   params.modelGridShape  = velObj->domain();
   params.finalImageShape = params.velModelShape;
   params.ndim = params.velModelShape.get_ndim();
   
   read_migration_param(&params,pars);
   read_twoway_prop_param(&params,pars);
   float maxFreq,vmax,vmin;
   //Read in the peak/center freq of the seismic wavelet
   maxFreq = waveletObj->get_float("peakFreq");
   waveletObj->setPeakFreq(maxFreq);
   
   vmax = velObj->maxval();
   vmin = velObj->minval();
   
   calcTimeParamsFor2wayModeling(&(params.timeParam),maxFreq,vmax,vmin,params.velModelShape);
   
   
   //Load the off Image parameters.
   bool bDoOffsetGather = params.mOffImgParams.bDoOffsetGather;
   
   PRV(params.bTake2ndTimeDerivForRTM);
   
   //Here, need to read in the mask as the random sparse sampling of the migrated gather.
   cubeData<float>* pImageMask = NULL;
   if (tagExist(csz_tag_img_rand_samp_mask)) {
      pImageMask = new cubeData<float>(csz_tag_img_rand_samp_mask,dm::IN,"image random sample mask");
      LOG(Enable Random Sparse Sampling of the migrated Image.);
      pImageMask->init(NULL); //read the mask in.
   }
   params.pImageMask = pImageMask;
   
   //params.pImages = new cubeData<float>(csz_tag_images,dm::OUT,"RTM_Images"); //per shot image. //not used.
   params.pImages = NULL;
   cubeData<float> finalOffImg(csz_tag_final_off_image,dm::OUT,"");
   cubeData<float> finalImg(csz_tag_final_image,dm::OUT,"");
   hypercube imageShape;
   if (bDoOffsetGather) {
     hypercube shapeOffImg;
     get_prestack_img_cube_shape(shapeOffImg,params.mOffImgParams,
                  params.finalImageShape, params.timeParam.dtComp);
     finalOffImg.init(&shapeOffImg);
     imageShape = shapeOffImg;
   } else {
     finalImg.init(&params.finalImageShape);
     imageShape = finalImg.domain();
   }
   if (pImageMask) {VERIFY(pImageMask->domain().sameSize(imageShape));}

   params.pSrcSnapshot 	= new cubeData<float>(csz_tag_srcSnapshot,dm::OUT,"src snapshot");
   params.pRecvSnapshot	= new cubeData<float>(csz_tag_recvSnapshot,dm::OUT,"recv snapshot");
   params.pRsfSeisRec	= new cubeData<float>(csz_tag_seisRec,dm::IN,"seis-record");
   
   
   //Init the operator and do the RTM
   modelingObj->initOperatorEx(&params, acMigration);
   if (!bUseOneWay) { //Use Time FD, two-way WE.
     //Get the size of seismic record
     params.pRsfSeisRec->init(NULL);
     //The geometry must agree
     VERIFY(params.m_nShots == params.pRsfSeisRec->domain().get_n(3));
   
     if (bDoOffsetGather) {
       modelingObj->migration_op(params.pRsfSeisRec,&finalOffImg);
       finalOffImg.unInit();
     } else {
       modelingObj->migration_op(params.pRsfSeisRec,&finalImg);
       finalImg.unInit();
     }
   
     //then save the necessary wavefields
     //After modeling, save the snapshot and seisrecord
     params.pSrcSnapshot->unInit();
     params.pRsfSeisRec->unInit();
     params.pRecvSnapshot->unInit();
   } else { //Use one way, can only use file tags string as input, because input
     //is COMPLEX number, cannot be loaded by dmRsf<float>?
     if (bDoOffsetGather) {
       modelingObj->migration_op(params.crecFileName,&finalOffImg);
       finalOffImg.unInit();
     } else {
       modelingObj->migration_op(params.crecFileName,&finalImg);
       finalImg.unInit();
     }
   }

   srcWav* pWav = waveletObj.release();
   //pWav->getWavelet()->dump("dump");
   delete pWav;
}
