#include "stdafx.h"
#include "dmRsf.hpp"

#include "srcWav.h"
#include "velFD2d.h"
#include "point3d.hpp"

#include "modelingFD2d_op.h"
#include "modelingFDgpu.h"
#include "waveOpsOneway.h"
#include "SR_manager.h"
#include "paddingFD.h"
#include "paddingFD_damping.h"
#include "boundaryFD.h"

#include "param_entry.h"
#include "myOMP.h"

#include "WEMVA.h"

#include "stackTraceC.h"



#include "wavePropagate.param.h"

/**
 *  Wave Equation Tomography, Adj opeator
 * calculation, CUDA enabled. Reflection case
 */


using namespace Yang;
using namespace std;
   
//Doing beam wave-equation tomography for reflection case
/// The input of the program: initial velocity model vel
/// observed data data_obs along with its geometry
/// Input image perturbation 
/// Output: slowness/velocity perturbation

int main (int argc, char *argv[])
{
   initCallStackTrace();
   //Read params, params need contain geometry info that is exact the same as the input data. In
   //synthetic this is easy to do.
   auto_ptr<sep_params> autoPars(new sep_params(argc,argv));
   sep_params* pars = autoPars.get();

   auto_ptr<vel2d> velObj(new vel2d(csz_tag_vel,dm::IN,"vel"));
   auto_ptr<srcWav> waveletObj(new srcWav1d(csz_tag_wavelet,dm::IN,"wavelet"));

   //Read in wavelet and initial velocity, input 2nd derivative of Gaussian wavelet 
   waveletObj->loadWavelet();
   velObj->loadVel();
   
   waveOps *modelingObj = NULL;
   bool bUseOneWay = pars->get_bool("bUseOneway",false);
   //Read in the parameters for the modeling Operator
   waveOpParams params;
   velAnaParam velAnaParams;
   //WEMVA_Param WEMVAparam;
   params.bUseOnewayPropagator = bUseOneWay;

   //might use one-way
   if (bUseOneWay) {
     modelingObj = new waveOpsOneway();
     //Read in the oneway file tags
     string sz_csou = "csou";
     VERIFY(tagExist(sz_csou));
    
     params.crecFileName = pars->get_string(csz_tag_dataObs); //!assume we use the same data every time
     params.csouFileName = pars->get_string("csou");
     velAnaParams.dObsFilename = pars->get_string(csz_tag_dataObs); //For Imaging, observed data tag
     params.sz_additional_mig_par = pars->get_string("add_migpar_1w");
   } else {
     //Create the modeling Obj for modeling and also RTM(for calculating the gradient)
     modelingFDgpu2d_op* modelingObj_ptr = new modelingFDgpu2d_op();
  
     int gpuID = pars->get_int(csz_gpuID,1);
     modelingObj_ptr->setGPU_ID(gpuID);
  
     //Create the padding and boundary processor
     SR_managerFD2d*  SR_managerObj = new SR_managerFD2d(waveletObj.get());
     
     //chain two padding method together
     //float dampingExtent = 0.5f;
     //auto_ptr<paddingFD2dDamping> paddingDampingObj ( new paddingFD2dDamping(NULL,dampingExtent));
     paddingFD2d*     paddingObj  = new paddingFD2d(NULL);
     //cosDampingFunc dampFunc;
     //paddingDampingObj->m_pDampingFunc = &dampFunc;
      
     //boundary condition
     boundaryFD2d*    boundaryObj = new boundaryFD2d();
  
     //assemble
     modelingObj_ptr->assemble(SR_managerObj, boundaryObj,paddingObj);
     modelingObj = modelingObj_ptr;
   }

   
   params.pVel = velObj.get();
   params.pWavelet = waveletObj.get();
   params.velModelShape   = velObj->domain();
   params.modelGridShape  = velObj->domain();
   params.finalImageShape = params.velModelShape;
   params.ndim = params.velModelShape.get_ndim();

   read_twoway_prop_param(&params,pars);

   float maxFreq,vmax,vmin;
   //Read in the peak/center freq of the seismic wavelet
   maxFreq = waveletObj->get_float("peakFreq");
   waveletObj->setPeakFreq(maxFreq);
   vmax = velObj->maxval();
   vmin = velObj->minval();
   
   //calc time parameters
   calcTimeParamsFor2wayModeling(&(params.timeParam),maxFreq,vmax,vmin,params.modelGridShape);
  
   read_migration_param(&params,pars);
   //Load the off Image parameters.
   bool bDoOffsetGather = params.mOffImgParams.bDoOffsetGather;

   //WEMVAparam.off_img_param = params.mOffImgParams;
   //WEMVAparam.imageShape    = params.finalImageShape;

   if (!bUseOneWay) {
     //load the observed data
     velAnaParams.pDataObs = new cubeData<float>(csz_tag_dataObs,dm::IN,"data observed");
     //Get the size of seismic record
     velAnaParams.pDataObs->init(NULL,true);
     velAnaParams.pDataObs ->domain().info("Data obs info");
     //Load the observed data
     velAnaParams.dObsCube.alloc(velAnaParams.pDataObs->domain());
     velAnaParams.pDataObs->seq_grabAll(&(velAnaParams.dObsCube));
     
     //Set the data for Tom operator
     params.pRsfSeisRec = velAnaParams.pDataObs;
   }

   params.pDeltaSs = new cubeData<float>("",dm::OUT,"");
   
   //Init the output
   cubeData<float> outDeltaS("",dm::OUT,"deltaS");
   outDeltaS.alloc(params.velModelShape);

   cubeData<float>* pDeltaIcube = NULL;
   
   //Input image perturbation
   if (bDoOffsetGather) {
     //!! Memory Leak here
     pDeltaIcube = new cubeData<float> ("deltaIh",dm::IN,"Image perturb");
   } else {
     pDeltaIcube = new cubeData<float> ("deltaI",dm::IN,"Image perturb");
   } 
   
  pDeltaIcube->init(NULL,true);
  //Init and call the operator
  modelingObj->initOperatorEx(&params,acTomAdj);
  pDeltaIcube->domain().info("input image perturbation info");
  modelingObj->TomAdj_op(pDeltaIcube,&outDeltaS);
   
  //After that the result are among those images
  outDeltaS.dump("out");
  params.pDeltaSs->dump("deltaSs");
  params.pDeltaSs->unInit();

  //We use auto_ptr now, no need to delete those ptrs.
  return 0;
}
