#ifndef GPU_WFLD_SEP_H
#define GPU_WFLD_SEP_H

struct FFT1dWfldSeparate {
  cufftHandle planR2C,planC2C;
  int nz,nx; //The actually size of the internal array. nz might be padded to be 2^x, therefore it would be different from the input
  cufftComplex *wfld_src,*wfld_recv; //For the R2C transform. half size
  cufftComplex *wfldNew_src,*wfldNew_recv; //For the C2R transform, full size nx*nz*2*floats
};

inline void FFT1dWfldSeparate_init(FFT1dWfldSeparate* p) {
  p->planR2C = p->planC2C = 0;
  p->nx = p->nz = 0;
  p->wfld_src = p->wfld_recv = NULL;
  p->wfldNew_recv = p->wfldNew_recv = NULL;
}

inline void FFT1dWfldSeparate_deAlloc(FFT1dWfldSeparate* p) {
  if (p->wfld_src) { //Lets assume this case the fft plan need to be destroyed
    cufftDestroy(p->planR2C);    cufftDestroy(p->planC2C);
  }
  cuda_call(cudaFree(p->wfld_src));
  cuda_call(cudaFree(p->wfld_recv));
  cuda_call(cudaFree(p->wfldNew_src));
  cuda_call(cudaFree(p->wfldNew_recv));
  FFT1dWfldSeparate_init(p);
}

inline void FFT1dWfldSeparate_alloc(FFT1dWfldSeparate* p,int nz_in, int nx_in){
  bool bReAlloc = false;
  //round nz_in up to the mulpilies of 128
  int nz_rnd = (nz_in-1)/128+1;
  nz_rnd *= 128;
  if (nx_in!=p->nx || nz_rnd!=p->nz) {
    bReAlloc = true;
  } else {
    VERIFYC(p->wfld_src != NULL);
    VERIFYC(p->wfld_recv!= NULL);
  }
  p->nx = nx_in;
  p->nz = nz_rnd;
  
  if (bReAlloc) {
    FFT1dWfldSeparate_deAlloc(p); //free memory first
    //Alloc
    cudaMalloc((void**)&(p->wfld_src),     sizeof(cufftComplex)*(p->nz/2+1)*(p->nx));
    cudaMalloc((void**)&(p->wfld_recv),    sizeof(cufftComplex)*(p->nz/2+1)*(p->nx));
    cudaMalloc((void**)&(p->wfldNew_src),  sizeof(cufftComplex)*(p->nz)*(p->nx));
    cudaMalloc((void**)&(p->wfldNew_recv), sizeof(cufftComplex)*(p->nz)*(p->nx));
    
    //Create both plans
    cufftPlan1d(&(p->planR2C),p->nz,CUFFT_R2C, p->nx);
    cufftPlan1d(&(p->planC2C),p->nz,CUFFT_C2C, p->nx);
  }
}

inline void FFT1dWfldSeparate_calcCmplxWfld(FFT1dWfldSeparate* pSep, const float* wfld_src_in, const float* wfld_recv_in,
                                     cudacomplex* wfld_src_out, cudacomplex* wfld_recv_out, int nz_in) {
  int nz = pSep->nz;
  int nx = pSep->nx;
  cufftComplex* pSrc    = pSep->wfld_src;
  cufftComplex* pRecv   = pSep->wfld_recv;
  cufftComplex* pNewSrc = pSep->wfldNew_src;
  cufftComplex* pNewRecv= pSep->wfldNew_recv;
  
  //Set zero wfld
  cuda_call(cudaMemset(pSrc,0, sizeof(cufftComplex)*(nz/2+1)*nx));
  cuda_call(cudaMemset(pRecv,0,sizeof(cufftComplex)*(nz/2+1)*nx));
  
  //copy from the input
  myCudaMemcpy2d(pSrc,(nz/2+1)*sizeof(cufftComplex)/sizeof(float),  wfld_src_in,nz_in,
                 nz_in,nx,cudaMemcpyDeviceToDevice,0,0,0,0,sizeof(float));
  myCudaMemcpy2d(pRecv,(nz/2+1)*sizeof(cufftComplex)/sizeof(float), wfld_recv_in,nz_in,
                 nz_in,nx,cudaMemcpyDeviceToDevice,0,0,0,0,sizeof(float));
  
  //Use the CUFFT plan to transform the signal
  cufftExecR2C(pSep->planR2C,(cufftReal*)pSrc,pSrc);
  cufftExecR2C(pSep->planR2C,(cufftReal*)pRecv,pRecv);
  
  //Now we need to cp the wfld to the first half of new wfld, which equals to zeroing the upper half spectrum.
  //First set zero
  cuda_call(cudaMemset(pNewSrc,0,sizeof(cufftComplex)*nz*nx));
  cuda_call(cudaMemset(pNewRecv,0,sizeof(cufftComplex)*nz*nx));

  myCudaMemcpy2d(pNewSrc, nz,   pSrc, (nz/2+1), nz/2, nx, cudaMemcpyDeviceToDevice,
                 0,0,0,0, sizeof(cufftComplex));
  myCudaMemcpy2d(pNewRecv, nz,  pRecv, (nz/2+1), nz/2, nx, cudaMemcpyDeviceToDevice,
                 0,0,0,0, sizeof(cufftComplex));

  //Transform back to get the complx wfld.
  cufftExecC2C(pSep->planC2C,pNewSrc, pNewSrc,  CUFFT_INVERSE);
  cufftExecC2C(pSep->planC2C,pNewRecv,pNewRecv, CUFFT_INVERSE);

  //Copy the tranformed wfld data to output,because the pitch might be different
  assert(sizeof(cufftComplex) == sizeof(float2));
  myCudaMemcpy2d(wfld_src_out,nz_in,pNewSrc,nz,nz_in,nx,cudaMemcpyDeviceToDevice,
                 0,0,0,0,sizeof(cufftComplex));
  myCudaMemcpy2d(wfld_recv_out,nz_in,pNewRecv,nz,nz_in,nx,cudaMemcpyDeviceToDevice,
                 0,0,0,0,sizeof(cufftComplex));
  return;
}


#endif // 

