#include "stdafxc.h"
#include <cutil.h>
#include <cuda.h>
#include "cudaErrors.cu.h"

#include "cudaModeling2d.h"

#include "gpu_cmplx.h"
#include "gpu_util.cu.h"
#include "gpu_wfldSeparate.cu.h"

__constant__ int nt_gpu;
__constant__ float dt_gpu;

//This gives the dimension the wavefield, n1gpu-FAT,n2gpu-FAT gives the dimension of velocity
__constant__ int n1gpu;
__constant__ int n2gpu;
__constant__ int npad_gpu;
int npad_cpu;
__constant__ bool bAbsorbTop_gpu;

//__constant__ float bc_agpu;
//__constant__ float bc_bgpu;
texture <float, 2, cudaReadModeElementType> vel_ref;
cudaChannelFormatDesc cf;
cudaArray *vel_array=NULL;

//1D array
cudaArray *dampArray_gpu=NULL;
texture <float, 1, cudaReadModeElementType> damp_ref;

#define min2(v1,v2) (((v1)<(v2))?(v1):(v2)) 
#define max2(v1,v2) (((v1)<(v2))?(v2):(v1)) 
__device__ int min3(int v1,int v2,int v3){return min2(min2(v1,v2),v3);}
__device__ int min4(int v1,int v2,int v3, int v4){return min2(min2(v1,v2),min2(v3,v4));}

//address in the big wfld coord
#define dloc(z,x) ((z)+(x)*(n1gpu))
//address in the general coord
#define gloc(z,x,zlen) ((z)+(x)*(zlen))


float *p0_gpu = NULL;
float *p1_gpu = NULL;
float *dpDt2_gpu = NULL; //The dp/dt^2 term, actually computed is dp/dx^2

float *source_gpu = NULL; //Inject source slice
int   *src_iz_gpu = NULL; //Inject z location

int wfld_j_cpu = 0;

//Receiver Geometry for each shot
int *recordRecvIxGpu = NULL;
int *recordRecvIzGpu = NULL;
float *recordDataGpu = NULL; //array of size ntRecrod by nRecv.

int nRecordRecvCpu = 0;
int record_j_cpu = 0;
int ntRecCpu = 0;

//For adjoint RTM
float* reflGpu = NULL; //refl model, has the size of n1gpu X n2gpu
float *pU0_gpu = NULL;
float *pU1_gpu = NULL; //upgoing wavefield

//object controls GPU memory
struct cubeLists_gpu {
   float ** pLists; //pointer to the array of wavefield slices on gpu
   int	num;
   int nx,nz;
};

struct slice2d_gpu{
  float* p_gpu;
  int nx,nz; //For gpu memory size
  float* p_cpu; //corresponding cpu memory, might not be the same size.
  slice2d_gpu():p_gpu(NULL),nx(0),nz(0),p_cpu(NULL)
  {}
};

void free_slice2d_gpu(slice2d_gpu* pSlice) {
  if (pSlice->p_gpu != NULL) {
    cuda_call(cudaFree(pSlice->p_gpu));
  }
  pSlice->p_gpu = NULL;
  pSlice->p_cpu = NULL;
  pSlice->nx = 0;
  pSlice->nz = 0;
}

void init_slice2d_gpu(slice2d_gpu* pSlice) {
  pSlice->p_gpu = NULL;
  pSlice->p_cpu = NULL;
  pSlice->nx = 0;
  pSlice->nz = 0;
}

void reAlloc_slice2d_gpu(slice2d_gpu* p,int nz, int nx) {
  free_slice2d_gpu(p);
  p->nx = nx;
  p->nz = nz;
  cuda_call(cudaMalloc((void**)&(p->p_gpu), nx*nz*sizeof(float)));
}

//For imaging/adjRTM/Tomo
/* GPU variables */
float*   recv_gpu = NULL; //slice for back projected data
int   	*recv_iz_gpu = NULL; //Inject z location

float	*image_gpu = NULL;
cubeLists_gpu wfList_gb; //stored wfList for RTM imaging, not use check pt yet
cube_c imgGatherHorzOffset_gb;

//Global variables
int jImaging_gb = 0;

cube_c imgCube_cpu_gb;

//For WEM Tomo operators.
cube_c imgCube_S_cpu_gb, imgCube_R_cpu_gb;
bool bDoImageGather_gb = false;

//For src,recv intensity
bool bDoSrcIntens_cpu = false;
bool bDoRecvIntens_cpu = false;

slice2d_gpu srcIntense2d_gpu =  slice2d_gpu();
slice2d_gpu recvIntense2d_gpu = slice2d_gpu();

//For wfld down/up separation
FFT1dWfldSeparate wfld_sep_gb;


void init_wfLists(cubeLists_gpu* pWavList) {
   pWavList->num = 0;
   pWavList->pLists = NULL;
   pWavList->nx = 0;   pWavList->nz = 0;
}

void uninit_wfLists(cubeLists_gpu* pWavList) {
   if (pWavList->pLists) {
	  int i = 0;
	  for (i=0; i<pWavList->num; i++) {
		 cuda_call(cudaFree(pWavList->pLists[i]));
		 pWavList->pLists[i] = NULL;
	  } 
   } 
   free(pWavList->pLists); 
   init_wfLists(pWavList);
}

//NOTICE: CANNOT ALLOCATE CONTINUOUS MEMORY SPACE, because of the link-list trick in RTM function (while saving the source wavefield)
void alloc_wfLists(cubeLists_gpu* pWavList, int nx, int nz, int nFrame) {
   uninit_wfLists(pWavList);
   
   //Allocate the pointers
   pWavList->pLists = (float**)malloc(nFrame*sizeof(float*));   
   //link-list way of allocating memory, does not need large block of spare GPU memory 
   for (int i=0; i < nFrame; i++) { 
	  cuda_call(cudaMalloc( (void**)&(pWavList->pLists[i]), nx*nz*sizeof(float)));
   }
   pWavList->num = nFrame;
   pWavList->nx = nx;   pWavList->nz = nz;
}

void zero_wfLists(cubeLists_gpu* pWavList) {
   if (pWavList->pLists) {
	  int nframe = pWavList->num;
	  assert(nframe!=0);
	  int nz=pWavList->nz;	  int nx=pWavList->nx;
	  for (int i=0;i<nframe;i++) {
		 cuda_call(cudaMemset(pWavList->pLists[i],0,nz*nx*sizeof(float)));
	  }
   }
}


//Method about cube_c
void dealloc_cube_c(cube_c* pCube_gpu){
   if (pCube_gpu->p) {
	  cuda_call(cudaFree(pCube_gpu->p));
	  pCube_gpu->p = NULL;
   }
   if (pCube_gpu->bUse_p2) {
	  cuda_call(cudaFree(pCube_gpu->p2));
	  pCube_gpu->p2 = NULL;
	  pCube_gpu->bUse_p2 = false;
   }
}

void uninit_cube_c(cube_c* pCube_gpu){
   dealloc_cube_c(pCube_gpu);
   init_cube_c(pCube_gpu);
}

void zero_cube_c(cube_c* pCube_gpu) {
   int nx = pCube_gpu->nx;
   int nz = pCube_gpu->nz;
   int nf = pCube_gpu->nframe;
   
   //set mem to zero
   if (pCube_gpu->p) {
	  cuda_call(cudaMemset(pCube_gpu->p,0,nx*nz*nf*sizeof(float)));
   }
   if (pCube_gpu->bUse_p2 && pCube_gpu->p2) {
	  cuda_call(cudaMemset(pCube_gpu->p2,0,nx*nz*nf*sizeof(float)));
   }
}

//alloc use the nx,nz,nframe information it contains.
void alloc_cube_c(cube_c* pCube_gpu, bool bAllocate2ndCube = false) {
   dealloc_cube_c(pCube_gpu);
   
   int nx = pCube_gpu->nx;
   int nz = pCube_gpu->nz;
   int nf = pCube_gpu->nframe;
   assert(nx>0 && nz>0 && nf>0);
   cuda_call(cudaMalloc((void**)&(pCube_gpu->p), nx*nz*nf*sizeof(float)));
	if (bAllocate2ndCube) {
	   pCube_gpu->bUse_p2 = true;
		cuda_call(cudaMalloc((void**)&(pCube_gpu->p2), nx*nz*nf*sizeof(float)));
	}
	zero_cube_c(pCube_gpu);
}

#include "wave_kernel2d8o.cu.h"

__device__ int iz_wfld2vel(int iz_wfld) {return iz_wfld-FAT;}
__device__ int ix_wfld2vel(int ix_wfld) {return ix_wfld-FAT;}
__device__ int iz_vel2wfld(int iz_vel) {return iz_vel+FAT;}
__device__ int ix_vel2wfld(int ix_vel) {return ix_vel+FAT;}

//utility functions
void get_ntzx_from_gpu(int *p_nt, int *p_nz, int *p_nx, int* p_wfld_j) {
	//retrieve the nt,nz,nx parameters from GPU
    cuda_call(cudaMemcpyFromSymbol(p_nt, nt_gpu, sizeof(int)));
    cuda_call(cudaMemcpyFromSymbol(p_nz, n1gpu, sizeof(int)));
    cuda_call(cudaMemcpyFromSymbol(p_nx, n2gpu, sizeof(int)));
    *p_wfld_j = wfld_j_cpu;
}

struct grid_layouts {
   dim3 dimGrid;
   dim3 dimGrid_bnd;
   dim3 dimGridImage;
   dim3 dimBlock;
   int inject_dimGrid;
   int inject_dimBlock;
   
   dim3 dimGridBndLR, dimGridBndTB;
   dim3 dimBlockBnd;
   
   int desired1DblockSize;
};


//grid_layouts gd_kn_lots_gb; //grid kernel layouts
void gen_grid_layout_for_kernels(grid_layouts* p_gdlo, int nt, int nz, int nx) {
	int block1Dsize = BLOCK_SIZE*BLOCK_SIZE/2;
	//Here the convention is reversed.
  int nblocksx=(nz-2*FAT-1)/BLOCK_SIZE+1;	int nblocksy=(nx-2*FAT-1)/BLOCK_SIZE+1;
  p_gdlo->dimGrid  = dim3(nblocksx,nblocksy);	
	p_gdlo->dimBlock = dim3(BLOCK_SIZE, BLOCK_SIZE);

	int nBlockBndx = (nz-1)/BLOCK_SIZE; nBlockBndx++;
	int nBlockBndy = (nx-1)/BLOCK_SIZE; nBlockBndy++;	
	p_gdlo->dimGrid_bnd = dim3(nBlockBndx,nBlockBndy);

  //Only wfld in the region exclude nImgBnd should be used in the imaging condition.
  float nImgBnd = npad_gpu + FAT;
	int nBlockImgx = int((nz-2*nImgBnd-1)/BLOCK_SIZE); nBlockImgx++;
	int nBlockImgy = int((nx-2*nImgBnd-1)/BLOCK_SIZE); nBlockImgy++;	
	p_gdlo->dimGridImage = dim3(nBlockImgx,nBlockImgy);
	
	p_gdlo->inject_dimGrid = int((nx-2*FAT-1)/block1Dsize)+1;
	p_gdlo->inject_dimBlock = block1Dsize;
	
	p_gdlo->dimGridBndLR = dim3((nz-2*FAT-1)/BLOCK_SIZE+1,1);
	p_gdlo->dimGridBndTB = dim3((nx-2*FAT-1)/BLOCK_SIZE+1,1);
	p_gdlo->dimBlockBnd = dim3(BLOCK_SIZE, 2*FAT);
	
	p_gdlo->desired1DblockSize = BLOCK_SIZE*BLOCK_SIZE/2;
}

//The source_gpu needs to be the size of nt*nx. And you cannot have 2 sources aligned vertically. This is not flexible, need to change.
//in this case, the x direction is the n2 direction
__global__ void inject_kernel(int* iz_gpu, int it, float *source_gpu, float *p)
{
    register int ix=blockIdx.x*blockDim.x+threadIdx.x;
    if(ix>=(n2gpu-2*FAT)) return;
    p[dloc(iz_vel2wfld(iz_gpu[ix]),ix_vel2wfld(ix))] += source_gpu[it+ix*nt_gpu];
}

__global__ void record_kernel(int *irecv_iz_gpu, int *irecv_ix_gpu , int nRecv, float *wfld, int ntRec, int itRec, float *pRec) 
{
   register int index = blockIdx.x*blockDim.x+threadIdx.x;
   if (index < nRecv) {
	  pRec[itRec+index*ntRec] = wfld[dloc(iz_vel2wfld(irecv_iz_gpu[index]),ix_vel2wfld(irecv_ix_gpu[index]))];
   }
}

extern "C" void bnd_abs_ord1(float *p2_gpu, float *p1_gpu, dim3 *pDimGridLR, dim3 *pDimGridTB, dim3 *pDimBlock) {
   //Do left and right
   kernel_exec(bnd1d_kernel2_0<<<*pDimGridLR,*pDimBlock>>>(p2_gpu,p1_gpu));
   //Do top and bottom
   kernel_exec(bnd1d_kernel2_1<<<*pDimGridTB,*pDimBlock>>>(p2_gpu,p1_gpu));
}

extern "C" void gpu2dInit(int *wfld_j, int* gpu_id, int bAbsorbTop=true) {
	cuda_call(cudaSetDevice(*gpu_id));
	//initialize the constant values
	int n1,n2;
	n1 = n2 = 0;
	cuda_call(cudaMemcpyToSymbol(n1gpu,&n1,sizeof(int)));
	cuda_call(cudaMemcpyToSymbol(n2gpu,&n2,sizeof(int)));
		
	//interval for wavfield snapshots
	wfld_j_cpu = *wfld_j;
  jImaging_gb = 0; npad_cpu=0;
  
	//Init all global variables.
	vel_array = NULL; dampArray_gpu = NULL;
	p0_gpu = p1_gpu = NULL;
	source_gpu = NULL; src_iz_gpu = NULL;
	dpDt2_gpu = NULL;
  
  pU0_gpu = NULL; pU1_gpu=NULL;
  p0_gpu = p1_gpu = NULL;
  
  recv_gpu = NULL; recv_iz_gpu = NULL;
  recordRecvIxGpu = recordRecvIzGpu = NULL;
  recordDataGpu = NULL;
  
  reflGpu = image_gpu = NULL;
	//Set zero for the img_off struct
	init_cube_c(&imgGatherHorzOffset_gb);
  init_cube_c(&imgCube_cpu_gb);
  
  init_wfLists(&wfList_gb);
  
  bDoSrcIntens_cpu = bDoRecvIntens_cpu = false;
  //clear the srcIntense and dstIntense
  init_slice2d_gpu(&srcIntense2d_gpu);
  init_slice2d_gpu(&recvIntense2d_gpu);
  
  bDoImageGather_gb = false;
  
  FFT1dWfldSeparate_init(&wfld_sep_gb);
}

//This routine operates on CPU.
//wfld_j, how many timesteps to dump the wavefield.
extern "C" void gpu2dModeling_init(int *wfld_j, int *record_j, int *gpu_id)
{
   gpu2dInit(wfld_j,gpu_id);
   source_gpu = NULL;
   src_iz_gpu = NULL;
   record_j_cpu = *record_j;
   recordRecvIxGpu=NULL;
   recordRecvIzGpu=NULL;
   recordDataGpu = NULL;
}

extern "C" bool gpu2dSetVelocity(float *cpu_space, int *new_nz, int *new_nx, int *nt, float *dt, float *dz, float *dx,
										  float *dampCoefArray, int *dampArrayLength, int* npad) 
{
   int n1_old, n2_old;
   //First retrieve the current velocity model size in device memory
   cuda_call(cudaMemcpyFromSymbol(&n1_old, n1gpu, sizeof(int))); 
   cuda_call(cudaMemcpyFromSymbol(&n2_old, n2gpu, sizeof(int)));

   bool bDoTopAbsorp = true;
   cuda_call(cudaMemcpyToSymbol(bAbsorbTop_gpu, &bDoTopAbsorp, sizeof(bool)));

   cuda_call(cudaMemcpyToSymbol(npad_gpu, npad, sizeof(float)));
   npad_cpu = *npad;
   
   int new_wf_n1 = *new_nz + 2*FAT; int new_wf_n2 = *new_nx + 2*FAT;
   bool bSizeNotChanged = (n1_old==new_wf_n1 && n2_old==new_wf_n2);
   
   if (!bSizeNotChanged) //needs to reallocate the velocity array
	{
	   //allocate damping Coef Array
	   if(*dampArrayLength < max2((new_wf_n1/2+1),(new_wf_n2/2+1)) ) {
		   fprintf(stderr,"\n The dampArrayLength is too small!!!!");
		   exit(-1);
		}

	   cuda_call(cudaFreeArray(dampArray_gpu));
	   cf = cudaCreateChannelDesc<float>();
	   cuda_call(cudaMallocArray(&dampArray_gpu,&cf,*dampArrayLength,1));
	   
		//allocate Two slices for wavefield propagation
		cuda_call(cudaFree(p0_gpu)); cuda_call(cudaFree(p1_gpu));
		cuda_call(cudaMalloc(    (void**) &p0_gpu,  (new_wf_n1)*(new_wf_n2)*sizeof(float)));
		cuda_call(cudaMalloc(    (void**) &p1_gpu,  (new_wf_n1)*(new_wf_n2)*sizeof(float)));

		cuda_call(cudaFree(source_gpu)); 
		//need to change later the way to inject data
		cuda_call(cudaMalloc(    (void**) &source_gpu,  (*nt)*(*new_nx)*sizeof(float)));
		cuda_call(cudaMalloc(    (void**) &src_iz_gpu,  (*new_nx)*sizeof(int)));

		//allocate texture memory as array
		cf = cudaCreateChannelDesc<float>();
		cuda_call(cudaFreeArray(vel_array));
		cuda_call(cudaMallocArray(&vel_array, &cf, *new_nz, *new_nx));
		
		//Copy the constants from CPU memory to device constant memory.
		cuda_call(cudaMemcpyToSymbol(n1gpu, &new_wf_n1, sizeof(int)));
		cuda_call(cudaMemcpyToSymbol(n2gpu, &new_wf_n2, sizeof(int)));
		cuda_call(cudaMemcpyToSymbol(nt_gpu,nt, sizeof(int)));
	   
	   //Calculate the coefficients of finite-difference stencils
		cuda_call(cudaMemcpyToSymbol(dx_gpu, dx, sizeof(float)));
		cuda_call(cudaMemcpyToSymbol(dz_gpu, dz, sizeof(float)));
	   float coeffs_cpu[COEFFS_SIZE]=get_coeffs(*dz,*dx);
	   //copy the stencils to GPU
	   cuda_call(cudaMemcpyToSymbol(coeffs, coeffs_cpu, COEFFS_SIZE*sizeof(float)));
	}
	//cp the velocity data from cpu to gpu
	cuda_call(cudaMemcpyToArray(vel_array, 0,0, cpu_space, (*new_nz)*(*new_nx)*sizeof(float), cudaMemcpyHostToDevice));
	//Rebind the reference to the texture memory.
	cuda_call(cudaUnbindTexture(vel_ref));
    cuda_call(cudaBindTextureToArray(vel_ref, vel_array));
	
	//cp the dampCoef from cpu to gpu
	cuda_call(cudaMemcpyToArray(dampArray_gpu, 0,0, dampCoefArray, (*dampArrayLength)*sizeof(float), cudaMemcpyHostToDevice));
	cuda_call(cudaUnbindTexture(damp_ref));
    cuda_call(cudaBindTextureToArray(damp_ref, dampArray_gpu));
	return !bSizeNotChanged;
}


//here assume the velocity model grid, does not change for each shot. Not good, should be able to change for each shot
//return true means the size has changed, need to relocate memory
extern "C" bool gpu2dModeling_setVelocity(float *cpu_space, int *new_nz, int *new_nx, int *nt, float *dt, float *dz, float *dx,
										  float *dampCoefArray, int *dampArrayLength, int* npad, int nRecv, int ntRec, int* ix_recv, int* iz_recv)
{
	bool bRet = gpu2dSetVelocity(cpu_space,new_nz,new_nx,nt,dt,dz,dx,dampCoefArray,dampArrayLength,npad);
	//Save the Receiver Geometry Information
	nRecordRecvCpu = nRecv;
	ntRecCpu = ntRec;
	//printf("\ntRec... =%d,%d,%d\n",ntRec,*nt,record_j_cpu);
	assert(ntRec == ((*nt-1)/record_j_cpu)+1);
	
	cuda_call(cudaFree(recordRecvIxGpu)); cuda_call(cudaFree(recordRecvIzGpu));
	cuda_call(cudaMalloc((void**)&recordRecvIxGpu,  nRecv*sizeof(float)));
	cuda_call(cudaMalloc((void**)&recordRecvIzGpu,  nRecv*sizeof(float)));
	cuda_call(cudaMemcpy(recordRecvIxGpu, ix_recv, nRecv*sizeof(float), cudaMemcpyHostToDevice));
	cuda_call(cudaMemcpy(recordRecvIzGpu, iz_recv, nRecv*sizeof(float), cudaMemcpyHostToDevice));
	cuda_call(cudaFree(recordDataGpu));
	cuda_call(cudaMalloc((void**)&recordDataGpu, nRecv*ntRec*sizeof(float)));
	return bRet;
}


//iz_cpu the inject source locations:iz along the ix grid axis
extern "C" void gpu2d_wave_prop(int *iz_cpu, int iz_cpu_len, float *source_cpu, float *wfld, float *recordBuffer)
{
    int nt, nz, nx, i3, wfld_j;
    float *ptemp;
	int nRecv, record_j, itRec, ntRec;
	
	//retrieve the nt,nz,nx parameters from GPU
    get_ntzx_from_gpu(&nt,&nz,&nx,&wfld_j);
	
	record_j = record_j_cpu;
	nRecv =  nRecordRecvCpu;
	ntRec =  ntRecCpu;

	grid_layouts gdlo;
	gen_grid_layout_for_kernels(&gdlo,nt,nz,nx);
		
	//The source inject slice copy to GPU, need to revise.
    cuda_call(cudaMemcpy(source_gpu, source_cpu, nt*(nx-2*FAT)*sizeof(float), cudaMemcpyHostToDevice));
	//printf("iz_cup_len=%d, nx=%d\n",iz_cpu_len,nx);
	assert(iz_cpu_len == nx - 2*FAT);
    cuda_call(cudaMemcpy(src_iz_gpu, iz_cpu, (nx-2*FAT)*sizeof(int), cudaMemcpyHostToDevice));
    
	//clean up the wavfield memory
    cuda_call(cudaMemset(p0_gpu, 0,nz*nx*sizeof(float)));
    cuda_call(cudaMemset(p1_gpu, 0,nz*nx*sizeof(float)));
	
	/* The ACTUAL WAVE PROPAGATION */
    i3 = 0; //Here at time it, calculate the wavfield values at time it+1
    //p0-->it, p1-->it+1
    for (int it=0; it<nt; it++)
    {
    //source at time 'it' should be applied to wavefield at time 'it+1'
    kernel_exec(inject_kernel<<<gdlo.inject_dimGrid,gdlo.inject_dimBlock>>>(src_iz_gpu, it, source_gpu, p1_gpu));
		//Absorbing boundary condition on wavefield at it+1, using wavefield at it(p0_gpu)
		//kernel_exec(bnd1d_kernel<<<dimGrid_bnd,dimBlock>>>(p0_gpu,p1_gpu));
		kernel_exec(paddingSponge1_kernel<<<gdlo.dimGrid,gdlo.dimBlock>>>(p1_gpu,p0_gpu));
		bnd_abs_ord1(p1_gpu, p0_gpu, &gdlo.dimGridBndLR, &gdlo.dimGridBndTB, &gdlo.dimBlockBnd);	
		//repeatively copy the wavefield memory out to host memory for snapshot
    if(it%wfld_j==0) //i.e jSnapshot
    {
        cuda_call(cudaMemcpy(&wfld[nz*nx*i3], p0_gpu, nz*nx*sizeof(float), cudaMemcpyDeviceToHost));
        i3++;
    }
		
		int block1D = gdlo.inject_dimBlock;
		//Harvest the data at time it;
		if(it%record_j==0) {
		   itRec = it / record_j;
		   kernel_exec(
			  record_kernel<<<(nRecv-1)/block1D+1,block1D>>>(recordRecvIzGpu,recordRecvIxGpu,nRecv,p0_gpu,ntRec,itRec,recordDataGpu));
		}
		
    kernel_exec(wave_kernel<<<gdlo.dimGrid,gdlo.dimBlock>>>(p0_gpu, p1_gpu, p0_gpu,NULL));
		//switch the wavefield memory pointer, afterthen p0->it+1, p1->it+2
    ptemp=p0_gpu; p0_gpu=p1_gpu;p1_gpu=ptemp;
  }
	//Copy the recorded data back to the host memory
	cuda_call(cudaMemcpy(recordBuffer,recordDataGpu,ntRec*nRecv*sizeof(float),cudaMemcpyDeviceToHost));
	return;
}


extern "C" void gpuModeling2d_uninit() {
	cuda_call(cudaFree(p0_gpu));
	cuda_call(cudaFree(p1_gpu));
	cuda_call(cudaUnbindTexture(vel_ref));
	cuda_call(cudaUnbindTexture(damp_ref));
  cuda_call(cudaThreadExit());
}


