#include "stdafx.h"
#include "dmRsf.hpp"

#include "srcWav.h"
#include "velFD2d.h"
#include "point3d.hpp"

#include "modelingFD2d_op.h"
#include "modelingFDgpu.h"

#include "SR_manager.h"
#include "paddingFD.h"
#include "paddingFD_damping.h"
#include "boundaryFD.h"

#include "param_entry.h"
#include "myOMP.h"

#include "FWI_FD.h"
#include "FWI_FD_beamFocus.h"
#include "FWI_FD_DSO.h"
#include "stackTraceC.h"
#include "FWI_objFunc.h"

#include "commonOps.h"
#include "solverNonlinearSearch.h"

#include "sep_params.h"
//Read params
#include "nonLinSolver.param.h"
#include "velAna.param.h"
#include "wavePropagate.param.h"
//#include "../misc.h"

/**
 * For beam focusing, Wave Equation Tomography, Just do the 
 * non-linear line search, Transmission case
 */


using namespace Yang;
using namespace std;
using boost::shared_ptr;
   
//Doing beam wave-equation tomography for transmission case
//Use cuda Wave modeling engine

/// The input of the program
/// initial velocity model vel
/// observed data data_obs along with its geometry

int main (int argc, char *argv[])
{
  initCallStackTrace();
  
  //Read params, params need contain geometry info that is exact the same as the input data. In
  //synthetic this is easy to do.
  shared_ptr<sep_params> autoPars(new sep_params(argc,argv));
  sep_params* pars = autoPars.get();
  
  shared_ptr<vel2d> velObj(new vel2d(csz_tag_vel,dm::IN,"vel"));
  shared_ptr<srcWav> waveletObj(new srcWav1d(csz_tag_wavelet,dm::IN,"wavelet"));
  
  //Read in wavelet and initial velocity, input 2nd derivative of Gaussian wavelet 
  waveletObj->loadWavelet();
  velObj->loadVel();
  
  //Create the modeling Obj for modeling and also RTM(for calculating the gradient)
  auto_ptr<modelingFDgpu2d_op> modelingObj(new modelingFDgpu2d_op());
  int gpuID = pars->get_int(csz_gpuID,0);
  modelingObj->setGPU_ID(gpuID);
  
  //Create the padding and boundary processor
  shared_ptr<SR_managerFD2d>  SR_managerObj(new SR_managerFD2d(waveletObj.get()));
  
  shared_ptr<paddingFD2d> paddingObj (new paddingFD2d(NULL));
  
  //boundary condition
  shared_ptr<boundaryFD2d> boundaryObj(new boundaryFD2d());
  
  //assemble
  modelingObj->assemble(SR_managerObj.get(), boundaryObj.get(),paddingObj.get());
  
  //Read in the parameters for the modeling Operator
  FDModelingParams params = FDModelingParams();
  velAnaParam velAnaParams;
  FWI_Param FWIparam;
  
  
  params.pVel = velObj.get();
  params.pWavelet = waveletObj.get();
  params.modelGridShape  = velObj->domain(); //local
  params.velModelShape   = velObj->domain();  //global
  params.finalImageShape = params.velModelShape;
  params.ndim = params.velModelShape.get_ndim();
   
  read_twoway_prop_param(&params,pars);
  float maxFreq,vmax,vmin;
  //Read in the peak/center freq of the seismic wavelet
  maxFreq = waveletObj->get_float("peakFreq");
  waveletObj->setPeakFreq(maxFreq);
  vmax = velObj->maxval();
  vmin = velObj->minval(); 
  //calc time parameters
  calcTimeParamsFor2wayModeling(&(params.timeParam),maxFreq,vmax,vmin,params.modelGridShape);
  read_migration_param(&params,pars);   
      
  read_velAna_param(&velAnaParams,pars);
  read_FWI_param(&FWIparam,pars);
   
  //load the observed data
  velAnaParams.pDataObs = new cubeData<float>(csz_tag_dataObs,dm::IN,"data observed");
  velAnaParams.pDataObs->init(NULL);
  velAnaParams.pDataObs ->domain().info("Data obs info");
  velAnaParams.dObsCube.alloc(velAnaParams.pDataObs->domain());
  velAnaParams.pDataObs->seq_grabAll(&(velAnaParams.dObsCube));
  params.pRsfSeisRec = &velAnaParams.dObsCube;
 
  //Create FWIObj
  FWI_FD* pFWIObj = NULL;
  pFWIObj =  create_FWI_obj(&velAnaParams,&params,modelingObj.get(),&FWIparam,pars);
  //After the initialization of params, init the non-linear solver with observed data
  objFunc* pObjFunc = new FWI_objFunc(pFWIObj);
  
  solverParam solvPar;
  read_nonLinSolver_par(&solvPar,pars);
  
  //solverNonlinearSearch(objFunc* pObj, const gVecComponent* pInitModel);
  gVecComponent initVelModelVec, finalModelVec;
  
  params.pVel->copyTo(initVelModelVec);
  initVelModelVec.reciprocal(); //convert to slowness
  params.pVel->allocMemoryTo(finalModelVec);
  solverNonlinearSearch nonLinSolver(pObjFunc, &initVelModelVec,&solvPar);
  
  //invert the model
  stdVecFloat finalErrorVec;
  int finalIter;
  
  SR_managerObj->init(&params);
  velAnaParams.velModelMask.alloc(params.velModelShape);
  SR_managerObj->getVelModelMask(velAnaParams.velModelMask);
  
  //velAnaParams.velModelMask.dump("dump5");
  
  vector<gVecComponent*> modelsVec, gradsVec;
  nonLinSolver.solve(finalModelVec, finalErrorVec, finalIter, 
                    &modelsVec, &gradsVec);
  LOG(m4);
  //Save result
  finalModelVec.dump("out");
  finalModelVec.unInit();
  
  //dump error
  int iSize = finalErrorVec.size(); 
  PRV(iSize);
  cubeData<float> finalErrCube(iSize);
  myMemcpy(finalErrCube.pval,&finalErrorVec[0],iSize);
  finalErrCube.dump("error");
  LOG(m5);
  //dump models and grads
  PRV(modelsVec.size());
  dumpVecModels(modelsVec,"models");
  dumpVecModels(gradsVec,"grads");
  LOG(deleteVecModels);
  deleteVecPointers(modelsVec);
  deleteVecPointers(gradsVec);
  
  //Dump the final data
  cubeData<float>& finalData = pObjFunc->getCalcData();
  finalData.dump("fData");
  
  //clean up
  velAnaParams.pDataObs->unInit();
  SAFE_DELETE(pFWIObj);
  SAFE_DELETE(pObjFunc);
  return 0;
}

