#include "stdafx.h"
#include "dmRsf.hpp"

#include "srcWav.h"
#include "velFD2d.h"
#include "point3d.hpp"

#include "waveOpParams.h"
#include "waveOpOneway.h"

#include "param_entry.h"
#include "WEMVA.h"

#include "stackTraceC.h"
#include "solverNonlinearSearch.h"
#include "commonOps.h"

#include "dipEstimator.h"
#include "ADCIG_Recon.h"
#include "WEMVA_RMO_Imit.h"
#include "util_seplib.h"
#include "rayTracer.h"
#include "RayTomoOp.h"

///Read parameters
#include "nonLinSolver.param.h"
#include "velAna.param.h"
#include "misc.h"


/**
 * For beam focusing, Wave Equation Tomography, Just do the 
 * non-linear line search 
 */

//Reflection case

#ifdef MPI_SUCCESS
#error
#endif

using namespace Yang;
using namespace std;
using namespace seplib;

using boost::shared_ptr;   
//Doing beam wave-equation tomography for reflection case.
//Use cuda Wave modeling engine

/// The input of the program
/// initial velocity model vel
/// observed data data_obs along with its geometry
int main (int argc, char *argv[])
{
  initCallStackTrace(); init_omp(1);
  //Read params, params need contain geometry info that is exact the same as the input data. In
  //synthetic this is easy to do.
  shared_ptr<sep_params> autoPars(new sep_params(argc,argv));
  sep_params* pars = autoPars.get();
  
  shared_ptr<vel2d>  velObj     (new vel2d(csz_tag_vel,dm::IN,"vel"));
  shared_ptr<srcWav> waveletObj (new srcWav1d(csz_tag_wavelet,dm::IN,"wavelet"));
  
  //Read in wavelet and initial velocity, input 2nd derivative of Gaussian wavelet 
  waveletObj->loadWavelet();
  velObj->loadVel(2); velObj->info("Initial velocity model");
  
  shared_ptr<ImgPar> imagingPar(new ImgPar());  ImgPar& imgPar = *imagingPar;
  //load acquisition parameters
  shared_ptr<AcqPar> acqPar(new AcqPar());
  acqPar->readParam(pars,velObj->domain());
  imgPar.readParam(pars,*acqPar); //Load imaging parameters

  shared_ptr<waveOpOneway>  modelingObj;
  //Read in the parameters for the modeling Operator
  shared_ptr<waveOpParams>  waveOpPar(new waveOpParams()); waveOpParams& params = *waveOpPar;
  
  velAnaParam   velAnaParams; WEMVA_Param WEMVAparam;
  params.readParam(pars);
  bool bUseOneWay = params.bOneway();
  //Here we just focus on use one-way
  if (bUseOneWay) {
    modelingObj.reset(new waveOpOneway(waveOpPar,acqPar,imagingPar,*velObj));
    
    params.read_oneway_prop_param(pars);
    velAnaParams.dObsFilename = pars->get_string(csz_tag_dataObs); //For Imaging, observed data tag
  } else {
    params.read_twoway_prop_param(pars);
    //calc time parameters
    float maxFreq,vmax,vmin;
    //Read in the peak/center freq of the seismic wavelet
    maxFreq = waveletObj->get_float("peakFreq");
    waveletObj->setPeakFreq(maxFreq);
    vmax = velObj->maxval();  vmin = velObj->minval();
    params.calcTimeParamsFor2wayModeling(maxFreq, vmax,vmin, velObj->domain());
  }
  modelingObj->velAnaParsIs(&velAnaParams); //set the velAnaParam to modeling obj.
  
  //Here, need to read in the mask as the random sparse sampling of the migrated gather.
  cubeData<float>* pImageMask = NULL;
  if (tagExist(csz_tag_img_rand_samp_mask)) {
    pImageMask = new cubeData<float>(csz_tag_img_rand_samp_mask,dm::IN,"image random sample mask");
    LOG('Enable Random Sparse Sampling of the migrated Image.');
    pImageMask->init(NULL); //read the mask in.
  }
  params.pImageMask = pImageMask;

  if (!bUseOneWay) { //if not use oneway, then read the recorded data
    //load the observed data
    velAnaParams.pDataObs = new cubeData<float>(csz_tag_dataObs,dm::IN,"data observed");
    velAnaParams.pDataObs->init(NULL);
    velAnaParams.pDataObs ->domain().info("Data obs info");
    velAnaParams.dObsCube.alloc(velAnaParams.pDataObs->domain());
    velAnaParams.pDataObs->seq_grabAll(&(velAnaParams.dObsCube));
  } //otherwise let Yaxun's program to take care of it.
  read_velAna_param(&velAnaParams,pars);
  read_WEMVA_param(&WEMVAparam,pars,imgPar);
  //assign the gradient mask to image mask. assume vel and img grids are aligned.
  WEMVAparam.imgh0_mask =  velAnaParams.gradMaskPreSmooth_;
  WEMVAparam.imgh0_mask -> info("WEMVAparam.imgh0_mask");
  
  
  //read_WEMVA_param(&WEMVAparam,pars); is called inside the WEMVA_obj function
  //shared_ptr<WEMVA_FD> pWEMVAobj(create_WEMVA_obj(&velAnaParams, modelingObj, &WEMVAparam,pars));
  const hypercube& shpOffimg= imgPar.finalSubOffImageShp(); axis ax_h = shpOffimg.get_axis(3);
  const hypercube& shp0img  = imgPar.finalImageShp();
  
  WEMVAparam.vmodTrue.reset(new cubeDataF()); WEMVAparam.vmodTrue->load("vtrue",2);
  
  //mask out the outermost part of the image.
  vecAxis axes = shp0img.get_axes();
  const axis &ax_z = axes[0], &ax_x = axes[1];
  //ax_z.print_info("axz"); ax_x.print_info("axx");
  //The params for marmousi model
  //axis az_crop = ax_z.return_truncated(680.0f,ax_z.get_coord(ax_z.n-1)); axis ax_crop = ax_x.return_truncated(6200.0f,11800.0f);
  //This is for BP model
  axis az_crop = ax_z.return_truncated(3.0,ax_z.get_coord(ax_z.n-1)); axis ax_crop = ax_x.return_truncated(1.0,29.0);
  axes[0]=az_crop; axes[1]=ax_crop;
  WEMVAparam.croppedImg0_.reset(new cubeDataF(hypercube(axes)));
  
  //Create WEMVA object
  shared_ptr<WEMVA_FD> pWEMVA_obj = create_WEMVA_obj(&velAnaParams,modelingObj,&WEMVAparam, pars);
  if (WEMVAparam.mWEMVAtype == WEMVA_Param::RMO) {
    LOGC("create WEMVA_RMO object");
    pWEMVA_obj = create_WEMVA_imit(pars,ax_h,shp0img,velAnaParams,
        modelingObj,WEMVAparam,acqPar, params.bOneway(),*velObj); 
  }
  VERIFY(NULL!=pWEMVA_obj.get());
  
  //Create the solver
  solverParam solvPar;  read_nonLinSolver_par(&solvPar,pars);
  gVecComponent initSlowModelVec, finalModelVec;
  
  velObj->copyTo(initSlowModelVec); initSlowModelVec.reciprocal();
  velObj->allocMemoryTo(finalModelVec);
  shared_ptr<nonLinSearchDirection> dirPredict(new nonLinSearchDirection(nonLinSearchDirection::PolakR,velObj->domain()));
  
  solverNonlinearSearch nonLinSolver(pWEMVA_obj, &initSlowModelVec,&solvPar,dirPredict);
  
  stdVecFloat finalErrorVec; int finalIter;
  //no need to generate gradient mask, beacause the deltaS is generated Windowed by the apertures of each shot respectively.
  vector<gVecComponent*> modelsVec, gradsVec, datsVec;
  nonLinSolver.solve( finalModelVec, finalErrorVec, finalIter, 
                     &modelsVec, &gradsVec, &datsVec);
  
  LOG("m4");
  //Save result, convert from slow to vel
  finalModelVec.reciprocal(); finalModelVec.dump("out"); 
  
  //dump error
  int iSize = finalErrorVec.size(); PRV(iSize);
  dumpStdVec(finalErrorVec,"error");
  LOG("m5");
  //dump models and grads
  PRV(modelsVec.size()); PRV(gradsVec.size());
  dumpVecModels(modelsVec,"mods");dumpVecModels(gradsVec,"grads");dumpVecModels(datsVec,"dats");
  LOG("deleteVecModels");
  deleteVecPointers(modelsVec);deleteVecPointers(gradsVec);
  LOG("dump final image");
  //Dump the final data
  pWEMVA_obj->postCalcData();
  const cubeData<float>& finalImage = WEMVAparam.Img0_ax_t; //get the angle domain image.
  finalImage.dump("fImage");
  
  LOG("final clean up");
  //clean up
  if (!bUseOneWay) {
    velAnaParams.pDataObs->unInit();
  }
  return 0;
}

