#include "stdafx.h"
#include "dmRsf.hpp"

#include "srcWav.h"
#include "velFD2d.h"
#include "point3d.hpp"

#include "waveOpParams.h"
#include "waveOpOneway.h"

#include "param_entry.h"
#include "WEMVA.h"

#include "stackTraceC.h"
#include "solverNonlinearSearch.h"
#include "commonOps.h"

#include "dipEstimator.h"
#include "ADCIG_Recon.h"
#include "WEMVA_RMO_Imit.h"
#include "util_seplib.h"
#include "rayTracer.h"
#include "RayTomoOp.h"

///Read parameters
#include "nonLinSolver.param.h"
#include "velAna.param.h"
#include "misc.h"

#include "my_mpi.h"
#include "waveOpMPI.h"
#include "my_timer.h"

//Need to build binary library files, annoying.
//#include <boost/mpi/environment.hpp>
//#include <boost/mpi/communicator.hpp>
//#include <boost/mpi/collectives.hpp>

/**
 * For beam focusing, Wave Equation Tomography, Just do the 
 * non-linear line search 
 */

//Reflection case

using namespace Yang;
using namespace std;
using namespace seplib;

using boost::shared_ptr;   
//Doing beam wave-equation tomography for reflection case.

/// The input of the program
/// initial velocity model vel
/// observed data data_obs along with its geometry
int main (int argc, char *argv[])
{
  
  mpi::environment MPI_Init(&argc,&argv);
  mpi::communicator world;

  shared_ptr<sep_params> autoPars(new sep_params(argc,argv));
  sep_params* pars = autoPars.get();
  
  int irank=world.rank(), nw = world.size();
  if (irank==1) {PRVEC(argv,argc);}

  //important, fix the stdin/stdout if no is provided.
  if (irank==0) mpi_sep_send_args(nw,10,0);
  else mpi_sep_receive_args();

  if (world.rank()==0) { //master node, execute all the main logic
    initCallStackTrace(); init_omp(0);
    //Read params, params need contain geometry info that is exact the same as the input data. In
    //synthetic this is easy to do.
    
    shared_ptr<vel2d>  velObj     (new vel2d(csz_tag_vel,dm::IN,"vel"));
    shared_ptr<srcWav> waveletObj (new srcWav1d(csz_tag_wavelet,dm::IN,"wavelet"));
    
    //Read in wavelet and initial velocity, input 2nd derivative of Gaussian wavelet 
    waveletObj->loadWavelet();
    velObj->loadVel(2); velObj->info("Initial velocity model");
    
    shared_ptr<ImgPar> imagingPar(new ImgPar());  ImgPar& imgPar = *imagingPar;
    //load acquisition parameters
    shared_ptr<AcqPar> acqPar(new AcqPar());
    acqPar->readParam(pars,velObj->domain());
    imgPar.readParam(pars,*acqPar); //Load imaging parameters
    LOG("A2");

    shared_ptr<waveOpOneway> modelingObj;
    //Read in the parameters for the modeling Operator
    shared_ptr<waveOpParams>  waveOpPar(new waveOpParams()); waveOpParams& params = *waveOpPar;
    
    velAnaParam   velAnaParams; WEMVA_Param WEMVAparam;
    params.readParam(pars);
    bool bUseOneWay = params.bOneway();
    //Here we just focus on use one-way
    LOG("A3.5");
    //split the jobs
    shared_ptr<workScheduler1d> ws(new workScheduler1d(acqPar->nShots(),world.size()-1,workScheduler1d::chunk));
    LOG("A4");
    Yang::timer clk("RMO_WEMVA");
    if (bUseOneWay) {
      modelingObj.reset(new waveOpOnewayMPI(waveOpPar,acqPar,imagingPar,*velObj,pars,ws));
      params.read_oneway_prop_param(pars);
      velAnaParams.dObsFilename = pars->get_string(csz_tag_dataObs); //For Imaging, observed data tag
      //split the csou files
      LOG("B2");
      int iax = 4; //sx is axis 4 for oneway program
      split_file_MPI(params.onewayPar()->csouFileName, params.suf_MPI_,4,*ws,"");
      string fn_crec = params.onewayPar()->crecFileName;
      //Need to split the data
      split_file_MPI(fn_crec,params.suf_MPI_,4,*ws,"");
      LOG("After spliting the csou/crec files"); cerr.flush();
    } else {
      params.read_twoway_prop_param(pars);
      //calc time parameters
      float maxFreq,vmax,vmin;
      //Read in the peak/center freq of the seismic wavelet
      maxFreq = waveletObj->get_float("peakFreq");
      waveletObj->setPeakFreq(maxFreq);
      vmax = velObj->maxval();  vmin = velObj->minval();
      params.calcTimeParamsFor2wayModeling(maxFreq, vmax,vmin, velObj->domain());
    }
    modelingObj->velAnaParsIs(&velAnaParams); //set the velAnaParam to modeling obj.
    
    //Here, need to read in the mask as the random sparse sampling of the migrated gather.
    cubeData<float>* pImageMask = NULL;
    if (tagExist(csz_tag_img_rand_samp_mask)) {
      pImageMask = new cubeData<float>(csz_tag_img_rand_samp_mask,dm::IN,"image random sample mask");
      LOG('Enable Random Sparse Sampling of the migrated Image.');
      pImageMask->init(NULL); //read the mask in.
    }
    params.pImageMask = pImageMask;
  
    if (!bUseOneWay) { //if not use oneway, then read the recorded data
      //load the observed data
      velAnaParams.pDataObs = new cubeData<float>(csz_tag_dataObs,dm::IN,"data observed");
      velAnaParams.pDataObs->init(NULL);
      velAnaParams.pDataObs ->domain().info("Data obs info");
      velAnaParams.dObsCube.alloc(velAnaParams.pDataObs->domain());
      velAnaParams.pDataObs->seq_grabAll(&(velAnaParams.dObsCube));
    } //otherwise let Yaxun's program to take care of it.
    read_velAna_param(&velAnaParams,pars);
    read_WEMVA_param(&WEMVAparam,pars,imgPar);
    //assign the gradient mask to image mask. assume vel and img grids are aligned.
    WEMVAparam.imgh0_mask =  velAnaParams.gradMaskPreSmooth_;
    WEMVAparam.imgh0_mask -> info("WEMVAparam.imgh0_mask");
    
    //read_WEMVA_param(&WEMVAparam,pars); is called inside the WEMVA_obj function
    //shared_ptr<WEMVA_FD> pWEMVAobj(create_WEMVA_obj(&velAnaParams, modelingObj, &WEMVAparam,pars));
    const hypercube& shpOffimg= imgPar.finalSubOffImageShp(); axis ax_h = shpOffimg.get_axis(3);
    const hypercube& shp0img  = imgPar.finalImageShp();
    
    WEMVAparam.vmodTrue.reset(new cubeDataF()); WEMVAparam.vmodTrue->load("vtrue",2);
    
    //mask out the outermost part of the image.
    vecAxis axes = shp0img.get_axes();
    const axis &ax_z = axes[0], &ax_x = axes[1];
    //ax_z.print_info("axz"); ax_x.print_info("axx");
    //The params for marmousi model
    //axis az_crop = ax_z.return_truncated(680.0f,ax_z.get_coord(ax_z.n-1)); axis ax_crop = ax_x.return_truncated(6200.0f,11800.0f);
    //This is for BP model
    axis az_crop = ax_z.return_truncated(3.0,ax_z.get_coord(ax_z.n-1)); 
    axis ax_crop = ax_x.return_truncated(0.0,28.0);
    axes[0]=az_crop; axes[1]=ax_crop;
    WEMVAparam.croppedImg0_.reset(new cubeDataF(hypercube(axes)));
    
    //Create WEMVA object
    shared_ptr<WEMVA_FD> pWEMVA_obj = create_WEMVA_obj(&velAnaParams,modelingObj,&WEMVAparam, pars);
    if (WEMVAparam.mWEMVAtype == WEMVA_Param::RMO) {
      LOGC("create WEMVA_RMO object");
      pWEMVA_obj = create_WEMVA_imit(pars,ax_h,shp0img,velAnaParams,
          modelingObj,WEMVAparam,acqPar, params.bOneway(),*velObj); 
    }
    VERIFY(NULL!=pWEMVA_obj.get());
    
    //Create the solver
    solverParam solvPar;  read_nonLinSolver_par(&solvPar,pars);
    gVecComponent initSlowModelVec, finalModelVec;
    
    velObj->copyTo(initSlowModelVec); initSlowModelVec.reciprocal();
    velObj->allocMemoryTo(finalModelVec);
    shared_ptr<nonLinSearchDirection> dirPredict(new nonLinSearchDirection(nonLinSearchDirection::steepest,velObj->domain()));
    
    solverNonlinearSearch nonLinSolver(pWEMVA_obj, &initSlowModelVec,&solvPar,dirPredict);
    
    stdVecFloat finalErrorVec; int finalIter;
    LOG("begin inversion"); cerr.flush();
    //no need to generate gradient mask, beacause the deltaS is generated Windowed by the apertures of each shot respectively.
    ocDumpF modelsVec("mods","model vecs",true),gradsVec("grads","grad vecs",true),datsVec("dats","dat vecs",false);
    nonLinSolver.solve( finalModelVec, finalErrorVec, finalIter,
                       &modelsVec, &gradsVec, &datsVec);
    
    LOG("m4");
    //Save result, convert from slow to vel
    finalModelVec.reciprocal(); finalModelVec.dump("fout"); 
    
    modelsVec.unInit(); gradsVec.unInit(); datsVec.unInit();

    //Dump the final data
    pWEMVA_obj->postCalcData();
    //const cubeData<float>& finalImage = WEMVAparam.Img0_ax_t; //get the angle domain image.
    const cubeData<float>& finalImage = pWEMVA_obj->getCalcData();
    finalImage.dump("fImage.H",true);
    
    LOG("final clean up");
    //clean up
    if (!bUseOneWay) {
      velAnaParams.pDataObs->unInit();
    }

    clk.check("RMO_WEMVA-mpi finish");
    //release the slave modes, send exit cmd.
    for (int ir=1; ir<world.size(); ir++) {
      world.send(ir,EXITAG,string(""));
    }
  }
  
  mpi::slaveModeMPI slaveMode;
  if (world.rank() > 0) { //the slaver node, just accept cmds, execute it, and wait for another cmd. Keep in mind that the master process is also a slave node.
    slaveMode.on();    
  }
  
  //finally synchronize, then exit.
  world.barrier();
  return 0;
}

