#include "stdafx.h"
#include "dmRsf.hpp"

#include "srcWav.h"
#include "velFD2d.h"
#include "point3d.hpp"

#include "waveOpParams.h"
#include "waveOpOneway.h"

#include "sep_params.h"
#include "param_entry.h"
#include "WEMVA.h"

#include "stackTraceC.h"
#include "solverNonlinearSearch.h"
#include "commonOps.h"

#include "dipEstimator.h"
#include "ADCIG_Recon.h"
#include "WEMVA_RMO_Imit.h"
#include "util_seplib.h"
#include "misc.h"

#include "rayTracer.h"
#include "RayTomoOp.h"

///Read parameters
#include "nonLinSolver.param.h"
#include "velAna.param.h"

/*
 * Input migrated offset gather(might be random subsampled,
 then compute the objective function and calculate the back projected data.
 */

//Reflection case

using boost::shared_ptr;
using namespace Yang;
using namespace std;
using namespace seplib;
   
//Doing beam wave-equation tomography for reflection case.
//Use cuda Wave modeling engine

/// The input of the program
/// initial velocity model vel, and the migrated image 0off and multi-offset
/// It calculated the backproj data.
int main (int argc, char *argv[])
{
  initCallStackTrace();
  init_omp(1);
  
  //Read params, params need contain geometry info that is exact the same as the input data. In
  //synthetic this is easy to do.
  shared_ptr<sep_params> autoPars(new sep_params(argc,argv));
  sep_params* pars = autoPars.get();
  
  shared_ptr<vel2d>  velObj     (new vel2d(csz_tag_vel,dm::IN,"vel"));
  shared_ptr<srcWav> waveletObj(new srcWav1d(csz_tag_wavelet,dm::IN,"wavelet"));
  
  //Read in wavelet and initial velocity, input 2nd derivative of Gaussian wavelet 
  waveletObj->loadWavelet();
  velObj->loadVel(2);
  velObj->info("Initial velocity model");
  
  shared_ptr<ImgPar> imagingPar(new ImgPar());
  ImgPar& imgPar = *imagingPar;
  //load acquisition parameters
  shared_ptr<AcqPar> acqPar(new AcqPar());
  acqPar->readParam(pars,velObj->domain());
  //Load imaging parameters
  imgPar.readParam(pars,*acqPar);

  shared_ptr<waveOpOneway> modelingObj;
  //Read in the parameters for the modeling Operator
  shared_ptr<waveOpParams>  waveOpPar(new waveOpParams());
  waveOpParams& params = *waveOpPar;
  
  velAnaParam   velAnaParams;
  WEMVA_Param   WEMVAparam;
  params.readParam(pars);
  bool bUseOneWay = params.bOneway();
  VERIFY(bUseOneWay);
  //Here we just focus on use one-way
  if (bUseOneWay) {
    modelingObj.reset(new waveOpOneway(waveOpPar,acqPar,imagingPar,*velObj));
    
    params.read_oneway_prop_param(pars);
    //velAnaParams.dObsFilename = pars->get_string(csz_tag_dataObs); //For Imaging, observed data tag
  } else {
    params.read_twoway_prop_param(pars);
    //calc time parameters
    float maxFreq,vmax,vmin;
    //Read in the peak/center freq of the seismic wavelet
    maxFreq = waveletObj->get_float("peakFreq");
    waveletObj->setPeakFreq(maxFreq);
    vmax = velObj->maxval();
    vmin = velObj->minval();
    params.calcTimeParamsFor2wayModeling(maxFreq, vmax,vmin, velObj->domain());
  }
    
  //Here, need to read in the mask as the random sparse sampling of the migrated gather.
  cubeData<float>* pImageMask = NULL;
  if (tagExist(csz_tag_img_rand_samp_mask)) {
    pImageMask = new cubeData<float>(csz_tag_img_rand_samp_mask,dm::IN,"image random sample mask");
    LOG('Enable Random Sparse Sampling of the migrated Image.');
    pImageMask->init(NULL); //read the mask in.
  }
  params.pImageMask = pImageMask;

  if (!bUseOneWay) { //if not use oneway, then read the recorded data
    //load the observed data
//  velAnaParams.pDataObs = new cubeData<float>(csz_tag_dataObs,dm::IN,"data observed");
//  velAnaParams.pDataObs->init(NULL);
//  velAnaParams.pDataObs ->domain().info("Data obs info");
//  velAnaParams.dObsCube.alloc(velAnaParams.pDataObs->domain());
//  velAnaParams.pDataObs->seq_grabAll(&(velAnaParams.dObsCube));
  } //otherwise let Yaxun's program to take care of it.
  
  read_velAna_param(&velAnaParams,pars);
  read_WEMVA_param(&WEMVAparam,pars,imgPar);

  //assign the gradient mask to image mask. assume vel and img grids are aligned.
  WEMVAparam.imgh0_mask = velAnaParams.gradMaskPreSmooth_;
    
  waveOpOneway* pModeling = modelingObj.get();
  //set the input into WEMVA obj, read the input image and offimg.
  if (!acqPar->b3D()) {
    WEMVAparam.Img0.load("img0",2);
    WEMVAparam.Img0_hx.load("imgh",3); //imgh should have axes (z,x,hx)
    stdVecInt reshape(3),mapping(3);
    reshape[0]=1;reshape[1]=2;reshape[2]=3;
    mapping[0]=1;mapping[1]=3;mapping[2]=2;
    shuffleAxis(WEMVAparam.Img0_hx_t,WEMVAparam.Img0_hx,reshape,mapping);
    WEMVAparam.Img0_hx_t.info("Img0_hx_t");
  } else {
    VERIFY(0);
  }
  
  const cubeDataF& offImg = WEMVAparam.Img0_hx_t, &off0Img = WEMVAparam.Img0;
  WEMVAparam.objFuncType_ = WEMVA_Param::StackPowerObjFunc;
  //make the cropped region of the migrated image.
  vecAxis axes = off0Img.domain().get_axes();
  const axis &ax_z = axes[0], &ax_x = axes[1];
  const axis &ax_h = offImg.domain().get_axis(3);
  //The params for marmousi model
  if (1) {
    //marmousi model
    //axis az_crop = ax_z.return_truncated(680.0f,ax_z.get_coord(ax_z.n-1)); axis ax_crop = ax_x.return_truncated(6200.0f,11800.0f);
    //BP model
    axis az_crop = ax_z.return_truncated(4.8,ax_z.get_coord(ax_z.n-1)); 
    axis ax_crop = ax_x.return_truncated(1.0,19.0);
    axes[0]=az_crop; axes[1]=ax_crop;
    WEMVAparam.croppedImg0_.reset(new cubeDataF(hypercube(axes)));
    cropAlign(*(WEMVAparam.croppedImg0_),WEMVAparam.Img0,false,true);
    cropAlign(WEMVAparam.Img0,*(WEMVAparam.croppedImg0_),false,true);
  }
  
  LOGC("create WEMVA_RMO object");
  shared_ptr<WEMVA_FD> pWEMVAobj = create_WEMVA_imit(pars,ax_h,off0Img.domain(),velAnaParams,
      modelingObj,WEMVAparam,acqPar, params.bOneway(),*velObj); 
                                    
  LOGC("fb2");
  //calcualte backProjData
  //pWEMVAobj->calcObjFunc();
  pWEMVAobj->calcBackProjData();
  
  LOG("fb3");
  //then dump the output
  WEMVAparam.backProjImg_hx.dump("fout.H",true);
  
  LOG("fb4");
  /* /Test AGC
  cubeDataF& img0 = WEMVAparam.Img0;
  const hypercube& shp0 = img0.domain();
  const axis& ax_z = shp0.get_axis(1), ax_x = shp0.get_axis(2);
  seplib::AGC_Param agc_par; agc_par.nw1_ = ax_z.n/2; agc_par.nw2_ = ax_x.n/10; //Window should be relatively big.
  shared_ptr<cubeDataF> img0offBal(new cubeDataF(shp0));
  seplib::AGC(img0, *img0offBal, &agc_par);
  img0.dump("dump5"); img0offBal->dump("dump6");
  */
  return 0;
}

