#include "stdafx.h"
#include "dmRsf.hpp"

#include "srcWav.h"
#include "velFD2d.h"
#include "point3d.hpp"

#include "waveOpParams.h"
#include "waveOpOneway.h"

#include "sep_params.h"
#include "param_entry.h"
#include "WEMVA.h"

#include "stackTraceC.h"
#include "solverNonlinearSearch.h"
#include "commonOps.h"

#include "dipEstimator.h"
#include "ADCIG_Recon.h"
#include "WEMVA_RMO_Imit.h"
#include "util_seplib.h"
#include "misc.h"

///Read parameters
#include "nonLinSolver.param.h"
#include "velAna.param.h"

/*
 * Input migrated offset gather(might be random subsampled,
 then compute the objective function and calculate the back projected data.
 */

//Reflection case

using boost::shared_ptr;
using namespace Yang;
using namespace std;
using namespace seplib;
   
//Doing beam wave-equation tomography for reflection case.
//Use cuda Wave modeling engine

/// The input of the program
/// initial velocity model vel, and the migrated image 0off and multi-offset
/// It calculated the backproj data.
int main (int argc, char *argv[])
{
  initCallStackTrace();
  init_omp(1);
  
  //Read params, params need contain geometry info that is exact the same as the input data. In
  //synthetic this is easy to do.
  shared_ptr<sep_params> autoPars(new sep_params(argc,argv));
  sep_params* pars = autoPars.get();
  
  shared_ptr<vel2d>  velObj     (new vel2d(csz_tag_vel,dm::IN,"vel"));
  shared_ptr<srcWav> waveletObj(new srcWav1d(csz_tag_wavelet,dm::IN,"wavelet"));
  
  //Read in wavelet and initial velocity, input 2nd derivative of Gaussian wavelet 
  waveletObj->loadWavelet();
  velObj->loadVel(2);
  velObj->info("Initial velocity model");
  
  shared_ptr<ImgPar> imagingPar(new ImgPar());
  ImgPar& imgPar = *imagingPar;
  //load acquisition parameters
  shared_ptr<AcqPar> acqPar(new AcqPar());
  acqPar->readParam(pars,velObj->domain());
  //Load imaging parameters
  imgPar.readParam(pars,*acqPar);

  shared_ptr<waveOpOneway> modelingObj;
  //Read in the parameters for the modeling Operator
  shared_ptr<waveOpParams>  waveOpPar(new waveOpParams());
  waveOpParams& params = *waveOpPar;
  
  velAnaParam   velAnaParams;
  WEMVA_Param   WEMVAparam;
  params.readParam(pars);
  bool bUseOneWay = params.bOneway();
  VERIFY(bUseOneWay);
  //Here we just focus on use one-way
  if (bUseOneWay) {
    modelingObj.reset(new waveOpOneway(waveOpPar,acqPar,imagingPar,*velObj));
    
    params.read_oneway_prop_param(pars);
    //velAnaParams.dObsFilename = pars->get_string(csz_tag_dataObs); //For Imaging, observed data tag
  } else {
    params.read_twoway_prop_param(pars);
    //calc time parameters
    float maxFreq,vmax,vmin;
    //Read in the peak/center freq of the seismic wavelet
    maxFreq = waveletObj->get_float("peakFreq");
    waveletObj->setPeakFreq(maxFreq);
    vmax = velObj->maxval();
    vmin = velObj->minval();
    params.calcTimeParamsFor2wayModeling(maxFreq, vmax,vmin, velObj->domain());
  }
    
  //Here, need to read in the mask as the random sparse sampling of the migrated gather.
  cubeData<float>* pImageMask = NULL;
  if (tagExist(csz_tag_img_rand_samp_mask)) {
    pImageMask = new cubeData<float>(csz_tag_img_rand_samp_mask,dm::IN,"image random sample mask");
    LOG('Enable Random Sparse Sampling of the migrated Image.');
    pImageMask->init(NULL); //read the mask in.
  }
  params.pImageMask = pImageMask;

  if (!bUseOneWay) { //if not use oneway, then read the recorded data
    //load the observed data
//  velAnaParams.pDataObs = new cubeData<float>(csz_tag_dataObs,dm::IN,"data observed");
//  velAnaParams.pDataObs->init(NULL);
//  velAnaParams.pDataObs ->domain().info("Data obs info");
//  velAnaParams.dObsCube.alloc(velAnaParams.pDataObs->domain());
//  velAnaParams.pDataObs->seq_grabAll(&(velAnaParams.dObsCube));
  } //otherwise let Yaxun's program to take care of it.
  read_velAna_param(&velAnaParams,pars);
  
  //set the input into WEMVA obj, read the input image and offimg.
  if (!acqPar->b3D()) {
    WEMVAparam.Img0.load("img0",2);
    WEMVAparam.Img0_hx_t.load("imgh",3); //imgh should have axes (z,x,hx)
  } else {
    VERIFY(0);
  }
  
  const cubeDataF& offImg_t = WEMVAparam.Img0_hx_t, &off0Img = WEMVAparam.Img0;
  LOG("a1");
  //read the axis information of the angle domain.
  axis ax; read_axis_from_par(ax,pars,string("ang"));
  axis_deg ax_gamma(ax); axis ax_h(offImg_t.domain().get_axis(2));
  shared_ptr<Param_o2a> par_o2a(new Param_o2a(ax_gamma, ax_h,slantstackYangAA,acqPar->b3D(), params.bOneway()));
  LOG("a2");
  shared_ptr<ADCIG_Recon> adcig_recon(new ADCIG_ReconStomp(par_o2a,true,pars));
  //Load the sub-samp mask for the hx
  WEMVAparam.sampMask_hx_t.reset(new cubeDataF());
  WEMVAparam.sampMask_hx_t->load("subSampMask",3);
  
  LOG("a3");
  shared_ptr<StructTensorParam> par_ST(new StructTensorParam());par_ST->readParam(pars);
  shared_ptr<dipEstimator> est(new dipEstimator(par_ST));
  
  LOG("a4");
  shared_ptr<AGC_Param>           agcPar(new AGC_Param()); agcPar->readParam(pars,off0Img.domain());
  shared_ptr<Param_RMO_Measurer>  par_RMO(new Param_RMO_Measurer()); par_RMO->readParam(pars,off0Img.domain());
  shared_ptr<RMO_Measurer>        RMO_measure(new RMO_Measurer(est,par_RMO,agcPar));
  
  LOG("a5");
  shared_ptr<WEMVA_FD> pWEMVAobj(new WEMVA_RMO_Imit(&velAnaParams, modelingObj, &WEMVAparam,RMO_measure,adcig_recon));
  
  LOGC("fb2");
  //calcualte backProjData
  pWEMVAobj->calcObjFunc();
  pWEMVAobj->calcBackProjData();
  
  //then dump the output
  WEMVAparam.backProjImg_hx.dump("out");
  
  /* //Test AGC
  cubeDataF& img0 = WEMVAparam.Img0;
  const hypercube& shp0 = img0.domain();
  const axis& ax_z = shp0.get_axis(1), ax_x = shp0.get_axis(2);
  seplib::AGC_Param agc_par; agc_par.nw1_ = ax_z.n/2; agc_par.nw2_ = ax_x.n/10; //Window should be relatively big.
  shared_ptr<cubeDataF> img0offBal(new cubeDataF(shp0));
  seplib::AGC(img0, *img0offBal, &agc_par);
  img0.dump("dump5"); img0offBal->dump("dump6");
  */
  return 0;
}

