#include "stdafx.h"
#include "sep_params.h"

#include "srcWav.h"
#include "velFD2d.h"

#include "point3d.hpp"

#include "../param_entry.h"

#include "modelingFD2d_op.h"
#include "SR_manager.h"
#include "paddingFD.h"
#include "paddingFD_damping.h"
#include "boundaryFD.h"
#include "myOMP.h"
#include "stackTraceC.h"

using namespace Yang;
using namespace std;

//Doing Reverse Time migration under the current scenario
int main (int argc, char *argv[])
{
   initCallStackTrace();
   //Init openMP, use 7 cores
   init_omp(1);

   //Read params, params need contain geometry info that is exact the same as the input data. In
   //synthetic this is easy to do.
   auto_ptr<sep_params> autoPars(new sep_params(argc,argv));
   sep_params* pars = autoPars.get();

   auto_ptr<vel2d> velObj(new vel2d(csz_tag_vel,dm::IN,"velocity"));

   auto_ptr<srcWav> waveletObj(new srcWav1d(csz_tag_wavelet,dm::IN,"wavelet"));

   //Read in wavelet and initial velocity
   waveletObj->loadWavelet();
   velObj->loadVel();

   velObj->domain().info("vel");
   
   //Create the modeling Obj for modeling and also RTM(for calculating the gradient)
   auto_ptr<modelingFD2d_op> modelingObj(new modelingFD2d_op());
   //Create the padding and boundary processor
   auto_ptr<SR_managerFD2d>  SR_managerObj(new SR_managerFD2d(waveletObj.get()));
   
   
   //chain two padding method together
   float dampingExtent = 0.01f;
   auto_ptr<paddingFD2dDamping> paddingObj ( new paddingFD2dDamping(NULL,dampingExtent));
   //auto_ptr<paddingFD2dSponge1> paddingObj (new paddingFD2dSponge1(paddingDampingObj.get()));
   cosDampingFunc dampFunc;
   paddingObj->m_pDampingFunc = &dampFunc;
      
   //boundary condition
   auto_ptr<boundaryFD2dAbsorb> boundaryObj(new boundaryFD2dAbsorb());

   //assemble
   modelingObj->assemble(SR_managerObj.get(), boundaryObj.get(),paddingObj.get());

   //read and set the parameters
   FDModelingParams params;
   
   float maxFreq,vmax,vmin;
   params.pVel = velObj.get();
   //Read in the peak/center freq of the seismic wavelet
   maxFreq = waveletObj->get_float("peakFreq");
   waveletObj->setPeakFreq(maxFreq);
   
   vmax = velObj->maxval();
   vmin = velObj->minval();
   params.velModelShape = velObj->domain();
   params.modelGridShape = params.velModelShape;
   
   //also get the time param from the params
   params.bFreeSurfaceOnTop = pars->get_bool(csz_freeSurface);
   params.stencilWidth = pars->get_int(csz_stenWidth);
   params.bUse4thOrder= false;
   
   params.ndim = params.velModelShape.get_ndim();
   params.offx = params.stencilWidth;params.offz = params.stencilWidth;
   
   params.m_nShots 	= pars->get_int(csz_nShots);
   params.mShot1_ix = pars->get_int(csz_shot1_ix);
   params.mShot1_iz = pars->get_int(csz_shot1_iz);
   params.mShotIncX = pars->get_int(csz_shotIncX);
      
   params.m_nPad = pars->get_int(csz_npad);

   //Read in Absorption type
   int itype = pars->get_int(csz_bdType);
   switch (itype)
   {
      case 1:
         params.m_bdType = abs_ord1;
         break;
      case 2:
         params.m_bdType = abs_ord2;
         break;
      case 3:
         params.m_bdType = abs_Dave;
         break;
   case 0:
	  params.m_bdType = abs_null;
	  break;
   default:
         seperr("!unknown boundary condition type!\n");
   }
   
   //RTM params
   params.mImposeRecvType = AsForce;

   params.pSrcSnapshot 	= new cubeData<float>(csz_tag_snapshot,dm::OUT,"src snapshot");
   params.pRecvSnapshot	= new cubeData<float>(csz_tag_recv_snapshot,dm::OUT,"src snapshot");
   params.pRsfSeisRec	= new cubeData<float>(csz_tag_seisRec,dm::IN,"seis-record");
   params.pImages	= new cubeData<float>(csz_tag_RTM_image,dm::OUT,"RTM_Image_Cube");

   //calc time parameters
   params.timeParam.dtComp = pars->get_float(csz_dtComp);
   params.timeParam.dtSnap = pars->get_float(csz_dtSnap);
   params.timeParam.dtImaging = pars->get_float(csz_dtImaging,params.timeParam.dtRecord);   
   //Get the size of seismic record
   params.pRsfSeisRec->init(NULL);
   //Always assume the recording starts from zero time. !! 
   params.timeParam.dtRecord = params.pRsfSeisRec->domain().get_d(1);
   params.timeParam.totalTime= (params.pRsfSeisRec->domain().get_n(1)-1) * params.timeParam.dtRecord;
   
   prt("dim of pRsfSeisRec = %d\n", params.pRsfSeisRec->domain().get_ndim());
   params.m_nShots = params.pRsfSeisRec->domain().get_n(3);
   
   calcTimeParamsFor2wayModeling(&(params.timeParam),maxFreq,vmax,vmin,params.modelGridShape);  
   VERIFY(equalf(params.pRsfSeisRec->domain().get_o(1), 0.0f));

   //The receiver geometry should also comes from the seismic record
   params.mTrace1ix = pars->get_int(csz_trace1_ix);
   params.mTrace1iz = pars->get_int(csz_trace1_iz);
   params.mTraceIncX = lround(params.pRsfSeisRec->domain().get_d(2) / velObj->domain().get_d(2));
   params.m_nTraces = params.pRsfSeisRec->domain().get_n(2);

   params.pFinalImage = new cubeData<float>();
   //Init the operator
   modelingObj->initOperatorEx(&params, acRTM);
   
   //do the modeling
   modelingObj->RTM();

   //then save the necessary wavefields
   //After modeling, save the snapshot and seisrecord
   params.pFinalImage->dump(csz_tag_final_image);
	  
   params.pSrcSnapshot->unInit();
   params.pRsfSeisRec->unInit();
   params.pRecvSnapshot->unInit();
   params.pFinalImage->unInit();
   
   
   srcWav* pWav = waveletObj.release();
   //pWav->getWavelet()->dump("dump");
   delete pWav;
   
}
