#include "stdafx.h"
#include "dmRsf.hpp"

#include "srcWav.h"
#include "velFD2d.h"
#include "point3d.hpp"

#include "waveOpParams.h"
#include "waveOpOneway.h"

#include "param_entry.h"
#include "WEMVA.h"

#include "stackTraceC.h"
#include "solverNonlinearSearch.h"
#include "commonOps.h"

///Read parameters
#include "nonLinSolver.param.h"
#include "velAna.param.h"

/**
 * For beam focusing, Wave Equation Tomography, Just do the 
 * non-linear line search 
 */

//Reflection case

using namespace Yang;
using namespace std;

   
//Doing beam wave-equation tomography for reflection case.
//Use cuda Wave modeling engine

/// The input of the program
/// initial velocity model vel
/// observed data data_obs along with its geometry

int main (int argc, char *argv[])
{
  initCallStackTrace();
  init_omp(1);
  
  //Read params, params need contain geometry info that is exact the same as the input data. In
  //synthetic this is easy to do.
  shared_ptr<sep_params> autoPars(new sep_params(argc,argv));
  sep_params* pars = autoPars.get();
  
  shared_ptr<vel2d>  velObj     (new vel2d(csz_tag_vel,dm::IN,"vel"));
  shared_ptr<srcWav> waveletObj(new srcWav1d(csz_tag_wavelet,dm::IN,"wavelet"));
  
  //Read in wavelet and initial velocity, input 2nd derivative of Gaussian wavelet 
  waveletObj->loadWavelet();
  velObj->loadVel(2);
  velObj->info("Initial velocity model");
  
  shared_ptr<ImgPar> imagingPar(new ImgPar());
  ImgPar& imgPar = *imagingPar;
  //load acquisition parameters
  shared_ptr<AcqPar> acqPar(new AcqPar());
  acqPar->readParam(pars,velObj->domain());
  //Load imaging parameters
  imgPar.readParam(pars,*acqPar);

  shared_ptr<waveOpOneway> modelingObj;
  //Read in the parameters for the modeling Operator
  shared_ptr<waveOpParams>  waveOpPar(new waveOpParams());
  waveOpParams& params = *waveOpPar;
  
  velAnaParam   velAnaParams;
  WEMVA_Param   WEMVAparam;
  params.readParam(pars);
  bool bUseOneWay = params.bOneway();
  //Here we just focus on use one-way
  if (bUseOneWay) {
    modelingObj.reset(new waveOpOneway(waveOpPar,acqPar,imagingPar,*velObj));
    
    params.read_oneway_prop_param(pars);
    velAnaParams.dObsFilename = pars->get_string(csz_tag_dataObs); //For Imaging, observed data tag
  } else {
    params.read_twoway_prop_param(pars);
    //calc time parameters
    float maxFreq,vmax,vmin;
    //Read in the peak/center freq of the seismic wavelet
    maxFreq = waveletObj->get_float("peakFreq");
    waveletObj->setPeakFreq(maxFreq);
    vmax = velObj->maxval();
    vmin = velObj->minval();
    params.calcTimeParamsFor2wayModeling(maxFreq, vmax,vmin, velObj->domain());
  }
    
  //Here, need to read in the mask as the random sparse sampling of the migrated gather.
  cubeData<float>* pImageMask = NULL;
  if (tagExist(csz_tag_img_rand_samp_mask)) {
    pImageMask = new cubeData<float>(csz_tag_img_rand_samp_mask,dm::IN,"image random sample mask");
    LOG('Enable Random Sparse Sampling of the migrated Image.');
    pImageMask->init(NULL); //read the mask in.
  }
  params.pImageMask = pImageMask;

  if (!bUseOneWay) { //if not use oneway, then read the recorded data
    //load the observed data
    velAnaParams.pDataObs = new cubeData<float>(csz_tag_dataObs,dm::IN,"data observed");
    velAnaParams.pDataObs->init(NULL);
    velAnaParams.pDataObs ->domain().info("Data obs info");
    velAnaParams.dObsCube.alloc(velAnaParams.pDataObs->domain());
    velAnaParams.pDataObs->seq_grabAll(&(velAnaParams.dObsCube));
  } //otherwise let Yaxun's program to take care of it.
  
  read_velAna_param(&velAnaParams,pars);
  
  waveOpOneway* pModeling = modelingObj.get();

  //Now compute the imitated tomo operator
  cubeData<float> imag,dimg,ds;
  imag.load("refl1",3); dimg.load("refl2",3);
  ds.alloc(velObj->domain());
  pModeling->TomoImit_op(&imag,&dimg,&ds);
  
  ds.dump("out");
  
  //clean up
  if (!bUseOneWay) {
    velAnaParams.pDataObs->unInit();
  }
  return 0;
}

