#include "stdafx.h"
#include "dmRsf.hpp"

#include "srcWav.h"
#include "velFD2d.h"
#include "point3d.hpp"

#include "modelingFD2d_op.h"
#include "SR_manager.h"
#include "paddingFD.h"
#include "paddingFD_damping.h"
#include "boundaryFD.h"

#include "param_entry.h"
#include "myOMP.h"

#include "stackTraceC.h"

#include "FWI_FD.h"
#include "FWI_FD_beamFocus.h"
#include "FWI_FD_DSO.h"
//#include "../misc.h"

/**
 * For beam focusing, Wave Equation Tomography, gradient 
 * calculation 
 */


using namespace Yang;
using namespace std;

   
//Doing beam wave-equation tomography for transmission case
/// The input of the program
/// initial velocity model vel
/// observed data data_obs along with its geometry

int main (int argc, char *argv[])
{
   //Init openMP, use 7 cores
   init_omp(1);
   signal(SIGSEGV,handler);
   signal(SIGABRT,handler);
   
   //Read params, params need contain geometry info that is exact the same as the input data. In
   //synthetic this is easy to do.
   auto_ptr<sep_params> autoPars(new sep_params(argc,argv));
   sep_params* pars = autoPars.get();

   auto_ptr<vel2d> velObj(new vel2d(csz_tag_vel,dm::IN,"vel"));
   auto_ptr<srcWav> waveletObj(new srcWav1d(csz_tag_wavelet,dm::IN,"wavelet"));

   //Read in wavelet and initial velocity, input 2nd derivative of Gaussian wavelet 
   waveletObj->loadWavelet();
   velObj->loadVel();
   
   //Create the modeling Obj for modeling and also RTM(for calculating the gradient)
   auto_ptr<modelingFD2d_op> modelingObj(new modelingFD2d_op());
   //Create the padding and boundary processor
   auto_ptr<SR_managerFD2d>  SR_managerObj(new SR_managerFD2d(waveletObj.get()));
   
   //chain two padding method together
   //float dampingExtent = 0.5f;
   //auto_ptr<paddingFD2dDamping> paddingDampingObj ( new paddingFD2dDamping(NULL,dampingExtent));
   auto_ptr<paddingFD2dSponge1> paddingObj (new paddingFD2dSponge1(NULL));
   //cosDampingFunc dampFunc;
   //paddingDampingObj->m_pDampingFunc = &dampFunc;
		
   //boundary condition
   auto_ptr<boundaryFD2dAbsorb> boundaryObj(new boundaryFD2dAbsorb());

   //assemble
   modelingObj->assemble(SR_managerObj.get(), boundaryObj.get(),paddingObj.get());

   //Create another RTM imaging operator to calculate the RTM
   //modelingFD2d_op modelingObj(new modelingFD2d_op(),acRTM);
   
   //Read in the parameters for the modeling Operator
   FDModelingParams params;
   FWI_FDParam FWIParam;
   //load the observed data
   FWIParam.pDataObs = new cubeData<float>(csz_tag_dataObs,dm::IN,"data observed");
   FWIParam.pDataObs ->domain().info("Data obs info");

   float maxFreq,vmax,vmin;
   params.pVel = velObj.get();
   params.pWavelet = waveletObj.get();
   //Read in the peak/center freq of the seismic wavelet
   maxFreq = waveletObj->get_float("peakFreq");
   waveletObj->setPeakFreq(maxFreq);
   
   vmax = velObj->maxval();
   vmin = velObj->minval();
   params.modelGridShape = velObj->domain();
   
   //also get the time param from the params
   params.bFreeSurfaceOnTop = pars->get_bool(csz_freeSurface);
   params.stencilWidth = pars->get_int(csz_stenWidth);
   params.bUse4thOrder= pars->get_bool(csz_use4thOrderInTime);

   //set size of mShapeWavfieeld
   setWavfieldShape(params.mShapeWavfield,params.stencilWidth,params.modelGridShape);
   
   params.ndim = params.mShapeWavfield.get_ndim();
   params.offx = params.stencilWidth;params.offz = params.stencilWidth;
   params.m_nShots 	= pars->get_int(csz_nShots);
   params.mShot1_ix = pars->get_int(csz_shot1_ix);
   params.mShot1_iz = pars->get_int(csz_shot1_iz);
   params.mShotIncX = pars->get_int(csz_shotIncX);

   params.mTrace1ix = pars->get_int(csz_trace1_ix);
   params.mTrace1iz = pars->get_int(csz_trace1_iz);
   params.mTraceIncX = pars->get_int(csz_traceIncX);
   params.m_nTraces = pars->get_int(csz_nTraces);
   params.m_nPad = pars->get_int(csz_npad);

   //Read in Absorption type
   int itype = pars->get_int(csz_bdType);
   switch (itype)
   {
      case 1:
         params.m_bdType = abs_ord1;
         break;
      case 2:
         params.m_bdType = abs_ord2;
         break;
      case 3:
         params.m_bdType = abs_Dave;
         break;
      default:
         seperr("!unknown boundary condition type!\n");
   }
   
   //Here I set the data rsf objects
   params.mImposeRecvType = AsForce;

   params.pSrcSnapshot 	= new cubeData<float>("",dm::INOUT,"src snapshot");
   params.pRecvSnapshot	= new cubeData<float>("",dm::INOUT,"recv snapshot");
   //The gradient of the local objective function
   params.pRTM_image	= new cubeData<float>("",dm::INOUT,"1st Gradient");

   //calc time parameters
   params.timeParam.dtComp = pars->get_float(csz_dtComp);
   params.timeParam.dtSnap = pars->get_float(csz_dtSnap);
   params.timeParam.dtImaging = pars->get_float(csz_dtImaging,params.timeParam.dtRecord);   
   //Get the size of seismic record
   FWIParam.pDataObs->init(NULL);
   //Always assume the recording starts from zero time. !! 
   params.timeParam.dtRecord = FWIParam.pDataObs->domain().get_d(1);
   params.timeParam.totalTime= (FWIParam.pDataObs->domain().get_n(1)-1) * params.timeParam.dtRecord;
   
   params.timeParam.dtComp = pars->get_float(csz_dtComp);
   params.timeParam.dtSnap = pars->get_float(csz_dtSnap);
   params.timeParam.dtImaging = pars->get_float(csz_dtImaging,params.timeParam.dtRecord);
   
   //calc time parameters
   calcTimeParamsFor2wayModeling(&(params.timeParam),maxFreq,vmax,vmin,params.modelGridShape);

   //Load the observed data
   FWIParam.dObsCube.alloc(FWIParam.pDataObs->domain());
   FWIParam.pDataObs->seq_grabAll(&(FWIParam.dObsCube));
   
   //Load the corrlation window
   FWIParam.iStartLag = pars->get_int(csz_corrStartLag);
   FWIParam.iEndLag = pars->get_int(csz_corrEndLag);
   FWIParam.pModelingParam = &params;
   FWIParam.beamSize = pars->get_int(csz_subGroupSize,-1); //e.g. beam size;
   
   int objFuncIndex = pars->get_int("objFunc",1);
   int moveOutFuncType = pars->get_int("moveOutType",2);
   switch(moveOutFuncType) {
   case 1: 
	  FWIParam.moveOutType = FWI_FDParam::Linear;
	  break;
   case 2:
	  FWIParam.moveOutType = FWI_FDParam::Quadratic;
	  break;
   default:
	  seperr("wrong moveOutType for beam focusing");
	  break;
   }
   
   //Create FWIObj
   FWI_FD* pFWIObj = NULL;

   bool bTaperEdge = true;
   
   float v0,vTrue;
   v0 = pars->get_float("v0",-1);
   vTrue = pars->get_float("vTrue",-1);
   FWI_FD_beamFocusY *p = NULL;
   switch(objFuncIndex) {
   case 1: 
	  prt("using classical FWI obj Function");
	  FWIParam.objFuncType = FWI_FDParam::classic; 
	  pFWIObj = new FWI_FD_classic(modelingObj.get(), &FWIParam);
	  break;
	  
   case 2: 
	  prt("using YiLuo FWI obj Function");
	  FWIParam.objFuncType = FWI_FDParam::YiLuo; 
	  pFWIObj = new FWI_FD_YL(modelingObj.get(), &FWIParam); break;
	  
   case 3: 
	  prt("using beam focusing1 FWI obj Function\n");
	  FWIParam.objFuncType = FWI_FDParam::beamFocusing1;
	  pFWIObj = new FWI_FD_beamFocus1(modelingObj.get(), &FWIParam,bTaperEdge); break;
	  
   case 4:
	  prt("using beam focusing2 FWI obj Function\n");
	  FWIParam.objFuncType = FWI_FDParam::beamFocusing2;
	  pFWIObj = new FWI_FD_beamFocus2(modelingObj.get(), &FWIParam,bTaperEdge); break;
   case 5:
	  prt("using beam focusing Yang's FWI obj Function, one param picking\n");
	  FWIParam.objFuncType = FWI_FDParam::beamFocusingY1;
	  p = new FWI_FD_beamFocusY1(modelingObj.get(), &FWIParam,bTaperEdge);
	  VERIFY(v0>0); p->m_v0 = v0; p->m_vTrue = vTrue;
	  pFWIObj = p; break;
   case 6:
	  prt("using beam focusing Yang's FWI obj Function, two param picking\n");
	  FWIParam.objFuncType = FWI_FDParam::beamFocusingY2;
	  p = new FWI_FD_beamFocusY2(modelingObj.get(), &FWIParam,bTaperEdge);
	  VERIFY(v0>0); p->m_v0 = v0; p->m_vTrue = vTrue;
	  pFWIObj = p; break;
   case 7: //DSO
	  prt("\n Using DSO objective function");
	  FWIParam.objFuncType = FWI_FDParam::DSO;
	  pFWIObj = new FWI_FD_DSO1(modelingObj.get(), &FWIParam,bTaperEdge); break;
   default:
	  seperr("wrong  objFunction index");
	  exit(-1);
   }
   
   LOG(calc the obj function);
   
   if (objFuncIndex==5 || objFuncIndex==6) {
   axis ax_a, ax_b;
   ax_a.o = pars->get_float("o_a");
   ax_a.d = pars->get_float("d_a");
   ax_a.n = pars->get_int("n_a");
   ax_b.o = pars->get_float("o_b");
   ax_b.d = pars->get_float("d_b");
   ax_b.n = pars->get_int("n_b");
   ax_a.label = "param_a";   ax_b.label = "param_b";
   pFWIObj->m_pFWIParam->ax_a = ax_a;
   pFWIObj->m_pFWIParam->ax_b = ax_b;
   }
   
   //Calculate the data
   pFWIObj->calcObjFunc();
   //FWIParam.dCalCube.dump("dump1");
   //calculate the bacjProject data
   
   pFWIObj->m_pFWIParam->C_cube.dump("corr");
   if (objFuncIndex==5 || objFuncIndex==6) {
	  pFWIObj->m_pFWIParam->pickedCurvatureCube.dump("aCube");
	  pFWIObj->m_pFWIParam->pickedSlopeCube.dump("bCube");
   }

   LOG(calc the back project data);
   pFWIObj->calcBackProjData();
   
   FWIParam.backProjCube.dump("backProjData");
   //calculate gradient this time
   FWIParam.velGradCube.alloc( params.modelGridShape);
   pFWIObj->calcVelGradient(FWIParam.velGradCube);
   
   LOG(after calc vel Grad);
   //Dump the src and recv wavefield.   
   cubeData<float>* pCubef= dynamic_cast<cubeData<float>*> (params.pSrcSnapshot);
   pCubef->dump(csz_tag_srcSnapshot);
   pCubef= dynamic_cast<cubeData<float>*> (params.pRecvSnapshot);
   pCubef->dump(csz_tag_recvSnapshot);
   
   //Save the gradient to see
   FWIParam.velGradCube.dump("grad");
   LOG(after dump the gradient);
   //clean up
   FWIParam.pDataObs->unInit();
   SAFE_DELETE(pFWIObj);
   
   //We use auto_ptr now, no need to delete those ptrs.
   return 0;
}

