#include "stdafx.h"
#include "dmRsf.hpp"

#include "srcWav.h"
#include "velFD2d.h"
#include "point3d.hpp"

#include "modelingFD2d_op.h"
#include "SR_manager.h"
#include "paddingFD.h"
#include "paddingFD_damping.h"
#include "boundaryFD.h"

#include "param_entry.h"
#include "myOMP.h"

#include "FWI_FD.h"
#include "FWI_objFunc.h"
#include "solverNonlinearSearch.h"
//#include "../misc.h"

/**
 * For beam focusing, Wave Equation Tomography, Just do the 
 * non-linear line search 
 */


using namespace Yang;
using namespace std;

   
//Doing beam wave-equation tomography for transmission case
/// The input of the program
/// initial velocity model vel
/// observed data data_obs along with its geometry

int main (int argc, char *argv[])
{
   //Init openMP, use 7 cores
   initCallStackTrace();

   //Read params, params need contain geometry info that is exact the same as the input data. In
   //synthetic this is easy to do.
   auto_ptr<sep_params> autoPars(new sep_params(argc,argv));
   sep_params* pars = autoPars.get();

   auto_ptr<vel2d> velObj(new vel2d(csz_tag_vel,dm::IN,"vel"));
   auto_ptr<srcWav> waveletObj(new srcWav1d(csz_tag_wavelet,dm::IN,"wavelet"));

   //Read in wavelet and initial velocity, input 2nd derivative of Gaussian wavelet 
   waveletObj->loadWavelet();
   velObj->loadVel();
   
   //Create the modeling Obj for modeling and also RTM(for calculating the gradient)
   auto_ptr<modelingFDgpu2d_op> modelingObj(new modelingFDgpu2d_op());
   int gpuID = pars->get_int(csz_gpuID,0);
   modelingObj->setGPU_ID(gpuID);
   
   //Create the padding and boundary processor
   auto_ptr<SR_managerFD2d>  SR_managerObj(new SR_managerFD2d(waveletObj.get()));
   
   //chain two padding method together
   //float dampingExtent = 0.5f;
   //auto_ptr<paddingFD2dDamping> paddingDampingObj ( new paddingFD2dDamping(NULL,dampingExtent));
   auto_ptr<paddingFD2dSponge1> paddingObj (new paddingFD2dSponge1(NULL));
   //cosDampingFunc dampFunc;
   //paddingDampingObj->m_pDampingFunc = &dampFunc;
		
   //boundary condition
   auto_ptr<boundaryFD2dAbsorb> boundaryObj(new boundaryFD2dAbsorb());

   //assemble
   modelingObj->assemble(SR_managerObj.get(), boundaryObj.get(),paddingObj.get());

   //Create another RTM imaging operator to calculate the RTM
   //modelingFD2d_op modelingObj(new modelingFD2d_op(),acRTM);
   
   //Read in the parameters for the modeling Operator
   FDModelingParams params;
   FWI_FDParam FWIParam;
   //load the observed data
   FWIParam.pDataObs = new cubeData<float>(csz_tag_dataObs,dm::IN,"data observed");
   //FWIParam.pDataObs ->domain().info("Data obs info");

   float maxFreq,vmax,vmin;
   params.pVel = velObj.get();
   params.pWavelet = waveletObj.get();
   //Read in the peak/center freq of the seismic wavelet
   maxFreq = waveletObj->get_float("peakFreq");
   waveletObj->setPeakFreq(maxFreq);
   
   vmax = velObj->maxval();
   vmin = velObj->minval();
   params.modelGridShape = velObj->domain();
   params.velModelShape = velObj->domain();
   
   //also get the time param from the params
   params.bFreeSurfaceOnTop = pars->get_bool(csz_freeSurface);
   params.stencilWidth = pars->get_int(csz_stenWidth);
   params.bUse4thOrder= false;

   //set size of mShapeWavfieeld
   setWavfieldShape(params.mShapeWavfield,params.stencilWidth,params.modelGridShape);
   
   params.ndim = params.mShapeWavfield.get_ndim();
   params.offx = params.stencilWidth;params.offz = params.stencilWidth;
   params.m_nShots 	= pars->get_int(csz_nShots);
   params.mShot1_ix = pars->get_int(csz_shot1_ix);
   params.mShot1_iz = pars->get_int(csz_shot1_iz);
   params.mShotIncX = pars->get_int(csz_shotIncX);

   params.mTrace1ix = pars->get_int(csz_trace1_ix);
   params.mTrace1iz = pars->get_int(csz_trace1_iz);
   params.mTraceIncX = pars->get_int(csz_traceIncX);
   params.m_nTraces = pars->get_int(csz_nTraces);
   params.m_nPad = pars->get_int(csz_npad);

   //Read in Absorption type
   int itype = pars->get_int(csz_bdType);
   switch (itype)
   {
      case 1:
         params.m_bdType = abs_ord1;
         break;
      case 2:
         params.m_bdType = abs_ord2;
         break;
      case 3:
         params.m_bdType = abs_Dave;
         break;
      default:
         seperr("!unknown boundary condition type!\n");
   }
   
   //Here I set the data rsf objects
   params.mImposeRecvType = AsForce;

   params.pSrcSnapshot 	= new cubeData<float>("",dm::INOUT,"src snapshot");
   params.pRecvSnapshot	= new cubeData<float>("",dm::INOUT,"recv snapshot");
   //The gradient of the local objective function
   params.pRTM_image	= new cubeData<float>("",dm::INOUT,"1st Gradient");

   //calc time parameters
   params.timeParam.dtComp = pars->get_float(csz_dtComp);
   params.timeParam.dtSnap = pars->get_float(csz_dtSnap);
   params.timeParam.dtImaging = pars->get_float(csz_dtImaging,params.timeParam.dtRecord);   
   //Get the size of seismic record
   FWIParam.pDataObs->init(NULL);
   //Always assume the recording starts from zero time. !! 
   params.timeParam.dtRecord = FWIParam.pDataObs->domain().get_d(1);
   params.timeParam.totalTime= (FWIParam.pDataObs->domain().get_n(1)-1) * params.timeParam.dtRecord;
   
   params.timeParam.dtComp = pars->get_float(csz_dtComp);
   params.timeParam.dtSnap = pars->get_float(csz_dtSnap);
   params.timeParam.dtImaging = pars->get_float(csz_dtImaging,params.timeParam.dtRecord);
   
   //calc time parameters
   calcTimeParamsFor2wayModeling(&(params.timeParam),maxFreq,vmax,vmin,params.modelGridShape);

   //Load the observed data
   FWIParam.dObsCube.alloc(FWIParam.pDataObs->domain());
   FWIParam.pDataObs->seq_grabAll(&(FWIParam.dObsCube));
   
   //Load the corrlation window
   FWIParam.iStartLag = pars->get_int(csz_corrStartLag);
   FWIParam.iEndLag = pars->get_int(csz_corrEndLag);
   FWIParam.pModelingParam = &params;
   FWIParam.beamSize = pars->get_int(csz_subGroupSize,-1); //e.g. beam size;
   
   int objFuncIndex = pars->get_int("objFunc",1);
   int moveOutFuncType = pars->get_int("moveOutType",2);
   switch(moveOutFuncType) {
   case 1: 
	  FWIParam.moveOutType = FWI_FDParam::Linear;
	  break;
   case 2:
	  FWIParam.moveOutType = FWI_FDParam::Quadratic;
	  break;
   default:
	  seperr("wrong moveOutType for beam focusing");
	  break;
   }
   
   //Create FWIObj
   FWI_FD* pFWIObj = NULL;

   switch(objFuncIndex) {
   case 1: 
	  prt("using classical FWI obj Function\n");
	  FWIParam.objFuncType = FWI_FDParam::classic; 
	  pFWIObj = new FWI_FD_classic(modelingObj.get(), &FWIParam);
	  break;
	  
   case 2: 
	  prt("using YiLuo FWI obj Function\n");
	  FWIParam.objFuncType = FWI_FDParam::YiLuo; 
	  pFWIObj = new FWI_FD_YL(modelingObj.get(), &FWIParam); break;
	  
   case 3: 
	  prt("using beam focusing1 FWI obj Function\n");
	  FWIParam.objFuncType = FWI_FDParam::beamFocusing1;
	  pFWIObj = new FWI_FD_beamFocus1(modelingObj.get(), &FWIParam); break;
	  
   case 4:
	  prt("using beam focusing2 FWI obj Function\n");
	  FWIParam.objFuncType = FWI_FDParam::beamFocusing2;
	  pFWIObj = new FWI_FD_beamFocus2(modelingObj.get(), &FWIParam); break;
	  
   default:
	  seperr("wrong  objFunction index\n");
	  exit(-1);
   }
   
   //After the initialization of params, init the non-linear solver with observed data
   objFunc* pObjFunc = new FWI_objFunc(pFWIObj);
   //solverNonlinearSearch(objFunc* pObj, const gVecComponent* pInitModel);
   gVecComponent initVelModelVec, finalModelVec;
   
   params.pVel->copyTo(initVelModelVec);
   params.pVel->allocMemoryTo(finalModelVec);
   
   solverNonlinearSearch nonLinSolver(pObjFunc, &initVelModelVec);
   
   //invert the model
   
   stdVecFloat finalErrorVec;
   int finalIter;
   int maxIter = pars->get_int("maxNonLinIter",5);
   float errorAllowed = pars->get_float("errorAllowed",1e-6);
   float maxPerturbScale = pars->get_float("maxPerturbScale",0.01); // 1 percent of the RMS value
   
   SR_managerObj->init(&params);
   SR_managerObj->getVelModelMask(FWIParam.velModelMask);
   
   nonLinSolver.solve( finalModelVec, finalErrorVec, finalIter, 
								maxIter, errorAllowed, maxPerturbScale);
   
   //Save result
   finalModelVec.dump("out");
   finalModelVec.unInit();

   //Dump the final data
   FWIParam.dCalCube.dump("dump9");
   //clean up
   FWIParam.pDataObs->unInit();
   SAFE_DELETE(pFWIObj);
   SAFE_DELETE(pObjFunc);
   
   return 0;
}

