/* 2D Time domain finite difference kernel
	By: Abdullah AlTheyab
	Ref: similar to Paulius code but for 2D
 */

#define BLOCK_SIZE 16
#define FAT 4
#define COEFFS_SIZE 9

#define C0  0
#define CZ1 1
#define CX1 2
#define CZ2 3
#define CX2 4
#define CZ3 5
#define CX3 6
#define CZ4 7
#define CX4 8

#define C_C00(d) (8.0/(5.0*(d)*(d)))
#define get_coeffs(d1,d2) {-1025.0/576.0*(C_C00(d1)+C_C00(d2)), C_C00(d1), C_C00(d2),-C_C00(d1)/8.0,-C_C00(d2)/8.0,C_C00(d1)/63.0,C_C00(d2)/63.0,-C_C00(d1)/896.0,-C_C00(d2)/896.0}

//define the constants for finite difference coefficients
__constant__ float coeffs[COEFFS_SIZE];
__constant__ float dx_gpu,dz_gpu;

//wave propagation kernel, input are the GPU memory slices for t-1,t,and t+1, also give dpdx2 seperately
//the last term sums up the wavefield energy of p0
extern "C" __global__ void wave_kernel(float *p0, float *p1, float *p2, float *dpdx2=NULL, float *p_sq=NULL){
   //define shared memory region (need to include the boundary
	__shared__ float p1s[BLOCK_SIZE+2*FAT][BLOCK_SIZE+2*FAT];
	
	//calculate the coordinates
	int ig=FAT+blockIdx.x*BLOCK_SIZE+threadIdx.x; 	// global z coordinates
	int jg=FAT+blockIdx.y*BLOCK_SIZE+threadIdx.y; 	// global x coordinates

	if (ig >= n1gpu) return;
	if (jg >= n2gpu) return;
	
	int il=FAT+threadIdx.x; 			  			// thread-local z coordinate
	int jl=FAT+threadIdx.y;				 			// thread-local x coordinate
	
	//find the index of local grid point (ig,jg) in global coordinates.
	int addr=dloc(ig,jg);		  		  			// linear memroy address

	//copy the current wavefield values to shared memory
	p1s[jl][il]=p1[addr];

	//For the threads on the boundary, they are also responsible for reading the boundary value pts.
	if(threadIdx.y < FAT){
		p1s[jl-FAT][il]=p1[dloc(ig,jg-FAT)];
		if ( jg+BLOCK_SIZE < n2gpu ) {
		   p1s[jl+BLOCK_SIZE][il]=p1[dloc(ig,jg+BLOCK_SIZE)];
		}
	}
	if(threadIdx.x < FAT){
		p1s[jl][il-FAT]=p1[dloc(ig-FAT, jg)];
		if (ig+BLOCK_SIZE < n1gpu) {
		   p1s[jl][il+BLOCK_SIZE]=p1[dloc(ig+BLOCK_SIZE, jg)];
		}
	}
	
	//when it comes to doing the calculation, tick out the points outside the inner grid
	if (ig >= n1gpu-FAT) return;
	if (jg >= n2gpu-FAT) return;
	
	//wait until all share memory load is ready.
	__syncthreads();
	
	//float temp=-expf(bc_bgpu+min4(jg,n2gpu-jg,n1gpu-ig,ig)*bc_agpu);
	//look for the look up table.
	float temp = 0.0f;
  float dpdxf = 0.0f;
	//temp = tex1D(damp_ref,min4(jg,n2gpu-jg,n1gpu-ig,ig));
	if (NULL != p_sq) {
		temp=p1s[jl][il];
		p_sq[addr] += temp*temp; 
	}
	temp = 0.0f;
	
	//float pc = p1s[jl][il];
	//This is an analytical expression for that temp, should use a search table instead.
	
	//if set temp=0.0, then there is no absorption, otherwise temp has to be negative.
	//The modified acoustic equation with damping now becomes
	//let deltaT be time step for discretization)
	// dp^2/dt^2*(1/v^2) = dp^2/dx^2 + (temp*deltaT)dp/dt
	// need to normalize deltaT.
    dpdxf = tex2D(vel_ref, ig-FAT, jg-FAT ) * (
			coeffs[CX4]*(p1s[jl-4][il]+p1s[jl+4][il])
			+ coeffs[CZ4]*(p1s[jl][il-4]+p1s[jl][il+4])
			+ coeffs[CX3]*(p1s[jl-3][il]+p1s[jl+3][il])
			+ coeffs[CZ3]*(p1s[jl][il-3]+p1s[jl][il+3])
			+ coeffs[CX2]*(p1s[jl-2][il]+p1s[jl+2][il])
			+ coeffs[CZ2]*(p1s[jl][il-2]+p1s[jl][il+2])
			+ coeffs[CX1]*(p1s[jl-1][il]+p1s[jl+1][il])
			+ coeffs[CZ1]*(p1s[jl][il-1]+p1s[jl][il+1])
			+ coeffs[C0]*p1s[jl][il]);
			//+ temp*(p1s[jl][il]-p0[addr]));
	p2[addr]= dpdxf+ (p1s[jl][il]+p1s[jl][il]-p0[addr]);
  if (NULL != dpdx2) {
    dpdx2[addr] = dpdxf;
  }
}

//boundary 1D parallel abs_ord1par, not very good
//update the boundary values in p2 wavefield(at it+1) based on the p1 wavefield (at it)
//The thread grid spans n1gpu*n2gpu
extern "C" __global__ void bnd1d_kernel(float *p2, float *p1) {
   	//calculate the coordinates
	int ig=blockIdx.x*BLOCK_SIZE+threadIdx.x; 		// global z coordinates
	int jg=blockIdx.y*BLOCK_SIZE+threadIdx.y; 	// global x coordinates

	//if not at boundary, return
	if (ig >= FAT && ig < n1gpu-FAT && jg >= FAT && jg < n2gpu-FAT ) return;	
	//if Out of range, return
	if (ig >= n1gpu) return;	
	if (jg >= n2gpu) return;
	
    
	float vdtdx,delp;
	int addr=dloc(ig,jg);
    
    //int il=FAT+threadIdx.x; 			  			// thread-local z coordinate
	//int jl=FAT+threadIdx.y;				 			// thread-local x coordinate
	
	//find the index of local grid point (ig,jg) in global coordinates.
    /*
	//Boundary Cond abs_ord1par
	//left, right
	if (jg < FAT) {
	   vdtdx = tex2D(vel_ref,ig-FAT,0); vdtdx = sqrt(vdtdx); vdtdx /= dx_gpu; delp = 1.2f * vdtdx;
	   p2[addr] = (1-delp)* (p1[addr]) + delp* p1[dloc(ig,jg+1)];
	}
	if (jg >= n2gpu-FAT ) {
	   vdtdx = tex2D(vel_ref,ig-FAT,n2gpu-2*FAT-1); vdtdx = sqrt(vdtdx); vdtdx /= dx_gpu; delp = 1.2f * vdtdx;
	   p2[addr] = (1-delp)* (p1[addr]) + delp* p1[dloc(ig,jg-1)];
	}
	
	//top bottom
	if (ig < FAT) {
		vdtdx = tex2D(vel_ref,0,jg-FAT); vdtdx = sqrt(vdtdx); vdtdx /= dz_gpu; delp = 1.2f * vdtdx;
		p2[addr] = (1-delp)* (p1[addr]) + delp* p1[dloc(ig+1,jg)];
	 }
	 if (ig >= n1gpu-FAT ) {
		vdtdx = tex2D(vel_ref,n1gpu-2*FAT-1,jg-FAT); vdtdx = sqrt(vdtdx); vdtdx /= dz_gpu; delp = 1.2f * vdtdx;
		p2[addr] = (1-delp)* (p1[addr]) + delp* p1[dloc(ig-1,jg)];
	 }
	*/
	 
	 // Alternative cond_abs_ord1 
	//left, right
	if (jg < FAT) {
	   vdtdx = tex2D(vel_ref,ig-FAT,0); vdtdx = sqrt(vdtdx); vdtdx /= dx_gpu; delp = 1.2f * vdtdx;
	   p2[addr] = (1-delp)/(1+delp)* (p1[addr]-p2[dloc(ig,jg+1)]) +
	   p1[dloc(ig,jg+1)];
	}	
	if (jg >= n2gpu-FAT ) {
	   vdtdx = tex2D(vel_ref,ig-FAT,n2gpu-2*FAT-1); vdtdx = sqrt(vdtdx); vdtdx /= dx_gpu; delp = 1.2f * vdtdx;
	   p2[addr] = (1-delp)/(1+delp)* (p1[addr]-p2[dloc(ig,jg-1)]) +
	   p1[dloc(ig,jg-1)];
	}
	
	//top bottom
	if (ig < FAT) {
		vdtdx = tex2D(vel_ref,0,jg-FAT); vdtdx = sqrt(vdtdx); vdtdx /= dz_gpu; delp = 1.2f * vdtdx;
	    p2[addr] = (1-delp)/(1+delp)* (p1[addr]-p2[dloc(ig+1,jg)]) +
	    p1[dloc(ig+1,jg)];
	 }
	 if (ig >= n1gpu-FAT ) {
		vdtdx = tex2D(vel_ref,n1gpu-2*FAT-1,jg-FAT); vdtdx = sqrt(vdtdx); vdtdx /= dz_gpu; delp = 1.2f * vdtdx;
	    p2[addr] = (1-delp)/(1+delp)* (p1[addr]-p2[dloc(ig-1,jg)]) +
	    p1[dloc(ig-1,jg)];
	 }
	 return;
}

//Boundary Cond, abs_ord1, do the left and right. Problem there is a zero freq component staying in the wavefield
//The thread grid is of size (n1gpu-2*FAT) X 2*FAT, block size BLOCK_SIZE X 2*FAT
extern "C" __global__ void bnd1d_kernel2_0(float *p2, float *p1) {
   	__shared__ float p2s[4*FAT][BLOCK_SIZE];
   	//calculate the coordinates
	int ic=blockIdx.x*BLOCK_SIZE+threadIdx.x; 		// global z coordinates
	int jc=blockIdx.y*BLOCK_SIZE+threadIdx.y; 	// global x coordinates

	//if Out of range, return
	if (ic >= n1gpu-2*FAT) return;
	
	int il=threadIdx.x; 			  			// thread-local z coordinate
	int jl=threadIdx.y;				 			// thread-local x coordinate

	int addr = 0; float vdtdx = 0;
	//coord in the wavefield
	int ig = ic + FAT;

	if (jc<FAT) {
	   //read p2(ig,jg+1) and p1(ig,jg+1) into shared memory
	   p2s[jl][il] = p2[dloc(ig,jc+1)];
	   p2s[jl+2*FAT][il] = p1[dloc(ig,jc+1)];
	   addr = dloc(ig,jc);
	   vdtdx = tex2D(vel_ref,ic,0);
	} 
	else //right bound
	{
	   //read p2(ig,jg-1) and p1(ig,jg-1) into shared memory
	   p2s[jl][il] = p2[dloc(ig,n2gpu-2*FAT-1+jc)];
	   p2s[jl+2*FAT][il] = p1[dloc(ig,n2gpu-2*FAT-1+jc)];
	   addr = dloc(ig,n2gpu-2*FAT+jc);
	   vdtdx = tex2D(vel_ref,ic,n2gpu-2*FAT-1);
	}
	
	__syncthreads();
	
   float delp;
   vdtdx = sqrt(vdtdx); vdtdx /= dx_gpu; delp = 1.2f * vdtdx;
   p2[addr] = (1-delp)/(1+delp)* (p1[addr]-p2s[jl][il]) + p2s[jl+2*FAT][il];
   //p2[addr] = (1-delp)* (p1[addr]) + delp*p2s[jl+2*FAT][il];
}

//Boundary Cond abs_ord1, do the top and bottom
//The thread grid is of size (n2gpu-2*FAT) X 2*FAT, block size BLOCK_SIZE X 2*FAT
extern "C" __global__ void bnd1d_kernel2_1(float *p2, float *p1) {
   	__shared__ float p2s[4*FAT][BLOCK_SIZE];
   	//calculate the coordinates
	int jc=blockIdx.x*BLOCK_SIZE+threadIdx.x; 	// global X coordinates 
	int ic=blockIdx.y*BLOCK_SIZE+threadIdx.y; 	// global Z coordinates

	//if Out of range, return
	if (jc >= n2gpu-2*FAT) return;
	
	int jl=threadIdx.x; 			  			// thread-local x coordinate
	int il=threadIdx.y;				 			// thread-local z coordinate

	int addr = 0; float vdtdx = 0;
	//coord in the wavefield
	int jg = jc + FAT;

	//Top boundary
	if (ic<FAT) {
	   if(bAbsorbTop_gpu) {
	   //read p2(ig+1,jg) and p1(ig+1,jg) into shared memory
	   p2s[il][jl] = p2[dloc(ic+1,jg)];
	   p2s[il+2*FAT][jl] = p1[dloc(ic+1,jg)];
	   } else {
		  p2s[il][jl] = 0.0f;
		  p2s[il+2*FAT][jl] = 0.0f;
	   }
	   addr = dloc(ic,jg);
	   vdtdx = tex2D(vel_ref,0,jc);
	}
	//bottom bound
	if (ic >= FAT) 
	{
	   //read p2(ig-1,jg) and p1(ig-1,jg) into shared memory
	   p2s[il][jl] = p2[dloc(n1gpu-2*FAT-1+ic,jg)];
	   p2s[il+2*FAT][jl] = p1[dloc(n1gpu-2*FAT-1+ic,jg)];
	   addr = dloc(n1gpu-2*FAT+ic,jg);
	   vdtdx = tex2D(vel_ref,n1gpu-2*FAT-1,jc);
	}
	
	__syncthreads();

   
   float delp;
   vdtdx = sqrt(vdtdx); vdtdx /= dz_gpu; delp = 1.2f * vdtdx;
   p2[addr] = (1-delp)/(1+delp)* (p1[addr]-p2s[il][jl]) + p2s[il+2*FAT][jl];
}

#define MCOEF_ABSPONGE1 0.15
//Absorbing padding layer
//Grid domian n1gpu-2*FAT X n2gpu-2*FAT. block size BLOCK_SIZEXBLOCK_SIZE
extern "C" __global__ void paddingSponge1_kernel(float *p2, float *p1) {
   	//calculate the coordinates
	int ig=blockIdx.x*BLOCK_SIZE+threadIdx.x; 		// global z coordinates
	int jg=blockIdx.y*BLOCK_SIZE+threadIdx.y; 		// global x coordinates

	//if not at the padding boundary, return
	if (ig >= npad_gpu && ig < (n1gpu-2*FAT-npad_gpu) && jg >= npad_gpu && jg < (n2gpu-2*FAT-npad_gpu) ) return;	
	//if Out of range, return
	if (ig >= n1gpu-2*FAT) return;	
	if (jg >= n2gpu-2*FAT) return;
		
	//find the index of local grid point (ig,jg) in global coordinates.
	
	//location in wavefield coord.
	int addr=dloc(ig+FAT,jg+FAT);

	float vdtDdx=0.0f, mCoef=0.0f;
	vdtDdx = tex2D(vel_ref,ig,jg);
	vdtDdx = sqrt(vdtDdx);
	float fp2 = p2[addr];
	//Boundary Cond abs_ord1par
	//left
	if (jg < npad_gpu) {
	   vdtDdx /= dx_gpu; 
	   mCoef = (npad_gpu-jg) / float(npad_gpu) * MCOEF_ABSPONGE1 * vdtDdx;
	   p2[addr] -= (mCoef * (fp2 - p1[addr] + vdtDdx*(p1[dloc(ig+FAT,jg-1+FAT)]-p1[addr])));
	}
	//right
	if (jg >= n2gpu-2*FAT-npad_gpu) {
	   vdtDdx /= dx_gpu;
	   mCoef = (jg-(n2gpu-2*FAT-npad_gpu)+1) / float(npad_gpu) * MCOEF_ABSPONGE1 * vdtDdx;
	   p2[addr] -= (mCoef * (fp2 - p1[addr] + vdtDdx*(p1[dloc(ig+FAT,jg+1+FAT)]-p1[addr])));
	}
	
	//bottom
	if (ig >= n1gpu-2*FAT-npad_gpu) {
		vdtDdx /= dz_gpu;
	    mCoef = (ig-(n1gpu-2*FAT-npad_gpu)+1) / float(npad_gpu) * MCOEF_ABSPONGE1 * vdtDdx;
		p2[addr] -= (mCoef * (fp2 - p1[addr] + vdtDdx*(p1[dloc(ig+1+FAT,jg+FAT)]-p1[addr])));
	 }
	if (ig < npad_gpu && bAbsorbTop_gpu) {
	   vdtDdx /= dz_gpu;
	   mCoef = (npad_gpu-ig) / float(npad_gpu) * MCOEF_ABSPONGE1 * vdtDdx;
	   p2[addr] -= (mCoef * (fp2 - p1[addr] + vdtDdx*(p1[dloc(ig-1+FAT,jg+FAT)]-p1[addr])));
	}
	return;
}
